""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import os
try: import json
except ImportError: import simplejson as json
import agent_util
import datetime
import random
import time
import re

KEY_FSR_DETAILS = 'fsr_details'
KEY_FSR_PAYLOAD = 'payload'

KEY_CONNECTOR_DETAILS = 'connector_details'

CONFIG_KEY_SERVICES = 'fortisoar_services'
CONFIG_KEY_EXTN_SME_SERVICES = 'fortisoar_external_sme_services'
CONFIG_KEY_CACHE_TIME = 'fortisoar_cache_time'
CONFIG_KEY_IS_EXTN_SME = 'fortisoar_is_external_sme'
CONFIG_KEY_MQ_CERT = 'fortisoar_mq_cert'
CONFIG_KEY_NGINX_CERT = 'fortisoar_nginx_cert'
CONFIG_KEY_CONNECTOR_EXCLUSION_LIST = 'fortisoar_connector_exclusion_list'

TEXTKEY_SERVICE_STATUS = 'fortisoar.service_status'
TEXTKEY_PRIMARY_DB_SIZE = 'fortisoar.primary_database_size'
TEXTKEY_WORKFLOW_DB_SIZE = 'fortisoar.playbook_logs_database_size'
TEXTKEY_AUDIT_DB_SIZE = 'fortisoar.audit_logs_database_size'
TEXTKEY_PLAYBOOK_EXECUTED_IN_LAST_ONE_HOUR = 'fortisoar.playbooks_triggered_per_hour'
TEXTKEY_ALERT_CREATED_PER_DAY = 'fortisoar.alerts_created_per_day'
TEXTKEY_REPLICATOR_LOG_SIZE = 'fortisoar.tenancy_data_replication_lag'
TEXTKEY_QUEUED_WORKFLOW_COUNT = 'fortisoar.queued_playbooks'
TEXTKEY_CERT_EXPIRY = 'fortisoar.cert_expiry'
TEXTKEY_CLUSTER_HEARTBEAT_LAG = 'fortisoar.cluster_heartbeat_lag'
TEXTKEY_REPLICATION_LAG = 'fortisoar.cluster_data_replication_lag'
TEXTKEY_SELF_HEARTBEAT_LAG = 'fortisoar.self_heartbeat_lag'
TEXTKEY_LICENSE_EXPIRY = 'fortisoar.license_expiry'
TEXTKEY_CONNECTOR_HEALTH = 'fortisoar.connector_health'
TEXTKEY_AGENT_HEALTH = 'fortisoar.agent_health'
TEXTKEY_WAL_SIZE = "fortisoar.wal_size"

DEFAULT_CACHE_TIME = 300
DEFAULT_CERT_MQ = "/opt/cyops/configs/rabbitmq/ssl/server/cert.pem"
DEFAULT_CERT_NGINX = "/etc/nginx/ssl/server.leaf.crt"
DEFAULT_CERT_ROOT = "/etc/pki/ca-trust/source/anchors/cs.loc.root.crt"

OPTION_CERT_MQ = "RabbitMQ"
OPTION_CERT_NGINX = "Nginx"
OPTION_CERT_ROOT = "Root"

CERT_LIST_ENTERPRISE = [OPTION_CERT_MQ, OPTION_CERT_NGINX, OPTION_CERT_ROOT]
CERT_LIST_SME = [OPTION_CERT_MQ]

NODE_HEALTH_REPLICATION_STAT = 'replication_stat'
NODE_HEALTH_MISC_STAT = 'misc_stat'

SERVICE_LIST = [
    "rabbitmq-server",
    "elasticsearch",
    "postgresql-14",
    "nginx",
    "php-fpm",
    "cyops-auth",
    "uwsgi",
    "celeryd",
    "celerybeatd",
    "cyops-tomcat",
    "cyops-search",
    "cyops-postman",
    "cyops-integrations-agent",
    "fsr-workflow",
    "fsr-api-consumer",
    "cyops-ha"
]

SERVICE_LIST_EXTN_SME = ["rabbitmq-server"]

license_info = None

FILE_FSR_DATABASE = '/var/lib/fm-agent/.fortisoar-db'

def _write_json(json_path, json_data):
    json_dump = json.dumps(json_data, indent=4)
    fh = open(json_path, "w")
    fh.write(json_dump)
    fh.close()

def _read_json(json_path):
    if not os.path.exists(json_path):
        return None
    try:
        with open(json_path, 'r') as json_contents:
            return json.load(json_contents)
    except Exception as e:
        return None

def write_database(json_data):
    _write_json(FILE_FSR_DATABASE, json_data)

def read_database(key):
    json_contents = _read_json(FILE_FSR_DATABASE)
    if json_contents and key in json_contents.keys():
        return json_contents[key]
    return None

def get_active_services_list():
    cmd = "sudo csadm services --get-active-list"
    ret, out = agent_util.execute_command(cmd)
    if ret != 0:
        return None
    return out.strip().split(" ")

def get_certificate_exiry(cert_file):
    cmd = "openssl x509 -in " + cert_file + " -noout -enddate"
    ret, out = agent_util.execute_command(cmd)
    if ret != 0:
        return None
    datestr = out.split('=')[-1]
    datespl = datestr.split('\n',1)[0]
    dateformat = datetime.datetime.strptime(datespl, '%b %d %H:%M:%S %Y %Z')
    diff = dateformat - datetime.datetime.utcnow()
    return diff.days

def convert_to_mb(input_size_in_bytes):
    try:
        return float("{:.2f}".format(input_size_in_bytes/1024.0**2))
    except Exception as e:
        pass
    return None

def merge_dict(a, b):
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dict(a[key], b[key])
            elif a[key] == b[key]:
                pass  # same leaf value
        else:
            a[key] = b[key]

def convert_to_dict(input_str, logger=None):
    try:
        payload = json.loads(input_str)
        if not payload:
            # payload is empty
            return None
    except Exception as e:
        if logger:
            logger.error("error while converting fortisoar string to dict [%s]" % str(e))
        return None
    return payload

def call_api(endpoint, logger=None, is_das=False):
    cmd = '/opt/cyops/scripts/api_caller.py --endpoint "' + endpoint + '" --silent'
    if is_das:
        cmd = cmd + " --das"
    ret, output = agent_util.execute_command(cmd)
    if ret != 0:
        if logger:
            logger.error(output)
        return None
    return convert_to_dict(output, logger)

def fetch_fsr_payload(logger):
    # We are using CLI instead of an API, since API fails on passive node.
    cmd = "sudo /bin/csadm ha show-health --json --no-local-db-point"
    ret, output = agent_util.execute_command(cmd)
    if ret != 0:
        logger.error("failed to get the fortisoar payload with response as [%s]" % output)
        return None
    response = convert_to_dict(output, logger)
    if response:
        return response[0]
    return None

def get_license_info(logger=None):
    global license_info
    if license_info:
        return license_info
    endpoint = "https://localhost:8443/license/?param=license_details"
    license_info = call_api(endpoint, logger, True)
    return license_info

def is_mssp(logger=None):
    license_info = get_license_info(logger)
    if license_info and license_info['details']['is_distributed']:
        return True
    return False

# license expiry is not applicable if license type is Trial(Extension) or Perpetual
def is_license_metric_applicable(logger=None):
    license_info = get_license_info(logger)
    if license_info and (license_info['type'].lower() not in ['trial(extension)', 'perpetual']):
        return True
    return False

def get_health_connector(config, logger=None):
    connector_name = config['name']
    connector_version = config['version']
    connector_config_id = config['configuration']['config_id']
    endpoint = "https://localhost/api/integration/connectors/healthcheck/" + connector_name + "/" + connector_version \
                + "/" + "?config=" + connector_config_id
    logger.info("Getting health for connector %s %s %s" %(connector_name, connector_version, connector_config_id))
    time_t1 = datetime.datetime.now()
    health = call_api(endpoint, logger)
    time_t2 = datetime.datetime.now()
    logger.info("Done health check for connector %s %s %s %s" %(connector_name, connector_version, connector_config_id, (time_t2-time_t1)))
    return health

def get_health_agent(agent_name, logger=None):
    # Observed that FortiMonitor ask for agent health with input as None when no agent is configured.
    # It unnecessary logs stacktrace in logfile. Hence check if agent_name, and then proceed.
    if not agent_name:
        return None
    endpoint = "https://localhost/api/3/agents?name=" + agent_name + "&__selectFields=configurationHealth"
    agent_health = call_api(endpoint, logger)
    if agent_health is None:
        return None
    if len(agent_health['hydra:member']) == 0:
        return None
    return agent_health['hydra:member'][0]['configurationHealth']['itemValue']

# TODO This API only list the connectors for self. It does not consider an agent connectors.
# If we need to monitor an agent connectors, we need to pass agent=<agent id > in query.
# For now, we are just checking self node connectors.
#
def get_configured_connectors(cls):
    api = "https://localhost/api/integration/connectors/?format=json&configured=true&page_size=1000"
    all_connectors = call_api(api, cls.log)
    if all_connectors is None:
        return None
    all_connectors = all_connectors['data']
    configured_connectors = {}
    for connector in all_connectors:
        if (not connector['config_count'] > 0) or (not connector['active']):
            continue
        connector_name = connector['name']
        connector_version = connector['version']
        for connector_config in connector['configuration']:
            configured_connectors[connector_name + " " + connector_config['name']] = {
                'name': connector_name,
                'version': connector_version,
                'configuration' : connector_config
            }
    return configured_connectors

def get_configured_agents(cls):
    api = "https://localhost/api/3/agents"
    all_agents = call_api(api, cls.log)
    if all_agents is None:
        return None
    all_agents = all_agents['hydra:member']
    configured_agents = []
    for agent in all_agents:
        configured_agents.append(agent['name'])
    return configured_agents

def get_connector_from_cache(cls, connector_name, cache_time=3600):
    configured_connectors = read_database(KEY_CONNECTOR_DETAILS)
    if configured_connectors and (connector_name in configured_connectors.keys()):
        return configured_connectors[connector_name]
    return None

def get_common_metadata(cls, config, status, msg, is_external_sme=False):
    certs_list = CERT_LIST_ENTERPRISE
    if is_external_sme:
        certs_list = CERT_LIST_SME
    metadata = {
        TEXTKEY_CERT_EXPIRY: {
            "label": "Certificate expiry",
            "options": certs_list,
            "status": status,
            "error_message": msg,
            "unit": "days"
        }
    }
    return metadata

def get_mssp_metadata(cls, config, status, msg):
    metadata = {
        TEXTKEY_REPLICATOR_LOG_SIZE: {
            "label": "Tenancy data replication lag",
            "options": None,
            "status": status,
            "error_message": msg,
            "unit": "MB"
        }
    }
    return metadata

def get_external_sme_metadata(cls, config, status, msg):
    service_list = SERVICE_LIST_EXTN_SME
    if CONFIG_KEY_EXTN_SME_SERVICES in config.keys():
        service_list = config[CONFIG_KEY_EXTN_SME_SERVICES].split(",")
    else:
        csadm_active_services = get_active_services_list()
        if csadm_active_services is not None:
            service_list =csadm_active_services
    cls.log.info("fortisoar services to monitor %s" % service_list)
    extn_sme_metadata = {
        TEXTKEY_SERVICE_STATUS: {
            "label": "Service status",
            "options": service_list,
            "unit": "boolean",
            "status": status,
            "error_message": msg
        }
    }
    common_metadata = get_common_metadata(cls, config, status, msg, True)
    merge_dict(extn_sme_metadata, common_metadata)
    return extn_sme_metadata

def get_enterprise_metadata(cls, config, status, msg):
    service_list = SERVICE_LIST
    if CONFIG_KEY_SERVICES in config.keys():
        service_list = config[CONFIG_KEY_SERVICES].split(",")
    else:
        csadm_active_services = get_active_services_list()
        if csadm_active_services is not None:
            service_list = csadm_active_services
    cls.log.info("fortisoar services to monitor %s" % service_list)

    other_cluster_nodes = []
    fsr_payload = fetch_fsr_payload(cls.log)
    enable_schema = True
    options_schema = None
    if enable_schema:
        options_schema = {
            'node': 'string',
            'status': 'string',
            'role': 'string',
            'resource': 'string'
        }
    if fsr_payload is not None:
        if 'heartbeat' in fsr_payload['misc_stat'].keys():
            is_primary = fsr_payload['misc_stat']['is_primary']
            nodes = fsr_payload['misc_stat']['heartbeat']
            if is_primary:
                # Primary monitor heartbeats of all secondary nodes
                for node in nodes:
                    if enable_schema:
                        other_cluster_nodes.append(
                            {
                                'node': node['node_name'],
                                'status': node['status'],
                                'role': node['role'],
                                'resource': node['node_name'] + "," + node['status'] + "-" + node['role']
                            }
                        )
                    else:
                        other_cluster_nodes.append(node['node_name'])
            else:
                # Secondary node monitor heartbeat if only primary node
                if enable_schema:
                    other_cluster_nodes.append(
                        {
                            'node': nodes[0]['node_name'],
                            'status': nodes[0]['status'],
                            'role': nodes[0]['role'],
                            'resource': nodes[0]['node_name'] + "," + nodes[0]['status']+ "-" + nodes[0]['role']
                        }
                    )
                else:
                    other_cluster_nodes.append(nodes[0]['node_name'])

    cls.log.info("fortisoar other cluster nodes %s" % other_cluster_nodes)

    connector_list = []
    configured_connectors = get_configured_connectors(cls)
    if configured_connectors is not None:
        connector_list = list(configured_connectors.keys())
        if CONFIG_KEY_CONNECTOR_EXCLUSION_LIST in config.keys():
            exclusion_connectors = config[CONFIG_KEY_CONNECTOR_EXCLUSION_LIST].split(",")
            for exclusion_connector in exclusion_connectors:
                cls.log.info("exclusion fortisoar connector %s" % exclusion_connector)
                connector_list = [value for value in connector_list if not value.startswith(exclusion_connector + " ")]
                pattern = re.compile(exclusion_connector + " ")
                matched_keys = {key for key, value in configured_connectors.iteritems() if pattern.match(key)}
                for key in matched_keys:
                    configured_connectors.pop(key, None)
    cls.log.info("fortisoar connectors list %s" % connector_list)
    write_database({KEY_CONNECTOR_DETAILS:configured_connectors})
    agent_list = []
    configured_agents = get_configured_agents(cls)
    if configured_agents is not None:
        agent_list = configured_agents
    cls.log.info("fortisoar agents list %s" % agent_list)

    # add dictionaries for each metric that can be measured.  Each key in the
    # metadata dictionary is a textkey for that metric (all lower case letters, numbers
    # and underscores, no spaces or special characters)
    enterprise_metadata = {
        TEXTKEY_SERVICE_STATUS: {
            "label": "Service status",
            "options": service_list,
            "unit": "boolean",
            "status": status,
            "error_message": msg
        },
        TEXTKEY_PRIMARY_DB_SIZE: {
            "label": "Primary database size",
            "options": None,
            "status": status,
            "error_message": msg,
            "unit": "MB"
        },
        TEXTKEY_WORKFLOW_DB_SIZE: {
            "label": "Playbook logs database size",
            "options": None,
            "status": status,
            "error_message": msg,
            "unit": "MB"
        },
        TEXTKEY_AUDIT_DB_SIZE: {
            "label": "Audit logs database size",
            "options": None,
            "status": status,
            "error_message": msg,
            "unit": "MB"
        },
        #TEXTKEY_PLAYBOOK_EXECUTED_IN_LAST_ONE_HOUR: {
        #    "label": "Playbooks triggered per hour",
        #    "options": None,
        #    "status": status,
        #    "error_message": msg,
        #    "unit": "count"
        #},
        #TEXTKEY_ALERT_CREATED_PER_DAY: {
        #    "label": "Alerts created per day",
        #    "options": None,
        #    "status": status,
        #    "error_message": msg,
        #    "unit": "count"
        #},
        TEXTKEY_QUEUED_WORKFLOW_COUNT: {
            "label": "Queued playbooks",
            "options": None,
            "status": status,
            "error_message": msg,
            "unit": "count"
        },
        TEXTKEY_CLUSTER_HEARTBEAT_LAG: {
            "label": "Cluster heartbeat lag",
            "options": other_cluster_nodes,
            "options_schema": options_schema,
            "unit": "seconds",
            "status": status,
            "error_message": msg
        },
        TEXTKEY_REPLICATION_LAG: {
            "label": "Cluster data replication lag",
            "options": other_cluster_nodes,
            "options_schema": options_schema,
            "unit": "MB",
            "status": status,
            "error_message": msg
        },
        TEXTKEY_SELF_HEARTBEAT_LAG: {
            "label": "Self heartbeat lag",
            "options": None,
            "unit": "seconds",
            "status": status,
            "error_message": msg
        },
        TEXTKEY_CONNECTOR_HEALTH: {
            "label": "Connector health",
            "options": connector_list,
            "unit": "boolean",
            "status": status,
            "error_message": msg
        },
        TEXTKEY_AGENT_HEALTH: {
            "label": "Agent health",
            "options": agent_list,
            "unit": "boolean",
            "status": status,
            "error_message": msg
        },
        TEXTKEY_WAL_SIZE: {
            "label": "PostgreSQL WAL size",
            "options": None,
            "unit": "GB",
            "status": status,
            "error_message": msg
        }
    }
    common_metadata = get_common_metadata(cls, config, status, msg)
    merge_dict(enterprise_metadata, common_metadata)
    if is_mssp(cls.log):
        mssp_metadata = get_mssp_metadata(cls, config, status, msg)
        merge_dict(enterprise_metadata, mssp_metadata)
    if is_license_metric_applicable(cls.log):
        enterprise_metadata[TEXTKEY_LICENSE_EXPIRY] = {
            "label": "License expiry",
            "options": None,
            "status": status,
            "error_message": msg,
            "unit": "days"
        }
    return enterprise_metadata

def get_boolean_config_key(config, key, default=False):
    value = config.get(key, default)
    if type(value) == bool:
        return value
    if value in [ "yes", "true", "t", "1"]:
        return True
    return False

# Class to an appropriate name, must end in "Plugin"
class FortiSOARPlugin(agent_util.Plugin):

    textkey = "fortisoar"
    label = "FortiSOAR"

    @classmethod
    def get_metadata(self, config):
        if not agent_util.which("csadm"):
            self.log.info("fortisoar is not installed")
            status = agent_util.UNSUPPORTED
            msg = "fortisoar is not installed on the instance"
            return {}

        status = agent_util.SUPPORTED
        msg = None
        is_external_sme = get_boolean_config_key(config, CONFIG_KEY_IS_EXTN_SME)
        if is_external_sme:
            return get_external_sme_metadata(self, config, status, msg)
        return get_enterprise_metadata(self, config, status, msg)

    def check(self, textkey, data, config):
        if not textkey.startswith("fortisoar."):
            return None
        ##########
        # Currently, there are 3 types of metrics.
        # TYPE 1: Metrics which are runtime and don't need any cache. for eg. services status, certificate expiry
        # TYPE 2: Metrics which using cache and not monitored in HA service. for eg. connector health
        # TYPE 3: Metrics which using cache and monitored in HA service. for eg. replication lag, heartbeat.
        #########

        #########
        # TYPE 1: Metrics below from here are runtime and don't need any cache
        #######
        if textkey == TEXTKEY_SERVICE_STATUS:
            srvc_name = str(data)
            cmd = "systemctl is-enabled " + srvc_name + ".service"
            ret, output = agent_util.execute_command(cmd)
            if ret != 0:
                # service is not enabled
                return None
            cmd = "systemctl --all | grep " + srvc_name + ".service | awk {'print $3'} | sort -u"
            ret, output = agent_util.execute_command(cmd)
            if ret != 0:
                # 0 means service not running
                self.log.info("service is not running " + srvc_name)
                return 0
            if output.strip() == "active":
                # 1 means service is running
                return 1
            self.log.info("service is not running " + srvc_name)
            return 0
        elif textkey == TEXTKEY_CERT_EXPIRY:
            input_cert = data
            if input_cert == OPTION_CERT_MQ:
                mq_cert = DEFAULT_CERT_MQ
                if CONFIG_KEY_MQ_CERT in config.keys():
                    mq_cert = config[CONFIG_KEY_MQ_CERT]
                return get_certificate_exiry(mq_cert)
            elif input_cert == OPTION_CERT_NGINX:
                nginx_cert = DEFAULT_CERT_NGINX
                if CONFIG_KEY_NGINX_CERT in config.keys():
                    nginx_cert = config[CONFIG_KEY_NGINX_CERT]
                return get_certificate_exiry(nginx_cert)
            elif input_cert == OPTION_CERT_ROOT:
                return get_certificate_exiry(DEFAULT_CERT_ROOT)
        elif textkey == TEXTKEY_LICENSE_EXPIRY:
            license_info = get_license_info(self.log)
            # If license type is Perpetual or Trial(Extension), we set the remaining days as -1
            if license_info and (int(license_info['remaining_days']) != -1):
                return license_info['remaining_days']
            return None

        #########
        # Metrics above from here are runtime and don't need any cache
        #######

        #######
        # TYPE 2: Metrics below from here using cache. These metrics are not monitored in HA service.
        # Cache is refreshed for it every metadata rebuild using flag file.
        #######

        if textkey == TEXTKEY_CONNECTOR_HEALTH:
            input_connector_name = data
            connector_config = get_connector_from_cache(self, input_connector_name)
            if connector_config is None:
                return None
            connector_health = get_health_connector(connector_config, self.log)
            if connector_health is None:
                return None
            connector_status = connector_health['status'].lower()
            if connector_status == "available":
                return 1
            return 0
        elif textkey == TEXTKEY_AGENT_HEALTH:
            input_agent_name = data
            agent_health = get_health_agent(input_agent_name, self.log)
            if agent_health is None:
                return None
            if agent_health.lower() == "remote node connected":
                return 1
            return 0

        #######
        # Metrics above from here using cache. These metrics are not monitored in HA service.
        # Cache is refreshed for it every metadata rebuild using flag file.
        #######

        #######
        # TYPE 3: Metrics below from here need cache. These metrics are monitored in HA service.
        # Cache is refreshed for it every DEFAULT_CACHE_TIME. i.e 5 mins
        #######

        cache_time = DEFAULT_CACHE_TIME
        if CONFIG_KEY_CACHE_TIME in config.keys():
            cache_time = int(config[CONFIG_KEY_CACHE_TIME])

        fsr_payload = self.get_fsr_payload_from_cache(textkey, cache_time)
        if fsr_payload is None:
            return None

        if textkey == TEXTKEY_PRIMARY_DB_SIZE:
            if (NODE_HEALTH_MISC_STAT in fsr_payload.keys()) and 'database' in fsr_payload[NODE_HEALTH_MISC_STAT].keys():
                return convert_to_mb(fsr_payload[NODE_HEALTH_MISC_STAT]['database']['venom']['size'])
            return None
        elif textkey == TEXTKEY_WORKFLOW_DB_SIZE:
            if (NODE_HEALTH_MISC_STAT in fsr_payload.keys()) and 'database' in fsr_payload[NODE_HEALTH_MISC_STAT].keys():
                return convert_to_mb(fsr_payload[NODE_HEALTH_MISC_STAT]['database']['sealab']['size'])
            return None
        elif textkey == TEXTKEY_AUDIT_DB_SIZE:
            if (NODE_HEALTH_MISC_STAT in fsr_payload.keys()) and 'database' in fsr_payload[NODE_HEALTH_MISC_STAT].keys():
                return convert_to_mb(fsr_payload[NODE_HEALTH_MISC_STAT]['database']['gateway']['size'])
            return None
        elif textkey == TEXTKEY_PLAYBOOK_EXECUTED_IN_LAST_ONE_HOUR:
            if 'workflow' in fsr_payload.keys():
                return int(fsr_payload['workflow']['playbook']['total'])
            return None
        elif textkey == TEXTKEY_ALERT_CREATED_PER_DAY:
            return random.randint(1, 900)
        elif textkey == TEXTKEY_REPLICATOR_LOG_SIZE:
            if 'remote_node_communication' in fsr_payload.keys():
                return convert_to_mb(int(fsr_payload['remote_node_communication']))
            return None
        elif textkey == TEXTKEY_QUEUED_WORKFLOW_COUNT:
            if 'workflow' in fsr_payload.keys():
                return int(fsr_payload['workflow']['mq']['total_size'])
            return None
        elif textkey == TEXTKEY_CLUSTER_HEARTBEAT_LAG:
            if not ('heartbeat' in fsr_payload[NODE_HEALTH_MISC_STAT].keys()):
                return None
            nodes = fsr_payload[NODE_HEALTH_MISC_STAT]['heartbeat']
            input_node_name = str(data).split(",")[0]
            for node in nodes:
                if input_node_name == node['node_name']:
                    return node['delay']
            return None
        elif textkey == TEXTKEY_REPLICATION_LAG:
            if not (NODE_HEALTH_REPLICATION_STAT in fsr_payload.keys()):
                return None
            nodes = fsr_payload[NODE_HEALTH_REPLICATION_STAT]
            if not nodes:
                return None
            input_node_name = str(data).split(",")[0]
            for node in nodes:
                if ('node_name' in node.keys()) and (input_node_name == node['node_name']):
                    return convert_to_mb(int(node['total_lag_in_bytes']))
            return None
        elif textkey == TEXTKEY_SELF_HEARTBEAT_LAG:
            return int(time.time()) - int(fsr_payload['timestamp'])
        elif textkey == TEXTKEY_WAL_SIZE:
            if not ('wal' in fsr_payload[NODE_HEALTH_MISC_STAT].keys()):
                return None
            return fsr_payload[NODE_HEALTH_MISC_STAT]['wal']['size_gb']

    def get_fsr_payload_from_cache(self, textkey, cache_time=DEFAULT_CACHE_TIME):
        cache = None
        try:
            cache = self.get_cache_results(KEY_FSR_DETAILS, KEY_FSR_PAYLOAD)[0]
        except Exception as e:
            self.log.error("error while getting fortisoar payload from cache [%s]" % str(e))

        # cache_time (in secs) is delta which tells about how much older cached this data is.
        if cache and cache[0] < cache_time:
            return cache[1]

        # Cache is older than cache_time, refresh it.
        fsr_payload = fetch_fsr_payload(self.log)
        if fsr_payload is not None:
            self.cache_result(KEY_FSR_DETAILS, KEY_FSR_PAYLOAD, fsr_payload, replace=True)
            self.log.info("caching fortisoar payload having epoch timestamp as %s" % fsr_payload['timestamp'])
            return fsr_payload
        #
        # Unable to fetch the FortiSOAR payload from the DB. This indicate DB is down.
        # Now, we should just report the heartbeat. If metric is other than heartbeat, then don't report it. i.e
        # return None for metric other than heartbeat. heartbeat can be self heartbeat or cluster heartbeat.
        # We are reporting only heartbeat values, since reporting other metric value is not correct, since those
        # are older than default cache_time.
        #
        if "heartbeat" not in textkey:
            return None

        fsr_payload = cache[1]
        altered_fsr_payload = {}
        if textkey == TEXTKEY_SELF_HEARTBEAT_LAG:
            altered_fsr_payload['timestamp'] = fsr_payload['timestamp']
            return altered_fsr_payload
        #
        # Report cluster heartbeat delay by calculating delay with current time and last_known_timestamp
        #
        altered_fsr_payload[NODE_HEALTH_MISC_STAT] = {'heartbeat':[]}
        if 'heartbeat' in fsr_payload[NODE_HEALTH_MISC_STAT].keys():
            nodes = fsr_payload[NODE_HEALTH_MISC_STAT]['heartbeat']
            for node in nodes:
                node['delay'] = int(time.time()) - int(node['last_known_timestamp'])
            altered_fsr_payload[NODE_HEALTH_MISC_STAT]['heartbeat'] = nodes
        return altered_fsr_payload
