#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import requests, urllib3, os, argparse, json
from cshmac.requests import HmacAuth
import urllib.parse
import base64, sys
from urllib.parse import urlparse
import time
import threading
import fsr.commonlibs.utils

#
# Please note that this script calls API's using localhost (variable declared 'CH_HOST').
# The environment which is executing this script may NOT have 'no_proxy' set with 'localhost'.
# In the environment in which 'https_proxy' is set and there is NO 'no_proxy' set for 'localhost',
# below API calls in this script will start failing. Hence, regardless of whether executing
# environment have 'no_proxy' set or not, define it here explicitly.
#
os.environ['no_proxy'] = "127.0.0.1, localhost"
KEY_PATH = '/opt/cyops/configs/keys/'
APPLIANCE_PUBLIC_KEY = ''
APPLIANCE_PRIVATE_KEY = ''
DAS_PUBLIC_KEY = ''
DAS_PRIVATE_KEY = ''

silent = False

KEY_METHOD = 'method'
KEY_ENDPOINT = 'endpoint'
KEY_PAYLOAD = 'payload'
KEY_DAS = 'is_das'
KEY_MONITOR_SP = 'is_monitor_sp'

SP_INSTALL_SUCCEED = True

def read_files(key_path):
    global APPLIANCE_PUBLIC_KEY
    global APPLIANCE_PRIVATE_KEY
    with open(os.path.join(key_path, 'APPLIANCE_PRIVATE_KEY'), 'r') as private:
        APPLIANCE_PRIVATE_KEY = private.read().strip()
    with open(os.path.join(key_path, 'APPLIANCE_PUBLIC_KEY'), 'r') as public:
        APPLIANCE_PUBLIC_KEY = public.read().strip()

def get_appliance_keys():
    read_files(KEY_PATH)

def read_das_keys():
    global DAS_PUBLIC_KEY
    global DAS_PRIVATE_KEY
    with open('/opt/cyops-auth/dashmac/keys/dasprivate.key', 'r') as private:
        DAS_PRIVATE_KEY = private.read().strip()
    with open('/opt/cyops-auth/dashmac/keys/daspublic.key', 'r') as public:
        DAS_PUBLIC_KEY = public.read().strip()

def print_me(msg):
    global silent
    if silent:
        return
    print(msg, flush=True)

def print_response(response):
    try:
        # First see if can be printed in JSON
        print(json.dumps(response.json(), indent=4), flush=True)
    except Exception as e:
        # Object is not converted to JSON. Happens due to non-json response.
        try:
            # See if can be printed in string utf-8
            print(response.content.decode("utf-8"), flush=True)
        except Exception as e:
            # Print content in byte.
            print(response.content, flush=True)

def monitor_sp_thread(input_args):
    response = input_args['response']
    try:
        json_response = response.json()
    except Exception as e:
        sys.exit(1)
    input_endpoint = input_args[KEY_ENDPOINT]
    parsed_uri = urlparse(input_endpoint)
    input_host = '{uri.scheme}://{uri.netloc}'.format(uri=parsed_uri)
    import_job_id = json_response['importJob']['@id']
    progess_api_endpoint = "{0}{1}".format(input_host, import_job_id)
    input_args[KEY_ENDPOINT] = progess_api_endpoint
    input_args[KEY_METHOD] = "GET"
    input_args[KEY_PAYLOAD] = None
    # In seconds
    sleep_between_retry = 30
    total_failure = 0
    # Continuous failures for 30 * 6 secs i.e 3 mins
    max_failure = 6
    is_running = True
    global SP_INSTALL_SUCCEED
    while is_running:
        time.sleep(sleep_between_retry)
        is_request_succeed = make_request(input_args)
        response = input_args['response']
        try:
            json_response = response.json()
            if ('@type' in json_response.keys()) and (json_response['@type'] == 'SystemUpdate'):
                # Skip failure if SystemUpdate is in progress
                continue
        except Exception as e:
            print(str(e), flush=True)
        if (not is_request_succeed) or (json_response is None) or json_response['status'] == 'Error' or json_response['status'] == 'Draft':
            total_failure = total_failure + 1
            if total_failure == max_failure:
                # Progress API failed 4 times in gap of 30 secs between each failure
                print("Max failure count of {0} is reached".format(max_failure), flush=True)
                if json_response and ('errorMessage' in json_response.keys()):
                    print(json_response['errorMessage'], flush=True)
                if json_response and ('logMessages' in json_response.keys()):
                    print("Printing logMessages:", flush=True)
                SP_INSTALL_SUCCEED = False
                return
            continue
        # Progress API succeed, reset failure count
        total_failure = 0
        import_status = json_response['status']
        import_progess_percent = json_response['progressPercent']
        fsr.commonlibs.utils.FSR_SPINNER_PROGRESS = import_progess_percent
        if import_status.lower() == 'import complete':
            is_running = False

def monitor_sp(input_args):
    thread = threading.Thread(target=monitor_sp_thread, args=(input_args,))
    thread.start()
    spinner = fsr.commonlibs.utils.FSRSpinner("Installing the SOAR Framework Solution Pack", full_printing=True, monitor_progress=True)
    spinner.start_spinner()
    thread.join()
    if SP_INSTALL_SUCCEED:
        spinner.stop_spinner()
        return
    spinner.stop_spinner("FAILED")
    sys.exit(1)

def make_request(input_args):
    endpoint = input_args[KEY_ENDPOINT]
    method = input_args[KEY_METHOD]
    payload = input_args[KEY_PAYLOAD]
    is_monitor_sp = input_args[KEY_MONITOR_SP]
    auth = None
    global silent
    try:
        urllib3.disable_warnings()
        get_appliance_keys()
        if method.lower() == "delete":
            auth = HmacAuth(endpoint, method, APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, APPLIANCE_PUBLIC_KEY)
        else:
            auth = HmacAuth(endpoint, method, APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, json.dumps(payload))
        print_me("Calling api endpoint {0} with method {1}".format(endpoint, method))
        if method.lower() == "delete":
            response = requests.delete(endpoint, auth=auth, verify=False)
        else:
            response = requests.request(method, endpoint, auth=auth, json=payload, verify=False)
        if not is_monitor_sp:
            print_response(response)
        input_args['response'] = response
        if response.status_code == 200 or response.status_code == 201 or response.status_code == 204:
            return True
        elif response.status_code == 503 or response.status_code == 403 or response.status_code == 404 or response.status_code == 400:
            if is_monitor_sp and (not silent):
                print_response(response)
            return False
    except Exception as e:
        print_me(str(e))
        return False

def make_request_das(input_args):
    endpoint = input_args[KEY_ENDPOINT]
    method = input_args[KEY_METHOD]
    payload = input_args[KEY_PAYLOAD]
    try:
        urllib3.disable_warnings()
        read_das_keys()
        body_payload = json.dumps(payload) if payload else DAS_PUBLIC_KEY
        auth = HmacAuth(endpoint, method, DAS_PUBLIC_KEY, DAS_PRIVATE_KEY, body_payload)
        kwargs = {"auth": auth, "verify": False}
        if method.lower() == "delete":
            response = requests.delete(endpoint, **kwargs)
        else:
            if payload is not None:
                kwargs['json'] = payload
            response = requests.request(method, endpoint, **kwargs)
        print_response(response)
        input_args['response'] = response
        if response.status_code == 200 or response.status_code == 201 or response.status_code == 204:
            return True
        elif response.status_code == 503 or response.status_code == 403 or response.status_code == 404:
            return False
    except Exception as e:
        print_me(str(e))
        return False

def call_api(input_args):
    endpoint = input_args[KEY_ENDPOINT]
    method = input_args[KEY_METHOD]
    is_das = input_args[KEY_DAS]
    is_monitor_sp = input_args[KEY_MONITOR_SP]
    function_to_call = "make_request"
    if is_das:
        function_to_call = "make_request_das"
    if not globals()[function_to_call](input_args):
        print_me("Failed")
        sys.exit(1)
    if is_monitor_sp:
        monitor_sp(input_args)
    print_me("Successfully executed API.")

def decode_command_payload(payload):
    if not payload:
        return payload
    if type(payload) is dict:
        return payload
    try:
        return json.loads(payload)
    except Exception as e:
        pass
    try:
        return json.loads(base64.b64decode(payload))
    except Exception as e:
        pass
    print_me("Payload is neither a valid json nor a base64 encoded")
    sys.exit(1)

def define_parser(parser):
    parser.add_argument('--method',
                        help="REST API method. Default set to 'GET'",
                        dest=KEY_METHOD,
                        type=str,
                        required=False,
                        default='GET')

    parser.add_argument('--endpoint',
                        help='Full path of endpoint. Eg. https://localhost/api/query/alerts?$limit=30',
                        dest=KEY_ENDPOINT,
                        type=str,
                        required=True)

    parser.add_argument('--payload',
                        help='Payload in json format. Eg. {"key" : "value"} or base64 encoded',
                        dest=KEY_PAYLOAD,
                        type=str,
                        required=False,
                        default=None)

    parser.add_argument('--silent',
                        help="Do not print informative statements",
                        action='store_true',
                        dest='is_silent',
                        default=silent)

    parser.add_argument('--das',
                        help="Use DAS HMAC",
                        action='store_true',
                        dest=KEY_DAS,
                        default=False)

    parser.add_argument('--monitor-sp',
                        help=argparse.SUPPRESS,
                        action='store_true',
                        dest=KEY_MONITOR_SP,
                        default=False)
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    define_parser(parser)
    args = parser.parse_args()
    silent = args.is_silent
    payload = decode_command_payload(args.payload)
    input_args = vars(args)
    input_args[KEY_PAYLOAD] = payload
    call_api(input_args)
    sys.exit(0)
