#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import os, sys
if os.getuid() != 0:
    print("This script must be run as 'root'")
    sys.exit(1)
import argparse, json, subprocess, re, shutil
sys.path.append("/opt/cyops/scripts/")
from csadm.logs.logger_defs import get_logger
MSG_EXECUTE_CMD = "Executing command [{0}]"
MSG_CMD_FAIL = "Failed to execute cmd [{0}] with return code as [{1}], output as [{2}] and error as [{3}]"
MSG_DRY_RUN = "Dry Run:{0}"
ENV_INTEGRATIONS = 'integrations'
VALID_ENVS = [ENV_INTEGRATIONS]
KEY_PIP = 'pip'
KEY_SITE_PACKAGES = 'site_packages'
KEY_REINSTALL_USER = 'reinstall_user'
ENV_DETAILS = {
    ENV_INTEGRATIONS: {
        KEY_PIP: '/opt/cyops-integrations/.env/bin/pip',
        KEY_SITE_PACKAGES: [
            '/opt/cyops/configs/integrations/packages/lib/python3.9/site-packages',
            '/opt/cyops-integrations/.env/lib/python3.9/site-packages'
        ],
        KEY_REINSTALL_USER: 'fsr-integrations'
    }
}
LOG_FILE = "/var/log/cyops/pip-cleanup.log"
logger = get_logger(__name__, LOG_FILE)

dry_run = False

def reinstall_packages(input_env_details, reinstall_list):
    if len(reinstall_list) == 0:
        return
    packages = ' '.join(reinstall_list)
    msg = "Reinstalling {0}".format(packages)
    if dry_run:
        msg = MSG_DRY_RUN.format(msg)
    logger.info(msg)
    if dry_run:
        return
    reinstall_cmd = "sudo -u {0} {1} install --force-reinstall --no-deps {2} 2>&1 | tee --append {3}".format(input_env_details[KEY_REINSTALL_USER],
                                                                 input_env_details[KEY_PIP],
                                                                 packages,
                                                                 LOG_FILE)
    os.system(reinstall_cmd)

def byte2human(no_of_bytes):
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if no_of_bytes >= prefix[s]:
            value = float(no_of_bytes) / prefix[s]
            return '%.1f%s' % (value, s)
    return "%B" % no_of_bytes

def execute_cmd(cmd, ignore_failure=False):
    pipes = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    out, err = pipes.communicate()
    ret = pipes.returncode
    out = out.strip().decode("utf-8")
    err = err.strip().decode("utf-8")
    if (ret != 0) and (not ignore_failure):
        logger.error(MSG_CMD_FAIL.format(cmd, ret, out, err))
        sys.exit(1)
    return (ret, out, err)

def get_installed_packages(pip_executable):
    format_cmd = "{0} list --format json"
    cmd = format_cmd.format(pip_executable)
    ret, out, err = execute_cmd(cmd)
    return json.loads(out.strip())

def get_older_packages(input_env_details, package_name, package_latest_version):
    list_site_packages = input_env_details[KEY_SITE_PACKAGES]
    package_name = package_name.replace("-", "_")
    files = []
    package_short_name = package_name + "-" + package_latest_version
    package_full_name = package_short_name + ".dist-info"
    for site_packages in list_site_packages:
        tmp_files = [f for f in os.listdir(site_packages) if re.match(re.escape(package_name) + r'-' + '.+dist-info', f)]
        if tmp_files:
            files.extend(tmp_files)
        tmp_files_namespace_package = [f for f in os.listdir(site_packages) if re.match(re.escape(package_name) + r'-' + '.+' + re.escape('.pth'), f)]
        for tmp_file_namespace_package in tmp_files_namespace_package:
            if not tmp_file_namespace_package.startswith(package_short_name+ "-"):
                files.append(tmp_file_namespace_package)
    #
    # Remove all matching entries for latest_package. We need to remove all matching entries, since
    # in case where multiple site-packages script cleaning up, and same version of package is available
    # at two places.
    #
    files = list(filter(lambda a: a != package_full_name, files))
    if files:
        msg = "Found older files {0} for package {1}".format(files, package_short_name)
        if dry_run:
            msg = MSG_DRY_RUN.format(msg)
        logger.info(msg)
    return files

def get_directory_size_in_bytes(path):
    return int(subprocess.check_output(['du','-sb', path]).split()[0].decode('utf-8'))

def process_env(input_env_details):
    installed_packages = get_installed_packages(input_env_details[KEY_PIP])
    list_site_packages = input_env_details[KEY_SITE_PACKAGES]
    cleanup_size_bytes = 0
    reinstall_list = []
    for installed_package in installed_packages:
        package_name = installed_package['name']
        package_latest_version = installed_package['version']
        older_packages = get_older_packages(input_env_details, package_name, package_latest_version)
        if older_packages:
            for older_package in older_packages:
                if '.dist-info' in older_package:
                    reinstall_list.append(package_name + "==" + package_latest_version)
                    break
            for old_package_name in older_packages:
                for site_packages in list_site_packages:
                    old_package_full_path= site_packages + "/" + old_package_name
                    if os.path.isdir(old_package_full_path) or os.path.isfile(old_package_full_path):
                        cleanup_size_bytes = cleanup_size_bytes + get_directory_size_in_bytes(old_package_full_path)
                        msg = "Deleting {0}".format(old_package_full_path)
                        if dry_run:
                            msg = MSG_DRY_RUN.format(msg)
                        logger.info(msg)
                        if not dry_run:
                            if os.path.isdir(old_package_full_path):
                                shutil.rmtree(old_package_full_path)
                            else:
                                os.remove(old_package_full_path)

    if cleanup_size_bytes != 0:
        msg = "Total cleaned up space {0}".format(byte2human(cleanup_size_bytes))
        if dry_run:
            msg = MSG_DRY_RUN.format(msg)
        logger.info(msg)
        reinstall_packages(input_env_details, reinstall_list)

def define_parser(parser):
    parser.add_argument('--env',
                        dest='input_env',
                        required=True,
                        choices=VALID_ENVS,
                        help='PIP environment')

    parser.add_argument('--dry-run',
                        dest='dry_run',
                        action='store_true',
                        default=False,
                        help='Dry run')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    define_parser(parser)
    args = parser.parse_args()
    dry_run = args.dry_run
    process_env(ENV_DETAILS[args.input_env])
    sys.exit(0)