#!/opt/cyops-auth/.env/bin/python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import sys, hmac, os, time, json
import requests, urllib3
urllib3.disable_warnings()
from fsr.portal_client import PortalClient
sys.path.append("/opt/cyops-auth")
from utilities.ha.stat_utils import collect_stats
from utilities.ha.consts import NODE_HEALTH_CPU, NODE_HEALTH_DISK, NODE_HEALTH_MEMORY
sys.path.append("/opt/cyops/scripts/")
from csadm.logs.logger_defs import get_logger
from csadm.utils import execute_cmd, read_json

#####
# define message constants/global variables after this
####

FSR_FCLOUD_HEARTBEAT_LOG_FILE = "/var/log/cyops/fcloud/heartbeat.log"

dir_path = os.path.dirname(FSR_FCLOUD_HEARTBEAT_LOG_FILE)
if not os.path.exists(dir_path):
    os.makedirs(dir_path)
logger = get_logger(__name__, FSR_FCLOUD_HEARTBEAT_LOG_FILE)

pc = PortalClient(logger)

DEVICE_DATA_DISK = "/dev/mapper/vgdata-lvdata"
KEY_VM_INSTANCE_ID = "uuid"
KEY_META_IN_META_DATA = "meta"
KEY_SECRET_KEY = "SECRET_KEY"
KEY_API_SERVER_HOST = "API_SERVER_HOST"
KEY_API_SERVER_PORT = "API_SERVER_PORT"
KEY_API_OUTPUT_ACTIONS = "actions"
KEY_API_OUTPUT_CODE = "code"
KEY_API_OUTPUT_CODE_DESC = "code_desc"
KEY_API_ACTION_TYPE = "type"
KEY_API_ACTION_DATA = "data"
KEY_API_ACTION_DATA_LICENSE_TEXT = "license_file"
API_ACTION_TYPE_NEW_LICENSE = "new_license"
FORTICLOUD_API_METHOD_HEARTBEAT = "heartBeat"
FORTICLOUD_FSR_PLATFORM_NAME = "FSRVMC"
FILE_FSR_VERSION = "/opt/cyops-ui/cyops_version.json"

#hs means heartbeat script
message_const_hs_1 = "{0} request for FortiCloud API method [{1}] returned 'code' in output json as [{2}]. \
                      Code description 'code_desc' received in output json is [{3}]."
message_const_hs_2 = "{0} on API URL [{1}] is failed with status code [{2}]"
message_const_hs_3 = "Calling FortiCloud [{0}] method"
message_const_hs_4 = "Unable to read file [{0}]"
#####
# define message constants/global variables before this
####

def convert_to_kb(input_size):

    to_kb = {'K' : 1,
             'KB': 1,
             'M' : 1024,
             'MB': 1024,
             'G' : 1024 * 1024,
             'GB': 1024 * 1024,
             'T' : 1024 * 1024 * 1024,
             'TB': 1024 * 1024 * 1024}

    size_format = input_size[-1]
    return float(input_size[:-1]) * to_kb[size_format.upper()]

def get_fsr_health_details():
    metrics = [
        NODE_HEALTH_CPU,
        NODE_HEALTH_DISK,
        NODE_HEALTH_MEMORY
    ]
    payload_fsr_health = collect_stats(metrics)
    if not payload_fsr_health:
        logger.error("Failed to collect the stats")
        sys.exit(1)
    return payload_fsr_health

def get_current_version():
    dict_fsr_version = read_json(FILE_FSR_VERSION)
    if dict_fsr_version is None:
        logger.error(message_const_hs_4.format(FILE_FSR_VERSION))
        sys.exit(1)

    #current_vesion will be like 6.4.4-33
    current_version = dict_fsr_version['version']
    current_release = current_version.split("-")[0]
    current_build_number = current_version.split("-")[1]
    #TODO findout the release date in fortmat DDMMYY for eg. 061020 ( 6th oct 2020)
    # Actually we need to send string like v6.4.4-build33 061020 (GA).
    # But since we are not able to find the release date, for now, we will send only v6.4.4-build33
    final_string_for_current_version = "v" + str(current_release) + "-build" + str(current_build_number)
    return final_string_for_current_version

def get_payload_for_heartbeat(payload_fsr_health):
    load_avg_mins_5 = payload_fsr_health[NODE_HEALTH_CPU]['load_avg_minute_5']
    cpu_usage_percent = payload_fsr_health[NODE_HEALTH_CPU]['percent']
    # The FSR cloud portal needs the cpu usage percent between 0-1.
    cpu_usage_percent = cpu_usage_percent / 100
    lvm_lv_for_disk = DEVICE_DATA_DISK
    disk_details = payload_fsr_health[NODE_HEALTH_DISK]
    disk_total = None
    disk_used = None
    for mount_point_details in disk_details:
        if mount_point_details['device'] == lvm_lv_for_disk:
            disk_total = mount_point_details['total']
            disk_used = mount_point_details['used']
            break 
    memory_details = payload_fsr_health[NODE_HEALTH_MEMORY]['ram']
    memory_total = memory_details['total']
    memory_used = memory_details['used']
 
    ts = int(time.time())

    meta_data = pc.get_metadata()
    vm_inst_id = meta_data[KEY_VM_INSTANCE_ID]
    payload = {
        "id": 12345,
        "jsonrpc": "2.0",
        "method": FORTICLOUD_API_METHOD_HEARTBEAT,
        "params": {
            "cpuloadavg": float(load_avg_mins_5),
            "cpuusage": float(cpu_usage_percent),
            "disktotal": int(convert_to_kb(disk_total)),
            "diskusage": int(convert_to_kb(disk_used)),
            "instid": vm_inst_id,
            "licensed": True,
            "memtotal": int(convert_to_kb(memory_total)),
            "memusage": int(convert_to_kb(memory_used)),
            "platform": FORTICLOUD_FSR_PLATFORM_NAME,
            "sig": pc.get_signature(ts),
            "ts": ts,
            "version": get_current_version()
        }
    } 
    return  payload

# Below timeout is in secs
def make_portal_api_call(endpoint, method, headers=None, payload=None, timeout=30):
    meta_data = pc.get_metadata()
    api_url = "https://" + meta_data[KEY_META_IN_META_DATA][KEY_API_SERVER_HOST] + ":" + str(meta_data[KEY_META_IN_META_DATA][KEY_API_SERVER_PORT]) + endpoint
    response = None
    try:
        response = requests.request(method,
                                    api_url,
                                    json=payload,
                                    verify=False,
                                    timeout=timeout,
                                    headers=headers)
    except Exception as e:
        logger.error(str(e))

    if (response is not None) and response.ok:
        return response.text
    logger.error(message_const_hs_2.format(method,
                                           api_url,
                                           str(response.status_code)))
    return None

def process_api_output_actions(api_output):
    if KEY_API_OUTPUT_ACTIONS not in api_output.keys():
        return
    for action in api_output[KEY_API_OUTPUT_ACTIONS]:
        if action[KEY_API_ACTION_TYPE] == API_ACTION_TYPE_NEW_LICENSE:
            # install new license
            pc.install_license(action[KEY_API_ACTION_DATA][KEY_API_ACTION_DATA_LICENSE_TEXT])
            pass

def start():
    payload_fsr_health = get_fsr_health_details()
    payload_heartbeat = get_payload_for_heartbeat(payload_fsr_health)
    headers = {
        'Content-Type': 'application/json'
    }
    logger.info(message_const_hs_3.format(FORTICLOUD_API_METHOD_HEARTBEAT))
    api_output = make_portal_api_call(endpoint="/jsonrpc",
                                      method="POST",
                                      headers=headers,
                                      payload=payload_heartbeat)
    if api_output is None:
        sys.exit(1)
    api_output = json.loads(api_output)
    api_output_code = api_output[KEY_API_OUTPUT_CODE]
    if api_output_code != 0:
        logger.error(message_const_hs_1.format("POST",
                                               FORTICLOUD_API_METHOD_HEARTBEAT,
                                               str(api_output_code),
                                               api_output[KEY_API_OUTPUT_CODE_DESC]))
        sys.exit(1)
    process_api_output_actions(api_output)

if __name__ == "__main__":
    start()

sys.exit(0)