#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import argparse
import subprocess
import sys
import base64
import json
import os
import requests
from pathlib import Path
import shutil
import glob

from cshmac.requests import HmacAuth

from fsr.portal_client import PortalClient
from fsr.commonlibs.utils import audit_system, GenericConfigReader
from fsr_utilities import decrypt

sys.path.append("/opt/cyops/scripts/")
from csadm.logs.logger_defs import get_logger

FSR_FCLOUD_INIT_CONFIG_LOG_FILE = "/var/log/cyops/fcloud/init-config.log"
INIT_CONFIG_START_FILE = "/tmp/.forticloud-init-config-start"
FLAGS_DIR = "/opt/cyops/configs/.flags"
FILE_FSR_INIT_CONFIG_COMPLETED = FLAGS_DIR + "/.forticloud-init-config-completed"
FILE_FSR_INIT_CONFIG_RESTRICT_RERUN = FLAGS_DIR + "/.restrict-rerun-for-forticloud-init-config"
# converted to prod flag file is used in sub-cmd csadm db --restore.
FILE_FSR_CONVERTED_TO_PROD = FLAGS_DIR + "/.forticloud-vm-converted-to-prod"
# Same directory is getting used in csadm system data-reset CLI. Ensure you change there is changes here.
DIR_DATA_RESET = "/var/lib/pgsql/.data-reset"

with open('/opt/cyops/configs/keys/APPLIANCE_PUBLIC_KEY', 'r') as private_key_file:
    APPLIANCE_PUBLIC_KEY = private_key_file.read().strip()
with open('/opt/cyops/configs/keys/APPLIANCE_PRIVATE_KEY', 'r') as public_key_file:
    APPLIANCE_PRIVATE_KEY = public_key_file.read().strip()

dir_path = os.path.dirname(FSR_FCLOUD_INIT_CONFIG_LOG_FILE)
if not os.path.exists(dir_path):
    os.makedirs(dir_path)
logger = get_logger(__name__, FSR_FCLOUD_INIT_CONFIG_LOG_FILE)

portal_client = PortalClient(logger)

######
# add helper functions below this
######

def execute_cmd(cmd, exit_on_failure=True, env=None):
    if env:
        os.environ.update(env)
    pipes = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    std_out, std_err = pipes.communicate()
    if (pipes.returncode != 0) and exit_on_failure:
        # Exit if the command fails
        logger.error("Failed to execute command [{0}] with error {1}".format(cmd, std_err.strip()))
        sys.exit(1)
    ret = pipes.returncode
    std_out = std_out.strip().decode("utf-8")
    std_err = std_err.strip().decode("utf-8")
    return ret, std_out, std_err

def _get_sni():
    account_id = portal_client.get_account_id()
    portal_url = portal_client.get_portal_url()
    sni = str(account_id) + "." + portal_url.replace("https://", "")
    return sni

def _get_host():
    portal_url = portal_client.get_portal_url()
    return portal_url.replace("https://", "")

def _make_rest_call(url, method, body):
    auth = HmacAuth(url, method, APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, json.dumps(body))
    response_obj = requests.request(method, url, auth=auth, json=body, verify=False)
    return response_obj

def _clear_super_admin_user():
    url = "https://localhost/api/auth/config"
    body = {"option": "super_admin", "value":""}
    response_obj = _make_rest_call(url, 'PUT', body)
    if response_obj.status_code != 200:
        logger.error("Failed to clear super admin user")
        return False, response_obj.text
    return True, ""

def start_initialization():
    Path(INIT_CONFIG_START_FILE).touch()

def end_initialization():
    Path(FILE_FSR_INIT_CONFIG_COMPLETED).touch()

def cache_clear_ch():
    logger.info("Clearing API cache")
    cmd = "sudo -u nginx php /opt/cyops-api/bin/console cache:clear --no-interaction"
    ret, out, err = execute_cmd(cmd, False)
    if ret != 0:
        logger.error("For command [{0}], got ret [{1}], out [{2}] and err [{3}]".format(cmd, ret, out, err))

######
# add helper functions above this
######

def enable_mq_mtls(input_args):
    # We are skipping the MQ service restart, as after this step, we run the change_hostname steps which
    # restart all services.
    #
    cmd = "csadm mq mtls enable --skip-service-restart"
    ret, out, err = execute_cmd(cmd, False)
    if ret != 0:
        logger.error("For cmd [" + cmd + "] got out as [" + out + "] and err as [" + err + "]")
        return False, err
    return True, ""

def add_mq_loopback_users(input_args):
    mq_conf = '/opt/cyops-rabbitmq/configs/ssl/rabbitmq-embedded-sme.conf'
    users_to_add_in_loop_back = ['fsr-cluster', 'admin']
    if not os.path.isfile(mq_conf):
        logger.error("File [" + mq_conf + "] does not exist.")
        return True, ""
    try:
        # Check if user already added in loopback
        fh = open(mq_conf, "r")
        file_contents = fh.read()
        not_added_users = []
        for user in users_to_add_in_loop_back:
            loopback_str = "loopback_users." + user + "=true"
            if loopback_str in file_contents:
                logger.info("MQ user [" + user + "] is already added in loopback")
            else:
                not_added_users.append(user)
        fh.close()

        if len(not_added_users) == 0:
            return True, ""

        fh = open(mq_conf, "a")
        fh.write("\n")
        for user in not_added_users:
            loopback_str = "loopback_users." + user + "=true"
            logger.info("Adding MQ user [" + user + "] in loopback")
            fh.write(loopback_str + "\n")
        fh.close()
        # We should restart the MQ after updating the conf file, but since, after this step,
        # we run the hostname change step, which restarts all services.
    except Exception as err:
        logger.error(str(err))
    return True, ""

def download_and_install_license(input_args):
    logger.info("Downloading and installing license")
    response = portal_client.download_license()
    if response and response['status_code'] == 200:
        license = response['output']['license_file']
        logger.info("License downloaded successfully")
        return True, portal_client.install_license(license)
    else:
        logger.error("license download failed with status code "+str(response['status_code'])+". Error occurred is: "+response['error'])
    return False, ""

def change_csadmin_ssh_default_password(input_args):
    logger.debug("Resetting csadmin user password")
    CSADMIN_ACCOUNT_AGING_INFO_CMD = "chage -l csadmin"
    CSADMIN_PASSWORD_EXPIRE_CMD = "passwd --expire csadmin"
    account_id = portal_client.get_account_id()
    CSADMIN_ACCOUNT_CHANGE_PWD_CMD = "echo -e \""+str(account_id)+"\\n"+str(account_id)+"\""+" | passwd csadmin"
    ret, output, error = execute_cmd(CSADMIN_ACCOUNT_AGING_INFO_CMD)
    output_lines = output.split("\n")
    for item in output_lines:
        if 'Password expires' in item:
            password_expires = item.split(':')[1].strip()
            if password_expires != 'never':
                execute_cmd(CSADMIN_ACCOUNT_CHANGE_PWD_CMD)
                logger.info("csadmin user password reset successful")
                logger.info("Marking password expired for csadmin")
                execute_cmd(CSADMIN_PASSWORD_EXPIRE_CMD)
                logger.info("Marked password expired for csadmin")
            else:
                logger.info("Password reset not required")
            break
    return True, ""

def update_sni_and_cert_in_router(input_args):
    logger.info("Updating SNI and certificate for router")
    router_url = "https://localhost/api/3/routers/52c5cee8-5c28-4ed2-a886-ec8bf4dc5993"

    sni = _get_sni()
    host = _get_host()
    with open('/opt/cyops/configs/rabbitmq/ssl/cyopsca/cacert.pem') as cacert_file:
        cert = cacert_file.read()
    logger.info("Updating SNI to "+sni)
    method = 'PUT'
    body = {
        "address": host,
        "sni": sni,
        "certificate": cert
    }
    auth =  HmacAuth(router_url, method, APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, json.dumps(body))
    response_obj = requests.put(router_url, auth=auth, json=body, verify=False)
    if response_obj.status_code != 200:
        logger.error("Failed to update SNI and certificate for router")
        logger.error(response_obj.text)
    logger.info("Updated SNI and certificate for router")
    cmd = "csadm secure-message-exchange update-exchange-event-listener-certs"
    logger.info("Updating the event exchange certificates for Embedded SME")
    ret, out, err = execute_cmd(cmd, False)
    if ret != 0:
        logger.error("Failed to update the event exchange certificates for Embedded SME")
    return True, ""

def trigger_heartbeat_and_heartbeat_cron_job(input_args):
    logger.info("Triggering heartbeat")
    execute_cmd("/opt/cyops/scripts/cloud/openstack/heartbeat.py", exit_on_failure=False)
    logger.info("Triggered heartbeat")

    logger.info("Creating heatbeat cron job")
    HEARTBEAT_CRON_FILE="/etc/cron.d/send_heartbeat"
    try:
        if os.path.isfile(HEARTBEAT_CRON_FILE):
            logger.info("Hearbeat cron already configured")
            return True, ""
        else:
            with open(HEARTBEAT_CRON_FILE, 'w') as hb_cron_file:
                hb_cron_file.write("*/5 * * * * root /opt/cyops/scripts/cloud/openstack/heartbeat.py\n")
            logger.info("Created heatbeat cron job")
            return True, ""
    except Exception as error:
        return  False, ""

def delete_monitor_init_config_cron_job(input_args):
    logger.info("Deleting monitor_init_config cron job")
    MONITOR_INIT_CONFIG_CRON_FILE = "/etc/cron.d/monitor-forticloud-init-config"
    if os.path.isfile(MONITOR_INIT_CONFIG_CRON_FILE):
        os.remove(MONITOR_INIT_CONFIG_CRON_FILE)
        logger.info("Deleted monitor_init_config cron job")
    else:
        logger.info("monitor_init_config cron job already deleted")
    return True, ""

def change_csadmin_gui_default_password_and_mark_inactive(input_args):
    pg_config = GenericConfigReader().reader_yaml(["/opt/cyops/configs/database/db_config.yml"]).get("postgres")
    pg_password = decrypt(pg_config.get('pg_password'))
    CSADMIN_DAS_UUID_SQL = "psql -U cyberpgsql -d das -qAt -c \"SELECT uuid FROM users WHERE login_id='csadmin'\""
    CSADMIN_CH_UUID_SQL = "psql -U cyberpgsql -d venom -qAt -c \"SELECT uuid FROM actors WHERE firstname='CS' and lastname='Admin'\""
    ret_code, das_uuid, error = execute_cmd(CSADMIN_DAS_UUID_SQL, env={'PGPASSWORD': pg_password})
    if ret_code != 0:
        return False, error
    ret, ch_uuid, error = execute_cmd(CSADMIN_CH_UUID_SQL, env={'PGPASSWORD': pg_password})
    if ret_code != 0:
        return False, error

    # csadmin is kept as super admin by default. Being super admin, it can not be marked inactive. We need to clear
    # super admin first before marking it inactive"
    status, message = _clear_super_admin_user()
    if not status:
        return status, message

    url = "https://localhost/api/3/resetpassword"
    account_id = portal_client.get_account_id()
    body = {
          "currentUUID": ch_uuid,
          "loginId": "csadmin",
          "password": "Fsr@"+str(account_id),
          "sendEmail": False
        }
    response_obj = _make_rest_call(url, 'POST', body)
    if response_obj.status_code != 200:
        logger.error("Failed to update csadmin GUI password")
        return False, response_obj.text
    logger.info("GUI password for csadmin user changed")
    body = {
          "update": {
            "uuid": das_uuid,
            "status": 2
          }
        }
    url = 'https://localhost/api/auth/users'
    response_obj = _make_rest_call(url, 'PUT', body)
    if response_obj.status_code != 200:
        logger.error("Failed to mark csadmin as inactive")
        return False, response_obj.text
    body = {
        "csActive": False
    }
    url = "https://localhost/api/3/people/3451141c-bac6-467c-8d72-85e0fab569ce"
    response_obj = _make_rest_call(url, 'PUT', body)
    if response_obj.status_code != 200:
        logger.error("Failed to mark csadmin as inactive")
        return False, response_obj.text
    logger.info("csadmin user marked inactive")
    return True, ""

def restrict_rerun(input_args):
    Path(FILE_FSR_INIT_CONFIG_RESTRICT_RERUN).touch()
    return True, ""

def change_hostname(input_args):
    sni = _get_sni()
    logger.info("Changing hostname to " + sni)
    HOSTNAME_CMD="hostname"
    ret, output, error = execute_cmd(HOSTNAME_CMD)
    if output.strip() == sni:
        logger.info("Hostname already changed")
        return True, ""
    CHANGE_HOSTNAME_CMD = "csadm hostname --set " + sni
    execute_cmd(CHANGE_HOSTNAME_CMD)
    logger.info("hostname changed successfully")
    return True, ""

def regenerate_certs_for_mq(input_args):
    logger.info("Regenerating mq certificate")
    MQ_GENERATE_CERT_CMD = "csadm mq certs generate"
    execute_cmd(MQ_GENERATE_CERT_CMD)
    logger.info("mq certificate regenerated successfully")
    return True, ""

def restrict_ws_connection(input_args):
    ws_conf_file_path = '/etc/cyops/config.yml'
    logger.info("Restricting allowed hosts for websocket")
    with open(ws_conf_file_path, 'r') as ws_conf_file:
        lines = ws_conf_file.readlines()
        for line in lines:
            if line and line.strip().startswith('allowed_host'):
                # allowed_hosts setting is already configured.
                return True, ""
    sni = _get_sni()
    allowed_hosts_setting = '\nallowed_host: https://' + sni
    with open(ws_conf_file_path, 'a') as ws_conf_file:
        ws_conf_file.write(allowed_hosts_setting)
        # Restart tomcat service to make the changes effective
        execute_cmd('systemctl restart cyops-tomcat')
    return True, ""



def refresh_device_uuid(input_args):
    cmd = "csadm license --refresh-device-uuid"
    ret, out, err = execute_cmd(cmd, False)
    if ret != 0:
        logger.error("For cmd [" + cmd + "] got out as [" + out + "] and err as [" + err + "]")
        return False, err
    if out and "Device UUID refreshed successfully" in out:
        logger.info("Restarting services to reflect Device UUID change.")
        cmd = "csadm services --restart"
        execute_cmd(cmd, False)
        cache_clear_ch()
    return True, ""

def backup_data_for_data_reset(input_args):
    _is_trial = portal_client.is_trial()
    if _is_trial is None:
        err_msg = "Unable to determine if VM is trial"
        logger.error(err_msg)
        return False, err_msg
    if not _is_trial:
        logger.info("VM is running with the production settings, skipping the data backup for data reset operation")
        return True, ""
    if os.path.isdir(DIR_DATA_RESET):
        backup_files = glob.glob(DIR_DATA_RESET + '/DR_BACKUP_*.tgz')
        if len(backup_files) > 0:
            logger.info("Backup file already exist for data reset operation at [" + DIR_DATA_RESET + "]")
            return True, ""
        shutil.rmtree(DIR_DATA_RESET)
    os.mkdir(DIR_DATA_RESET, mode=0o755)
    # We are skipping the site-packages in the backup, since it consume more space.
    # site-packages can be skipped since we are data reset is done on same instance.
    #
    logger.info("Taking backup for data reset operation at [" + DIR_DATA_RESET + "]")
    cmd = "csadm db --backup " + DIR_DATA_RESET + " --skip-site-packages"
    ret, out, err = execute_cmd(cmd, False)
    if ret != 0:
        logger.error("For cmd [" + cmd + "] got out as [" + out + "] and err as [" + err + "]")
        return False, err
    return True, ""

def convert_to_prod(input_args):
    if os.path.isfile(FILE_FSR_CONVERTED_TO_PROD):
        logger.info("VM is already converted to prod config, nothing to do")
        return True, ""
    _is_trial = portal_client.is_trial()
    if _is_trial is None:
        err_msg = "Unable to determine if VM is trial"
        logger.error(err_msg)
        return False, err_msg
    if _is_trial:
        logger.info("VM is running in the trial mode, nothing to do")
        return True, ""
    cmd = "csadm system config --mode prod --force"
    skip_service_restart = False
    if 'skip_service_restart' in input_args.keys():
        skip_service_restart = input_args['skip_service_restart']
    if skip_service_restart:
        cmd = cmd + " --skip-service-restart"
    logger.info("Running cmd [" + cmd + "]")
    ret, out, err = execute_cmd(cmd, False)
    logger.error("For cmd [" + cmd + "] got out as [" + out + "] and err as [" + err + "]")
    Path(FILE_FSR_CONVERTED_TO_PROD).touch()
    message = {
        "operation": "Update",
        "title": "FortiSOAR Cloud VM converted from Trial to Prod"
    }
    audit_system(message, logger)
    return True, ""

def refresh_fcloud_metadata(input_args):
    portal_client.refresh_metadata()
    return True, ""

def handle_subsequent_boot():
    steps = {
        "refresh_fcloud_metadata": {},
        "refresh_device_uuid": {},
        "trigger_heartbeat_and_heartbeat_cron_job": {},
        #'convert_to_prod': {"skip_service_restart": False},
        "delete_monitor_init_config_cron_job": {},
        "restrict_rerun": {}
    }
    for step in steps.keys():
        # Keep retrying step until successful
        while True:
            logger.info("Executing subsequent boot step: " + step)
            status, err_msg = globals()[step](steps[step])
            if status:
                break
            else:
                logger.error("Subsequent boot step " + step + " failed with error " + err_msg)
                logger.error("Retrying")
    logger.info("Completed Subsequent boot steps")

def init_config():
    start_initialization()
    sni = _get_sni()
    logger.info("SNI for the machine is: " + sni)
    init_steps = {
        'change_csadmin_ssh_default_password': {},
        'add_mq_loopback_users': {},
        'enable_mq_mtls': {},
        'download_and_install_license': {},
        'change_csadmin_gui_default_password_and_mark_inactive': {},
        #'convert_to_prod': {'skip_service_restart': True},
        'change_hostname': {},
        'regenerate_certs_for_mq': {},
        'update_sni_and_cert_in_router': {},
        'restrict_ws_connection': {},
        'trigger_heartbeat_and_heartbeat_cron_job': {},
        #'backup_data_for_data_reset': {},
        'delete_monitor_init_config_cron_job':{},
        'restrict_rerun':{}
    }

    for step in init_steps.keys():
        # Keep retrying step until successful
        while True:
            logger.info("Executing step: " + step)
            status, message = globals()[step](init_steps[step])
            if status:
                break
            else:
                logger.error("step " + step + " failed with error " + message)
                logger.error("Retrying")
    end_initialization()
    sys.exit(0)

def define_parser(parser):
    parser.add_argument("--subsequent-boot",
                        dest="is_subsequent_boot",
                        default=False,
                        action="store_true",
                        help="Execute Subsequent boot steps")

if __name__ == "__main__":
    if os.path.isfile(FILE_FSR_INIT_CONFIG_RESTRICT_RERUN):
        logger.error("File [" + FILE_FSR_INIT_CONFIG_RESTRICT_RERUN + "] exists. Exiting with 1.")
        sys.exit(1)
    parser = argparse.ArgumentParser()
    define_parser(parser)
    args = parser.parse_args()
    if args.is_subsequent_boot:
        handle_subsequent_boot()
        sys.exit(0)
    init_config()
