#!/bin/bash
# Copyright start
# Copyright (C) 2008 - 2024 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end

add_date_to_log(){
    echo "" >> $f_log
    date "+%d-%b-%Y %H:%M:%S" >> $f_log
}

my_system(){
    cmd="$1"
    echo "$cmd" >> $f_log
    $cmd >> $f_log 2>&1
    if [ $? -ne 0 ]; then
        echo "Error: Failed to run the command: $cmd" | tee -a $f_log
        my_exit 1
    fi
}

my_exit(){
    echo "---------------------------------------------------------------------------------------------------------" >> $f_log
    exit $1
}

print_summary(){
    s_total_size_logical_volume_after=$(df -kh | sed "s/--/-/g" | grep "$s_volume_group-$s_logical_volume" | awk '{print $2}')
    s_free_disk_space_logical_volume_after=$(df -kh | sed "s/--/-/g" | grep "$s_volume_group-$s_logical_volume" | awk '{print $4}')

    echo "Before the expansion,"
    echo "    1. Total size of the logical volume \"$s_logical_volume\" - $s_total_size_logical_volume_before"
    echo "    2. Free disk space in the logical volume \"$s_logical_volume\" - $s_free_disk_space_logical_volume_before"
    echo "After the expansion,"
    echo "    1. Total size of the logical volume \"$s_logical_volume\" - $s_total_size_logical_volume_after"
    echo "    2. Free disk space in the logical volume \"$s_logical_volume\" - $s_free_disk_space_logical_volume_after"
    echo "Information: In case total size of the logical volume after the expansion"
    echo "             is more than expected, it is due to csadm command fully occupies"
    echo "             the volume group for simplicity of the disk calculation"
}

check_lvm_mounted(){
    if [ -z "$s_filesystem_logical_volume" ]; then
        echo "Error : Unable to expand '$s_logical_volume'. Please verify if '$s_logical_volume' is currently mounted." | tee -a $f_log
        echo "        Exiting." | tee -a $f_log
        my_exit 1
    fi
}

check_new_disk_empty(){

    local l_disk_block_types=$(lsblk --bytes --noheadings /dev/$s_disk | awk '{print $6}')
    echo "$l_disk_block_types" | grep "lvm" > /dev/null
    local flag_lvm_present=$?
    echo "$l_disk_block_types" | grep "part" > /dev/null
    local flag_partition_present=$?
    if [ $flag_lvm_present -eq 0 -a $flag_partition_present -eq 1 ]; then
        echo "Error: The disk '$s_disk' has a volume group mounted on it without any partition being created" | tee -a $f_log
        echo "       This subcommand does not support the automated handling of volume groups created on raw disk blocks due to the complications involved." | tee -a $f_log
        echo "       Either provide a new disk and re-run this command or or use the disk commands provided by the Operating System to handle the requirement." | tee -a $f_log
        echo "Exiting now." | tee -a $f_log
        my_exit 1
    fi

}

validate_summary(){
    echo "Validating the changes to be done by 'csadm system disk expand-lv' command:"
    local s_expansion_type=""
    local flag_expansion_size_not_available="false"
    if [ $flag_use_disk = "true" ];then
        # Is it new disk (by searching for part in lsblk: lsblk | grep $s_disk | grep part)
        lsblk | grep $s_disk | grep -w --silent part
        if [ $? -eq 0 ];then
        # This is the existing disk
            s_expansion_type="existing-disk"
        else
            check_new_disk_empty
            s_expansion_type="new-disk"
        fi
    else
        s_expansion_type="vg"
    fi

    if [ "$s_expansion_type" = "vg" ];then
        s_expansion_resource="$s_volume_group"
    else
        s_expansion_resource="$s_disk"
    fi

    local i_free_disk_space_available_mb=$(get_free_space $s_expansion_type $s_expansion_resource)
    local i_free_disk_space_available_gb=$(echo "scale=2; $i_free_disk_space_available_mb/$i_one_kb" | bc)
    local i_total_disk_size_requested_gb="0"
    if [ $flag_use_disk_size = "true" ];then
        i_total_disk_size_requested_gb=$i_disk_size_requested_gb
    else
        i_total_disk_size_requested_gb=$i_free_disk_space_available_gb
    fi
    local i_total_size_logical_volume_expected=$(echo "scale=2; $i_total_disk_size_requested_gb+$(echo "$s_total_size_logical_volume_before" | sed 's/[^0-9|.]*//g')" | bc)
    local i_free_disk_space_available_expected_gb=$(echo "scale=2; $i_free_disk_space_available_gb-$i_total_disk_size_requested_gb" | bc)
    echo "Before expansion:"
    printf "%-70s %-20s\n" "    1. Total free disk space on \"$s_expansion_resource\" " "${i_free_disk_space_available_gb}G"
    printf "%-70s %-20s\n" "    2. Total disk space of the logical volume \"$s_logical_volume\" " "$s_total_size_logical_volume_before"
    printf "%-70s %-20s\n" "    3. Free disk space in the logical volume \"$s_logical_volume\" " "$s_free_disk_space_logical_volume_before"
    printf "%-70s %-20s\n" "    4. Total disk space requested for expansion " "${i_total_disk_size_requested_gb}G"

    if awk -v n1=$i_free_disk_space_available_gb -v n2=$i_total_disk_size_requested_gb 'BEGIN { exit !(n1 < n2) }'; then
        flag_expansion_size_not_available="true"
        echo "ERROR : Disk space requested for expansion of \"$s_logical_volume\" is less than the free disk space available on \"$s_expansion_resource\""
        echo "        Please adjust the --size argument accordingly and re-run the 'csadm system disk expand-lv' command."
        my_exit 1
    fi
    echo "Expected disk space after expansion:"
    printf "%-70s %-20s\n"  "    1. Expected total free disk space on \"$s_expansion_resource\" " "${i_free_disk_space_available_expected_gb}G"
    printf "%-70s %-20s\n"  "    2. Expected total disk space of the logical volume \"$s_logical_volume\" " "${i_total_size_logical_volume_expected}G"
    echo ""
    echo "If this looks good, re-run the above command without --validate option to expand the \"$s_logical_volume\" logical volume."
    my_exit 1

}

# Returns the free disk space in MBs
get_free_space(){
    s_type=$1
    s_resource=$2

    if [ "$s_type" = "vg" ];then
        # Get free space for the volume group in MBs
        i_free_space=$(vgs --noheadings --no-suffix --units b $s_resource | awk '{print $7}' | cut -d. -f1)
    elif [ "$s_type" = "existing-disk" ];then
        # Get available unpartitioned free space for the disk
        i_free_space=$(parted /dev/$s_resource unit b print free | sed -r '/^\s*$/d' | tail -1 | grep "Free Space" | awk '{print $3}' | sed 's/[^0-9]*$//')
    elif [ "$s_type" = "new-disk" ];then
        # For new disk need to get size of the entire disk
        i_free_space=$(lsblk --bytes --noheadings /dev/$s_resource | awk '{print $4}')
    fi

    # To convert byte count into (MBs)
    # i_divisor is initialized in init function
    i_free_space=$(echo $i_free_space | xargs printf "%d/$i_divisor\n" | bc)
    echo $i_free_space
}

# Checks if requested space is available in the target
# Exits with error if not
check_required_space(){
    s_type=$1

    if [ "$s_type" = "vg" ];then
        s_resource="$s_volume_group"
        s_display_string="volume group"
    else
        s_resource="$s_disk"
        s_display_string="disk"
    fi

    echo "Checking the availability of the required space in the $s_display_string \"$s_resource\""

    i_free_disk_space_mb=$(get_free_space $s_type $s_resource)
    if [ $i_free_disk_space_mb -ge $i_one_kb ];then
        i_free_disk_space_gb=$(echo "scale=2; $i_free_disk_space_mb/$i_one_kb" | bc)
        s_display_size="${i_free_disk_space_gb}GB"
    else
        s_display_size="${i_free_disk_space_mb}MB"
    fi

    # Exit if available free disk space is less than
    # the minimum disk limit
    if [ $i_free_disk_space_mb -lt $i_minimum_disk_limit_mb ];then
        echo "Error: Currently there is $s_display_size free disk space available on the $s_display_string \"$s_resource\"" | tee -a $f_log
        echo "       Minimum ${i_minimum_disk_limit_gb}GB of free disk space is required on it to expand the logical volume." | tee -a $f_log
        echo "       Logical volume was not expanded due to above problem." | tee -a $f_log
        my_exit 1
    fi

    # Exit if available free disk space is less than the requested space
    if [ $flag_use_disk_size = "true" -a \
         $i_free_disk_space_mb -lt $i_disk_size_requested_mb ];then
        echo "Error: Currently there is $s_display_size free disk space available on the $s_display_string \"$s_resource\"," | tee -a $f_log
        echo "       which is less than the requested disk space of ${i_disk_size_requested_gb}GB." | tee -a $f_log
        echo "       Logical volume was not expanded due to above problem." | tee -a $f_log
        my_exit 1
    fi
}

create_partition(){
    s_type=$1
    s_operation=$2

    if [ "$s_type" = "existing-disk" ];then
        if [ "$s_operation" = "expand" ];then
            echo "Expanding the partition \"/dev/${s_disk}${i_partition_number}\" $s_tmp"
            my_system "parted /dev/$s_disk resizepart $i_partition_number ${i_resize_total}"
        else
            # Get the starting GB number for the new partition
            i_gb_start=$(parted /dev/$s_disk unit GB print free | sed -r '/^\s*$/d' | tail -1 | grep "Free Space" | awk '{print $1}' | sed 's/[^0-9]*$//')

            # Add the requested size number in GB to it
            i_gb_end=$(echo "scale=2; $i_gb_start + $i_disk_size_requested_gb" | bc)

            echo "Creating the primary partition on \"$s_disk\""
            my_system "parted --script --align optimal /dev/$s_disk mkpart primary ${i_gb_start}GB ${i_gb_end}GB"

            # Get the latest partition number
            i_partition_number=$(parted /dev/$s_disk print | grep primary | awk '{print $1}' | sort | tail -1)
        fi
    elif [ "$s_type" = "new-disk" ];then
        echo "Creating the primary partition on \"$s_disk\""
        my_system "parted --script \
               --align optimal \
               /dev/$s_disk \
               mklabel $s_partition_table_type \
               mkpart primary 0% $s_disk_allocation"

        # It will be first partition for new disk
        i_partition_number=1
    fi
    sleep 5
}

create_physical_volume(){
    s_operation=$1

    if [ "$s_operation" = "expand" ];then
        echo "Resizing the physical volume \"/dev/${s_disk}${i_partition_number}\""
        my_system "pvresize /dev/${s_disk}${i_partition_number}"
    else
        s_new_physical_volume="/dev/${s_disk}${i_partition_number}"
        echo "Creating the physical volume \"$s_new_physical_volume\""
        my_system "pvcreate $s_new_physical_volume"
    fi
}

extend_volume_group(){
    echo "Extending the volume group \"$s_volume_group\" to the physical volume \"$s_new_physical_volume\""
    my_system "vgextend $s_volume_group $s_new_physical_volume"
}

extend_logical_volume(){
    # Extend the logical volume
    if [ $flag_use_disk_size = "false" -o \
         $flag_use_disk = "true" ]; then
        # --size not requested so occupy 100%
        # In case of disk and custom size option, requested disk size is
        # already carved out from the disk in the form of the partition
        # So just extend the volume fully to occupy available space in the volume group
        # That newly available space itself is the requested disk size
        echo "Extending the logical volume \"$s_logical_volume\" to fully occupy the space in the volume group \"$s_volume_group\""
        my_system "lvextend -l +100%FREE $f_logical_volume_path"
    else
        echo "Extending the logical volume \"$s_logical_volume\" by ${i_disk_size_requested_gb}${s_disk_space_extension_unit} in the volume group \"$s_volume_group\""
        my_system "lvextend -L +${i_disk_size_requested_gb}${s_disk_space_extension_unit} $f_logical_volume_path"
    fi

    # Grow the filesystem to fully occupy the logical volume
    my_system "xfs_growfs $s_filesystem_logical_volume"
}

add_disk(){
    # Disk option handling logic
    # Is it new disk (by searching for part in lsblk: lsblk | grep $s_disk | grep part)
    #        no
    #            does this disk has the logical volume's vg?
    #                yes
    #                    does it have enough free disk space
    #                        yes
    #                            resize partition, pvresize, lvextend
    #                        no
    #                            exit with error
    #                no
    #                    exit with error
    #        yes
    #            does it have enough free disk space
    #                yes
    #                    create partition, pv, vgextend etc
    #                no
    #                    exit with error

    # Is it new disk (by searching for part in lsblk: lsblk | grep $s_disk | grep part)
    lsblk | grep $s_disk | grep -w --silent part
    if [ $? -eq 0 ];then
        # This is the existing disk

        pvs | grep $s_volume_group | grep --silent $s_disk
        if [ $? -eq 0 ];then
            # Volume group is present on the disk

            check_required_space "existing-disk"

            # User needs to handle the expansion if the volume group is found on more than one partition
            i_partition_count=$(pvs | grep $s_volume_group | grep $s_disk | wc -l)
            if [ $i_partition_count -gt 1 ];then
                echo "Error: Logical volume $s_logical_volume's volume group $s_volume_group was found on more than one partition on this disk." | tee -a $f_log
                echo "       Specify other option or handle the logical volume expansion manually" | tee -a $f_log
                my_exit 0
            fi

            # Find the partition number
            i_partition_number=$(pvs | grep $s_volume_group | grep $s_disk | awk '{print $1}' | awk '{print substr($0,length,1)}')

            # Find the size occupied by the primary partition of the VG
            i_disk_size_occupied_bytes=$(parted /dev/$s_disk unit b print free | grep primary | awk '{print $3}' | tail -1 | sed 's/[^0-9]*$//')

            if [ $flag_use_disk_size = "true" ];then
                # If size mentioned, add the requested size number to the occupied size number
                i_disk_size_requested_bytes=$(expr $i_disk_size_requested_mb \* $i_divisor)
                i_disk_size_total_bytes=$(expr $i_disk_size_requested_bytes + $i_disk_size_occupied_bytes)

                # Resize by bytes, hence "b" is appended
                i_resize_total=${i_disk_size_total_bytes}b
                s_tmp="by ${i_disk_size_requested_gb}GB"
            else
                # Else resize fully
                i_resize_total="100%"
                s_tmp="to fully occupy the space"
            fi

            create_partition "existing-disk" "expand"

            # Resize the physical volume to fully occupy the partition
            create_physical_volume "expand"

            extend_logical_volume
        else
            # Volume group is not present on the disk
            echo "Warning: The logical volume \"$s_logical_volume\"'s volume group \"$s_volume_group\" is not present on the disk \"$s_disk\"." | tee -a $f_log
            echo "         The disk is already being used by some other volume group(s)." | tee -a $f_log
            echo "         This subcommand does not support the automation of handling of multiple partitions due to complications involved." | tee -a $f_log
            echo "Either re-run this subcommand with different options or use the disk commands provided by the Operating System to handle the requirement." | tee -a $f_log
            echo "Exiting now." | tee -a $f_log
            my_exit 0

            # Create new partition as the volume group is not present on this disk
            # Logic
            # Get the count of primary and extended
            #   parted /dev/sdf unit GB print free | grep --extended-regexp "primary|extended"
            #   4 primary or 3 primary and 1 extended allowed. Total 4
            # Is the count equal to 4?
            #   yes
            #       exit with error: cannot have more than 4 primary partitions
            #   no
            #       check_required_space "existing-disk"
            #       get the starting GB number
            #           parted /dev/sdf unit GB print free | grep "Free Space" | tail -1 | awk '{print $1}' | sed 's/[^0-9]*$//'
            #       add the requested size number in GB to it
            #           echo "2.15 + 2.00" | bc -l
            #       create partition
            #           parted --script --align optimal /dev/$s_disk mkpart primary 1.00GB 2.00GB
            #       create PV
            #       extend VG
            #       extend_logical_volume

#            # Get the count of primary and extended
#            i_count_partition=$(parted /dev/$s_disk print | grep --extended-regexp "primary|extended" | wc -l)
#            if [ $i_count_partition -lt $i_count_partition_limit ];then
#                # Check for available disk space
#                check_required_space "existing-disk"
#
#                # Create the primay partition: unit GB
#                create_partition "existing-disk"
#
#                # Create the physical volume
#                create_physical_volume
#
#                # Extend the volume group
#                extend_volume_group
#
#                extend_logical_volume
#            else
#                echo "Error: Cannot create the partition" | tee -a $f_log
#                echo "       Disk \"$s_disk\" already has $i_count_partition primary/extended partitions" | tee -a $f_log
#                echo "       Logical volume was not expanded due to above problem." | tee -a $f_log
#                my_exit 1
#            fi
        fi

    else
        # This is the new disk
        check_new_disk_empty

        check_required_space "new-disk"

        if [ $flag_use_disk_size = "true" ];then
            # Disk size to expand is explicitly mentioned
            s_disk_allocation="${i_disk_size_requested_gb}GB"
        else
            # Disk size to expand is not mentioned, so expand 100%
            s_disk_allocation="100%"
        fi

        # Create the partition
        create_partition "new-disk"

        # Create the physical volume
        create_physical_volume

        # Add the newly created physical volume to the volume group
        extend_volume_group

        extend_logical_volume
    fi
    echo "Successfully expanded the logical volume \"$s_logical_volume\" on the disk \"$s_disk\"" | tee -a $f_log
    print_summary
}

expand_lv(){
    echo "Logical volume \"$s_logical_volume\" is on the volume group \"$s_volume_group\"" | tee -a $f_log
    if [ "$flag_use_vg" = "true" ];then
        check_required_space "vg"

        # Extend the logical volume by the requested size
        extend_logical_volume

        echo "Successfully expanded the logical volume \"$s_logical_volume\" on the volume group \"$s_volume_group\"" | tee -a $f_log
        print_summary
    elif [ "$s_disk" != "" ];then
        add_disk $s_disk
    fi
}

pre_check(){
    # The following condition checks whether the logged-in user is root or not.
    if [[ $EUID -ne 0 ]]; then
        echo "This script must be run as root. Please sudo su and re-run script"
        my_exit 1
    fi
}

init(){
    f_log="/var/log/cyops/csadm/system.log"
    flag_expand_lv="false"
    flag_use_vg="false"
    flag_use_disk="false"
    flag_use_disk_size="false"
    flag_validate_expand_lv="false"
    s_disk=""
    i_count_partition_limit=4

    i_disk_size_requested_gb=0
    i_disk_size_requested_mb=0

    i_one_kb=1024

    # To get free space in MB
    i_divisor_to_get_mb=$(expr $i_one_kb \* $i_one_kb)
    i_divisor=$i_divisor_to_get_mb

    # Extend in gigabytes
    s_disk_space_extension_unit="G"

    # Set minimum required disk space in GBs and MBs
    i_minimum_disk_limit_gb=1
    i_minimum_disk_limit_mb=$(expr $i_minimum_disk_limit_gb \* $i_one_kb)

    # In case partition table type is msdos then need to set the variable accordingly
    s_partition_table_type="gpt"
    parted --list | grep Partition | grep -q msdos
    if [ $? -eq 0 ];then
        s_partition_table_type="msdos"
    fi
}

process_parameters(){
    OPTS=$(getopt -o ab:c:de:f --long expand-lv,logical-volume:,disk:,use-vg,size:,validate -n "disk" -- "$@")

    s_csadm_command="csadm system disk"
    eval set -- "$OPTS"

    while true; do
        case "$1" in
            -a|--expand-lv)
                flag_expand_lv="true"
                s_csadm_command="$s_csadm_command expand-lv"
                shift
            ;;

            -b|--logical-volume)
                s_logical_volume=$2
                s_csadm_command="$s_csadm_command --logical-volume $2"
                shift 2
            ;;

            -c|--disk)
                flag_use_disk="true"
                s_disk=$2
                s_csadm_command="$s_csadm_command --disk $2"
                shift 2
            ;;

            -d|--use-vg)
                flag_use_vg="true"
                s_csadm_command="$s_csadm_command --use-vg"
                shift
            ;;

            -e|--size)
                flag_use_disk_size="true"
                i_disk_size_requested_gb=$2
                i_disk_size_requested_mb=$(expr $i_disk_size_requested_gb \* $i_one_kb)
                s_csadm_command="$s_csadm_command --size $2"
                shift 2
            ;;

            -f|--validate)
                flag_validate_expand_lv="true"
                s_csadm_command="$s_csadm_command --validate"
                shift
            ;;

            -- ) shift
                 break ;;

            * ) break ;;
      esac
    done

    # Get the volume group name from the logical volume
    s_volume_group=$(lvs -o lv_name,vg_name | grep -w $s_logical_volume | awk '{print $2}')
    f_logical_volume_path=$(lvs -o lv_name,lv_path | grep -w $s_logical_volume | awk '{print $2}')

    # Get the file system path on which the logical volume is mounted
    s_filesystem_logical_volume=$(df -kh | sed "s/--/-/g" | grep -w "$s_volume_group-$s_logical_volume" | awk '{print $6}')

    # Get the size before making the changes, to show in the summary
    s_total_size_logical_volume_before=$(df -kh | sed "s/--/-/g" | grep -w "$s_volume_group-$s_logical_volume" | awk '{print $2}')

    # Check if size returned is in MB. if it is in MB then convert it to GB.
    if [[ $s_total_size_logical_volume_before == *"M" ]]; then
        i_total_size_logical_volume_before=$(echo "$s_total_size_logical_volume_before" | sed "s/M//g")
        s_total_size_logical_volume_before=$(echo "scale=2; $i_total_size_logical_volume_before/$i_one_kb" | bc)"G"
    fi
    s_free_disk_space_logical_volume_before=$(df -kh | sed "s/--/-/g" | grep -w "$s_volume_group-$s_logical_volume" | awk '{print $4}')

    add_date_to_log
    echo "Input command: $s_csadm_command" >> $f_log
}

# Main
init

pre_check

process_parameters $@

if [ "$flag_validate_expand_lv" = "true" ];then
    check_lvm_mounted
    validate_summary 2>>$f_log
    my_exit 0
fi

if [ "$flag_expand_lv" = "true" ];then
    check_lvm_mounted
    expand_lv       2>>$f_log
# Add elifs for more actions (like "expand-lv"), here
fi
my_exit 0
