#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import argparse, sys, os
import OpenSSL
import datetime
from shutil import copyfile
import glob
import pem
from dateutil import parser
from fsr.commonlibs.utils import CustomHelpFormatter
sys.path.append("/opt/cyops/scripts/")
from csadm.utils import print_error, execute_cmd_with_shell, write_file, is_safe_path, read_file, execute_cmd

DIR_TRUSTSTORE = "/opt/cyops/configs/rabbitmq/truststore/"
CA_CERT = "/opt/cyops/configs/rabbitmq/ssl/cyopsca/cacert.pem"

def is_ca_cert_already_added(ca_cert):
    current_ca_cert_contents = read_file(CA_CERT)
    ca_cert_contents = read_file(ca_cert)
    if ca_cert_contents in current_ca_cert_contents:
        return True
    return False

def mq_serv_op(op):
    serv_name = "rabbitmq-server"
    cmd = "systemctl " + op + " " + serv_name
    msg = "Executing " + serv_name + " " + op
    print(msg, flush=True)
    ret, out, err = execute_cmd(cmd)
    if ret != 0:
        print("Failed to " + op + " " + serv_name)
        sys.exit(1)

def trusttore_write(ca_file_contents, append_if_already_exists=True):
    print("Refreshing the truststore", flush=True)
    write_file(CA_CERT, ca_file_contents, append_if_already_exists)
    mq_serv_op("restart")

def get_ca_certs(ca_bundle=CA_CERT):
    certs = pem.parse_file(ca_bundle)
    if (certs is None) or len(certs) == 0:
        return None

def get_common_name(cert):
    subject_components = cert.get_subject().get_components()
    if subject_components is None:
        return None
    for subject_component in subject_components:
        if subject_component[0].decode("utf-8") == 'CN':
            return subject_component[1].decode("utf-8")

def print_ca_details(ca_bundle):
    certs = pem.parse_file(ca_bundle)
    if (certs is None) or len(certs) == 0:
        return
    KEY_STRING_SIZE = 13
    print("Name".ljust(KEY_STRING_SIZE) + ": " +  os.path.basename(ca_bundle))

    msg = "Subject".ljust(KEY_STRING_SIZE) + ": " + "{0}" + "\n"
    msg = msg + "Serial Number".ljust(KEY_STRING_SIZE) + ": " + "{1}" + "\n"
    msg = msg + "Expiry date".ljust(KEY_STRING_SIZE) + ": " + "{2}" + "\n"
    msg = msg + "Expired?".ljust(KEY_STRING_SIZE) + ": " + "{3}"

    for cert in certs:
        _cert =  OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, str.encode(str(cert)))
        subject = _cert.get_subject()
        subject_str = "".join("/{0:s}={1:s}".format(name.decode(), value.decode()) for name, value in subject.get_components())
        sn = str(hex(_cert.get_serial_number()))
        datetime_struct_to = parser.parse(_cert.get_notAfter().decode("UTF-8"))
        expiry_date = datetime_struct_to.strftime('%b-%d-%Y %H:%M:%S')
        expired = str(_cert.has_expired())
        print(msg.format(subject_str, sn, expiry_date, expired))

def truststore_copy_ca_bundle(ca_bundle):
    ca_list = glob.glob(DIR_TRUSTSTORE + "/" + "*.pem")
    input_ca_bundle_contents = read_file(ca_bundle)
    for ca_cert in ca_list:
        ca_cert_contents = read_file(ca_cert)
        if input_ca_bundle_contents == ca_cert_contents:
            return
    current_date = datetime.datetime.now()
    current_date_str = current_date.strftime('%Y-%m-%d-%H-%M-%S')
    target_file = DIR_TRUSTSTORE + current_date_str + ".pem"
    print("Adding [" + ca_bundle + "] to the truststore as [" + os.path.basename(target_file) + "]", flush=True)
    copyfile(ca_bundle, target_file)

def truststore_add(ca_bundle):
    if not os.path.isfile(ca_bundle):
        print_error("File [" + ca_bundle + "] does not exist")
        sys.exit(1)
    try:
        OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, open(ca_bundle).read())
    except Exception as e:
        print_error("Unable to read the CA certificate chain [" + ca_bundle + "]. Check if file contains valid certificate chain in PEM format.")
        sys.exit(1)
    if is_ca_cert_already_added(ca_bundle):
        print("Already added to the truststore")
        #
        # This copy is required for the cases where user try to add main CA bundle again after deploying MQ cert.
        # Also, it is possible that user append new CA bundle manually to the main CA file and then try to add that
        # bundle again using CLI. In this case, it should copy the new CA bundle file to the truststore dir, so that
        # it will be listed on 'list'.
        #
        truststore_copy_ca_bundle(ca_bundle)
        sys.exit(0)
    truststore_copy_ca_bundle(ca_bundle)
    truststore_refresh_add()
    print("Added [" + ca_bundle + "] to the truststore successfully")

def truststore_list():
    ca_list = glob.glob(DIR_TRUSTSTORE + "/" + "*.pem")
    if (ca_list is None) or len(ca_list) == 0:
        print("CA certificate is not added to the truststore")
        return
    for ca in ca_list:
        print_ca_details(ca)
        print("")

def truststore_remove(comma_separated_ca_bundle_name):
    ca_bundle = comma_separated_ca_bundle_name.split(",")
    certs_to_remove = []
    for ca_bundle_name in ca_bundle:
        full_path_ca = DIR_TRUSTSTORE + "/" + ca_bundle_name
        if not os.path.isfile(full_path_ca):
            print_error("Unable to find the CA certificate [" + ca_bundle_name + "] in the truststore. Check 'csadm mq truststore list'.")
            sys.exit(0)
        print("Removing the CA certificate from the truststore [" + ca_bundle_name + "]", flush=True)
        if not is_safe_path(DIR_TRUSTSTORE, full_path_ca):
            print_error('Not allowed!\n')
            sys.exit(1)
        certs_to_remove.append(full_path_ca)
    truststore_refresh_remove(certs_to_remove)
    print("Removed the CA certificate from the truststore")

def truststore_refresh_add():
    ca_list = glob.glob(DIR_TRUSTSTORE + "/" + "*.pem")
    current_ca_cert_contents = read_file(CA_CERT)
    new_ca_certs = []
    for ca_cert in ca_list:
        ca_cert_contents = read_file(ca_cert)
        if ca_cert_contents not in current_ca_cert_contents:
            new_ca_certs.append(ca_cert_contents)
    if len(new_ca_certs) != 0:
        new_ca_certs_contents = "\n".join(new_ca_certs) + "\n"
        trusttore_write(new_ca_certs_contents, append_if_already_exists=True)
    print("Truststore refreshed successfully")

def truststore_refresh_remove(certs_to_remove=[]):
    if len(certs_to_remove) == 0:
        print("Truststore refreshed successfully")
        return
    current_ca_cert_contents = read_file(CA_CERT)
    for cert_to_remove in certs_to_remove:
        cert_to_remove_contents = read_file(cert_to_remove)
        os.remove(cert_to_remove)
        if cert_to_remove_contents in current_ca_cert_contents:
            current_ca_cert_contents = current_ca_cert_contents.replace(cert_to_remove_contents, "\n")
    if current_ca_cert_contents.strip() != "":
        #
        # Only single CA cert to remove, and if it is same as main CA.
        # This situation occurs when user tries to add main CA bundle(i.e current MQ CA) using
        # truststore add and then try to remove same.
        #
        trusttore_write(current_ca_cert_contents, append_if_already_exists=False)
    print("Truststore refreshed successfully")

# Below sub-routine is getting used in /opt/cyops/adm/subcmd/mq
def define_parser(parser, subparser):
    parser_trust_store = subparser.add_parser('truststore', help="Manages the RabbitMQ truststore")
    visible_options = "{add,remove,list,refresh}"
    sub_parser_trust_store = parser_trust_store.add_subparsers(metavar=visible_options)
    parser_add = sub_parser_trust_store.add_parser('add', help="Adds the CA certificate to the truststore")
    parser_add.set_defaults(operation='operation-mq-truststore-add')
    parser_add.add_argument('--ca-cert',
                            help="Path to the CA certificate PEM file (including the file name) that contains the complete chain of CA certificates",
                            dest="ca_cert_path",
                            type=str,
                            required=True)

    parser_remove = sub_parser_trust_store.add_parser('remove', help="Removes the CA certificate from the truststore")
    parser_remove.set_defaults(operation='operation-mq-truststore-remove')
    parser_remove.add_argument('--ca-cert-name',
                               help="The CA certificate bundle PEM file name from 'csadm mq truststore list'",
                               dest="ca_cert_name",
                               type=str,
                               required=True)

    parser_list = sub_parser_trust_store.add_parser('list', help="Lists the CA certificate from the truststore")
    parser_list.set_defaults(operation='operation-mq-truststore-list')

    parser_refresh = sub_parser_trust_store.add_parser('refresh', help="Refreshes the truststore")
    parser_refresh.set_defaults(operation='operation-mq-truststore-refresh')

    for name, subp in sub_parser_trust_store.choices.items():
        subp._positionals.title = 'subcommand options are'
        subp._optionals.title = 'subcommand options are'

def main():
    parser = argparse.ArgumentParser(prog="csadm mq truststore", formatter_class=CustomHelpFormatter)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'
    subparser = parser.add_subparsers()
    define_parser(parser, subparser)

    args = parser.parse_args()

    if not 'operation' in dir(args):
        parser.print_help(sys.stderr)
        sys.exit(1)

    if str(args.operation) == 'operation-mq-truststore-add':
        truststore_add(args.ca_cert_path)

    elif str(args.operation) == 'operation-mq-truststore-remove':
        truststore_remove(args.ca_cert_name)

    elif str(args.operation) == 'operation-mq-truststore-refresh':
        truststore_refresh_add()

    elif str(args.operation) == 'operation-mq-truststore-list':
        truststore_list()

if __name__ == '__main__':
    main()
    sys.exit(0)