#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import argparse, sys, socket, json, os, re, subprocess, logging
sys.path.append("/opt/cyops/scripts/")
from csadm.models.proxy_config import setHttpProxy, setNoProxy, setHttpsProxy, \
                                      update_etc_yum_conf, get_full_proxy_url, update_etc_environment, \
                                      list_proxy, restore_proxy_settings, remove_proxy_settings

from fsr.commonlibs.utils import CustomHelpFormatter
from fsr.commonlibs.utils import get_list_appliance_flavor
from fsr.commonlibs.consts import APPLIANCE_FLAVOR_FORTICLOUD

if os.getuid() != 0:
    print("This script must be run as 'root'")
    exit(1)


program_name = "csadm network"

def add_ipv6_subparser(subparsers):
    parser_ipv6 = subparsers.add_parser('ipv6', help='Handle IPv6 operations')
    parser_ipv6.set_defaults(operation='ipv6')
    parser_ipv6.add_argument('--enable', action='store_true', help='Enable IPv6 protocol. Requires system reboot.', dest='enable_ipv6')
    parser_ipv6.add_argument('--skip-services-restart',
                            action='store_true',
                            help=argparse.SUPPRESS,
                            dest='skip_services_restart')

def validate_host(value):
    # Regular expression pattern to match ipv4 address
    ipv4_pattern = r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
    # Regular expression pattern to match ipv6 address
    ipv6_pattern = r"^(?:[0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}$"
    # Regular expression pattern to match hostname
    hostname_pattern = r"^(?!-)(?!(?:\d+\.)+\d+$)[A-Za-z0-9-]+(?:\.[A-Za-z0-9-]+)*$"

    if re.match(ipv4_pattern, value):
        return value
    elif re.match(ipv6_pattern, value):
        return value
    elif re.match(hostname_pattern, value):
        return value
    
    # If not matches with IP address or valid hostname
    print("Error: Invalid value for --host. The value must be an IPv4 address, uncompressed IPv6 address, or hostname.")
    exit(1)

def add_set_https_proxy_subparser(subparsers):
    parser_sethttpsproxy = subparsers.add_parser('set-https-proxy', help='Set proxy to be used for outbound https requests')
    parser_sethttpsproxy.set_defaults(operation='set-https-proxy')
    parser_sethttpsproxy.add_argument('--host', help='Proxy hostname', dest='proxy_hostname', required=True, type=validate_host)
    parser_sethttpsproxy.add_argument('--port', help='Proxy port', dest='proxy_port', required=True, type=int)
    parser_sethttpsproxy.add_argument('--protocol', help=argparse.SUPPRESS, dest='proxy_protocol', default='http')
    parser_sethttpsproxy.add_argument('--user', help='Proxy username', dest='proxy_username', default=None, required='--password' in sys.argv)
    parser_sethttpsproxy.add_argument('--password', help='Proxy password', dest='proxy_password', default=None)
    parser_sethttpsproxy.add_argument('--update-env-only', help=argparse.SUPPRESS, dest='update_env_only', action = 'store_true', default=False)


def add_set_http_proxy_subparser(subparsers):
    parser_sethttpproxy = subparsers.add_parser('set-http-proxy', help='Set proxy to be used for outbound http requests')
    parser_sethttpproxy.set_defaults(operation='set-http-proxy')
    parser_sethttpproxy.add_argument('--host', help='Proxy hostname', dest='proxy_hostname', required=True, type=validate_host)
    parser_sethttpproxy.add_argument('--port', help='Proxy port', dest='proxy_port', required=True, type=int)
    parser_sethttpproxy.add_argument('--protocol', help=argparse.SUPPRESS, dest='proxy_protocol', default='http')
    parser_sethttpproxy.add_argument('--user', help='Proxy username', dest='proxy_username', default=None, required='--password' in sys.argv)
    parser_sethttpproxy.add_argument('--password', help='Proxy password', dest='proxy_password', default=None)
    parser_sethttpproxy.add_argument('--update-env-only', help=argparse.SUPPRESS, dest='update_env_only', action = 'store_true', default=False)


def add_set_no_proxy_subparser(subparsers):
    parser_setnoproxy = subparsers.add_parser('set-no-proxy', help='List of addresses not be routed through proxy')
    parser_setnoproxy.set_defaults(operation='set-no-proxy')
    parser_setnoproxy.add_argument('--host', help='Comma-separated list of hostnames', dest='hostname', required=True)
    parser_setnoproxy.add_argument('--update-env-only', help=argparse.SUPPRESS, dest='update_env_only', action = 'store_true', default=False)


def add_list_proxy_subparser(subparsers):
    parser_list_proxy = subparsers.add_parser('list-proxy', help='List proxy details')
    parser_list_proxy.set_defaults(operation='list-proxy')

def add_restore_proxy_settings_subparser(subparsers):
    parser_restore_proxy = subparsers.add_parser('restore-proxy')
    parser_restore_proxy.set_defaults(operation='restore-proxy')

def add_remove_proxy_subparser(subparsers):
    parser_remove_proxy = subparsers.add_parser('remove-proxy', help='Remove the configured proxy')
    parser_remove_proxy.set_defaults(operation='remove-proxy')
    parser_remove_proxy.add_argument('--update-env-only',
                                     help=argparse.SUPPRESS,
                                     dest='update_env_only',
                                     action='store_true',
                                     default=False)


def define_parser_common(parser, subparsers):
    visible_options = "{set-no-proxy,list-proxy}"
    subparsers.metavar = visible_options
    add_set_no_proxy_subparser(subparsers)
    add_list_proxy_subparser(subparsers)

def define_parser_enterprise(parser, subparsers):
    visible_options = "{ipv6,set-https-proxy,set-http-proxy,set-no-proxy,list-proxy, remove-proxy}"
    subparsers.metavar = visible_options
    add_ipv6_subparser(subparsers)
    add_set_https_proxy_subparser(subparsers)
    add_set_http_proxy_subparser(subparsers)
    add_restore_proxy_settings_subparser(subparsers)
    add_remove_proxy_subparser(subparsers)

def define_parser(parser):

    subparsers = parser.add_subparsers()

    define_parser_common(parser, subparsers)

    list_appliance_flavor = get_list_appliance_flavor()

    if list_appliance_flavor is None:
        define_parser_enterprise(parser, subparsers)
    elif APPLIANCE_FLAVOR_FORTICLOUD not in list_appliance_flavor:
            define_parser_enterprise(parser, subparsers)

    for name, subp in subparsers.choices.items():
        subp._positionals.title = 'subcommand options are'
        subp._optionals.title = 'subcommand options are'


def get_dash_string():

    return "------------------------------------------------"

def print_note(msg):

    dash_string = get_dash_string()
    note = dash_string + "\n"
    note = note + "Note:" + "\n"
    note = note + msg + "\n"
    note = note + dash_string + "\n"
    print (note)
    return

def print_reload_note(need_to_restart_services_manually=True):
    note_msg = "Proxy Setting applied successfully."
    if not need_to_restart_services_manually:
        print_note(note_msg)
        return
    note_msg = note_msg + " For the settings to take effect, you need to logout and connect to the session again to load the latest environment."
    print_note(note_msg)

def start():
    parser = argparse.ArgumentParser(prog=program_name, formatter_class=CustomHelpFormatter)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'

    define_parser(parser)

    args = parser.parse_args()

    if not 'operation' in dir(args):
        parser.print_help(sys.stderr)
        exit(1)

    if str(args.operation) == 'ipv6':
        if args.enable_ipv6:
            cmd = "/opt/cyops/scripts/ipv6.sh --enable"
            if args.skip_services_restart:
                cmd = cmd + " --skip-services-restart"
            ret = os.system(cmd)
            if ret != 0:
                exit(1)
            return

    elif str(args.operation) == 'set-https-proxy':
        protocol = args.proxy_protocol
        host = args.proxy_hostname
        port = args.proxy_port
        username = args.proxy_username
        password = args.proxy_password
        update_env_only = args.update_env_only

        if '//' in host:
            host = host.split('//')[1]

        proxy = protocol + "://" + host + ":" + str(port)
        update_etc_yum_conf(proxy, username, password)
        new_value = get_full_proxy_url(protocol, host, port, username, password)
        key = "https_proxy"
        # The users having to login shell will get proxy variables in env when we update /etc/environment.
        # For example (sudo -u nginx <pip_command_here>)
        update_etc_environment(key, new_value, append_to_existing_value=False)

        if update_env_only:
            return
        need_to_restart_services_manually = True
        if not setHttpsProxy(protocol + "://" + host, port, username, password):
            #Failed to set https proxy list
            need_to_restart_services_manually = False
        print_reload_note(need_to_restart_services_manually)
        return

    elif str(args.operation) == 'set-http-proxy':

        protocol = args.proxy_protocol
        host = args.proxy_hostname
        port = args.proxy_port
        username = args.proxy_username
        password = args.proxy_password
        update_env_only = args.update_env_only

        if '//' in host:
            host = host.split('//')[1]

        proxy = protocol + "://" + host + ":" + str(port)
        update_etc_yum_conf(proxy, username, password)
        new_value = get_full_proxy_url(protocol, host, port, username, password)
        key = "http_proxy"
        update_etc_environment(key, new_value, append_to_existing_value=False)
        if update_env_only:
            return
        need_to_restart_services_manually = True
        if not setHttpProxy(protocol + "://" + host, port, username, password):
            #Failed to set http proxy list
            need_to_restart_services_manually = False
        print_reload_note(need_to_restart_services_manually)
        return 

    elif str(args.operation) == 'set-no-proxy':
        proxy = args.hostname
        update_env_only = args.update_env_only

        key = "no_proxy"
        update_etc_environment(key, proxy, append_to_existing_value=False)
        if update_env_only:
            return
        need_to_restart_services_manually = True
        if not setNoProxy(proxy):
           #Failed to set no proxy list
           need_to_restart_services_manually = False
        print_reload_note(need_to_restart_services_manually)
        return

    elif str(args.operation) == 'list-proxy':
        list_proxy()
        return

    elif str(args.operation) == 'restore-proxy':
        restore_proxy_settings()
        return

    elif str(args.operation) == 'remove-proxy':
        is_proxy_removed = remove_proxy_settings(args.update_env_only)
        if is_proxy_removed:
            note_msg = "For the settings to take effect, you need to logout and connect to the session again to load the latest environment."
            print_note(note_msg)
        return

    # Execution should not come to this stage. If comes, then print help
    parser.print_help(sys.stderr)

if __name__ == '__main__':
    start()

exit(0)
