#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import argparse
import os
import sys
import validators

if os.getuid() != 0:
    print("This script must be run as 'root'")
    exit(1)
sys.path.append("/opt/cyops/scripts/")
from csadm.utils import *
from fsr.commonlibs.utils import CustomHelpFormatter

def main() :
    parser = argparse.ArgumentParser(prog='csadm package', formatter_class=CustomHelpFormatter)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'
    subparsers = parser.add_subparsers(dest='sub_command', help='subcommand help')
    subparsers.required = True
    install_parser = subparsers.add_parser('install', help='Install RPM Package')
    install_parser.set_defaults(func=yum_operation)
    install_parser.add_argument('-t', '--type',
                                type=str,
                                action='store',
                                choices=['connector'],
                                required=True,
                                help='Type of the package to install')

    install_parser.add_argument('-n', '--name',
                                type=str,
                                action='store',
                                required=True,
                                help='Name of the RPM to install (e.g. the connector name)')
    install_parser.add_argument('--no-interaction',
                                action='store_true',
                                help='No confirmation required')

    update_parser = subparsers.add_parser('update', help='Update RPM package')
    update_parser.set_defaults(func=yum_operation)
    update_parser.add_argument('-t', '--type',
                               type=str,
                               action='store',
                               choices=['connector'],
                               required=True,
                               help='Type of the package to update')

    update_parser.add_argument('-n', '--name',
                               type=str,
                               action='store',
                               required=True,
                               help='Name of the RPM to update (e.g. the connector name)')

    update_parser.add_argument('--no-interaction',
                                action='store_true',
                                help='No confirmation required')

    remove_parser = subparsers.add_parser('remove', help='Remove RPM package')
    remove_parser.set_defaults(func=yum_operation)
    remove_parser.add_argument('-t', '--type',
                               type=str,
                               action='store',
                               choices=['connector'],
                               required=True,
                               help='Type of the package to remove')

    remove_parser.add_argument('-n', '--name',
                               type=str,
                               action='store',
                               required=True,
                               help='Name of the RPM to remove (e.g. the connector name)')
    remove_parser.add_argument('--no-interaction',
                                action='store_true',
                                help='No confirmation required')
    
    content_hub = subparsers.add_parser('content-hub', help='Perform a full synchronization for content-hub')
    content_hub.set_defaults(func=handle_content_hub)
    
    visible_options = "{sync}"
    sub_parser_content_hub = content_hub.add_subparsers(metavar=visible_options)

    sub_parser_sync = sub_parser_content_hub.add_parser("sync", help="Synchronize content-hub")
    sub_parser_sync.set_defaults(operation='content-hub-sync', func=handle_content_hub)
    sub_parser_sync.add_argument("--force",dest='force_sync', help="Force synchronization of content-hub", action="store_true")
    
    args = parser.parse_args()
    
    if 'func' in dir(args):
        args.func(args)
        return
    
def handle_content_hub(args):
    if 'operation' not in dir(args) or args.operation is None or not args.operation.startswith('content-hub'):
        os.system('csadm package content-hub -h')
        sys.exit(1)
    if args.operation == 'content-hub-sync':
        sync_content_hub(args.force_sync)
        
def sync_content_hub(force_sync=False):
    sync_cmd='sudo -u nginx php /opt/cyops-api/bin/console fortisoar:contenthub:sync'
    if force_sync == True:
        sync_cmd = sync_cmd + ' -fs'
    ret = os.system(sync_cmd)
    if ret != 0:
        print("Synchronization to content-hub failed")
        sys.exit(1)
    print("Synchronization to content-hub is successful")
        
def is_connector_installed(connector_name):
    # No need sudoers permission for /bin/rpm as nginx can run the query on rpm.
    cmd = "/bin/rpm -q " + connector_name + " &>/dev/null"
    ret = os.system(cmd)
    if ret != 0:
        return False
    return True

def yum_ops(sub_command, pkg_name):
    try:
        cmd = 'yum -y {} {}'.format(sub_command, pkg_name)
        #print(cmd)
        os.system(cmd)
    except Exception as err:
        print("Error: " + str(err))

#The yum_operation function will use for install, update, remove functions.
def yum_operation(args):
    conn_name_starts = 'cyops-connector'
    try:
        url_bool = validators.url(args.name)
        #print(name_validation)
        #the condition will exit if the --name is url (not only the connector name)
        if url_bool:
            print('command only accepts connector name.')
            exit(1)
        #the condition will exit if the connector name not starts with cyops-connector
        if not args.name.startswith(conn_name_starts):
            print('No connector with the name "{}" is available in the FortiSOAR connector repository. Ensure that the connector name provided is in the format {} and try again.'.format(args.name, conn_name_starts))
            exit(1)

        else:
        #The operation will go ahead without user's confirmation if user added --no-interaction argument.
        #if there is no "--no-interaction" argument in the command the command will ask for users second confirmation.
        #if args.type == 'connector' and args.name.startswith(conn_name_starts) and not url_bool:
            yum_op_name = args.sub_command
            if yum_op_name == 'install' and is_connector_installed(str(args.name)):
                yum_op_name = "reinstall"
            if args.no_interaction:
                yum_ops(yum_op_name, args.name)

            else:
                message="Do you want to {} {}? (Enter yes/no)?" .format(yum_op_name, args.name)
                result = take_choice_input(validation_msg=message)
                if result.lower() == 'yes':
                    yum_ops(yum_op_name, args.name)

    except Exception as err:
        print("Error: " + str(err))


if __name__ == '__main__' :
    main()
