#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import sys,os,psutil,argparse
import glob
import signal
if os.getuid() != 0:
    print("This script must be run as 'root'")
    sys.exit(1)
sys.path.append("/opt/cyops/scripts/")
from fsr.commonlibs.utils import CustomHelpFormatter, is_forticloud_instance, is_enterprise_instance, is_docker
from csadm.logs.logger_defs import get_logger
from csadm.utils import *
import fsr.fortimonitor.handlers.cli
LOG_FILE = "/var/log/cyops/csadm/system.log"
MODE_TRIAL = 'trial'
MODE_PROD = 'prod'
MODE_ALL = [MODE_TRIAL, MODE_PROD]
RESOURCE_ES = "elasticsearch"
RESOURCE_PG = "postgres"
RESOURCE_TOMCAT = "tomcat"
RESOURCE_PHP = "php"
RESOURCE_MQ = "mq"
RESOURCE_CELERY = "celery"
RESOURCE_HARDWARE = "hardware"
DIR_DATA_RESET = "/var/lib/pgsql/.data-reset"
SYSTEM_CONFIG = "/opt/cyops/configs/common/system-config.yml"

logger = get_logger(__name__, LOG_FILE)
set_logger_for_utils(logger)

config = None
args = None
services_to_restart = []
######
# add helper functions below this
#########

def check_positive_integer(i):
    i = int(i)
    if i <= 0:
        raise argparse.ArgumentTypeError("%s needs to be a positive integer" % i)
    return i

def signal_handler(signal, frame):
    sys.exit(1)

signal.signal(signal.SIGINT, signal_handler)

def populate_config():
    global config
    config = read_yaml(SYSTEM_CONFIG)

def get_config(resource):
    mode = args.mode
    if mode not in config.keys():
        return None
    if resource not in config[mode].keys():
        return None
    return config[mode][resource]

def get_percent_diff(required, current):
    change_percent = ((float(required) - current)/required) * 100
    return float('%.2f' % (change_percent))

def get_cpu_core_count():
    return int(os.cpu_count())

def get_ram():
    physical_memory = psutil.virtual_memory()
    physical_memory_details = dict(physical_memory._asdict())
    in_gb = float(physical_memory_details['total'])/(1024*1024*1024)
    return float('%.2f' % (in_gb))

def postgres_version():
    cmd = "psql --version | egrep -o '[0-9]{1,}\.' | cut -d'.' -f1"
    ret, out, err = execute_cmd_with_shell(cmd)
    return out

def restart_services():
    global services_to_restart
    if len(services_to_restart) == 0:
        return
    serv_list = ' '.join(services_to_restart)
    cmd = "systemctl restart " + serv_list
    logger.info("Restarting services [" + serv_list + "]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to restart services with error [" + str(err) + "]")

def enable_celery_autoscale():
    celery_conf = "/etc/celery/celeryd.conf"
    cmd = "sed -i \"s/^\s*#\\+\s*\(CELERYD_OPTS\s*=.*autoscale.*$\)/\\1/g\" " + celery_conf
    logger.info("Enabling the celery autoscaling")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to enable the celery autoscaling with error [" + str(err) + "]")

def disable_celery_autoscale():
    celery_conf = "/etc/celery/celeryd.conf"
    cmd = "sed -i \"s/^\s*CELERYD_OPTS\s*=.*autoscale/#&/g\" " + celery_conf
    logger.info("Disabling the celery autoscaling")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to disable the celery autoscaling with error [" + str(err) + "]")

######
# add helper functions above this
#########

def configure_es(xms_size_in_gb=4, xmx_size_in_gb=4):
    config_es = get_config(RESOURCE_ES)
    if config_es is not None:
        xms_size_in_gb = int(config_es['xms'])
        xmx_size_in_gb = int(config_es['xmx'])
    jvm_options_es = "/etc/elasticsearch/jvm.options.d/fsr.options"
    logger.info("Configuring the Elasticsearch Xms memory [" + str(xms_size_in_gb) + "GB]")
    cmd = "sed -i \"s/^-Xms.*$/-Xms" + str(xms_size_in_gb) + "g/\" " + jvm_options_es
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xms value for the Elasticsearch with error [" + str(err) + "]")
    cmd = "sed -i \"s/^-Xmx.*$/-Xmx" + str(xmx_size_in_gb) + "g/\" " + jvm_options_es
    logger.info("Configuring the Elasticsearch Xmx memory [" + str(xmx_size_in_gb) + "GB]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xmx value for the Elasticsearch with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("elasticsearch")

def configure_pg(shared_buffer_in_mb=2048, work_mem_in_mb=16):
    config_pg = get_config(RESOURCE_PG)
    if config_pg is not None:
        shared_buffer_in_mb = int(config_pg['shared_buffer'])
        work_mem_in_mb = int(config_pg['work_mem'])
    PG_VERSION = postgres_version()
    PG_DIR = "/var/lib/pgsql"
    PG_DATA_DIR = PG_DIR + "/" + str(PG_VERSION) + "/data"
    PG_CONF = PG_DATA_DIR + "/postgresql.conf"
    cmd = "sed -i 's/^\s*shared_buffers .*/shared_buffers = " + str(shared_buffer_in_mb) + "MB/' " + PG_CONF
    logger.info("Configuring the PostgreSQL shared buffer [" + str(shared_buffer_in_mb) + "MB]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure the PostgreSQL shared buffer with error [" + str(err) + "]")

    cmd = "sed -i 's/^\s*work_mem .*/work_mem = " + str(work_mem_in_mb) + "MB/' " + PG_CONF
    logger.info("Configuring the PostgreSQL work memory [" + str(work_mem_in_mb) + "MB]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure the PostgreSQL work memory with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("postgresql-" + str(PG_VERSION))

def configure_tomcat(xms_size_in_mb=2048, xmx_size_in_mb=4096):
    config_tomcat = get_config(RESOURCE_TOMCAT)
    if config_tomcat is not None:
        xms_size_in_mb = int(config_tomcat['xms'])
        xmx_size_in_mb = int(config_tomcat['xmx'])
    conf_tomcat = "/opt/cyops-tomcat/bin/setenv.sh"
    logger.info("Configuring the Tomcat Xms memory [" + str(xms_size_in_mb) + "MB]")
    cmd = "sed -i \"s/-Xms.*M/-Xms" + str(xms_size_in_mb) + "M/\" " + conf_tomcat
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xms value for the Tomcat with error [" + str(err) + "]")
    cmd = "sed -i \"s/-Xmx.*M/-Xmx" + str(xmx_size_in_mb) + "M/\" " + conf_tomcat
    logger.info("Configuring the Tomcat Xmx memory [" + str(xmx_size_in_mb) + "MB]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xmx value for the Tomcat with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("cyops-tomcat")

def configure_php(acpu_memory_size_in_mb=512):
    config_php = get_config(RESOURCE_PHP)
    if config_php is not None:
        acpu_memory_size_in_mb = int(config_php['acpu'])
    conf_php = "/etc/php.d/40-apcu.ini"
    logger.info("Configuring the PHP APCu memory [" + str(acpu_memory_size_in_mb) + "MB]")
    cmd = "sed -i \"s/apc.shm_size\s*=.*$/apc.shm_size = \\\"" + str(acpu_memory_size_in_mb) + "M\\\"/\" " + conf_php
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure the PHP APCu memory with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("php-fpm")

def configure_mq(number_of_scheduler=4):
    config_mq = get_config(RESOURCE_MQ)
    if config_mq is not None:
        number_of_scheduler = int(config_mq['scheduler'])
    conf_mq = "/etc/rabbitmq/rabbitmq-env.conf"
    config_key = "RABBITMQ_SERVER_ADDITIONAL_ERL_ARGS"
    cmd = "sed -i '/" + config_key + "/d' " + conf_mq
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to remove the existing value for the RabbitMQ scheduler")
        return
    # add empty line first. This ensure that newline character gets added, before we write any configuration.
    cmd = "echo '' >> " + conf_mq
    execute_cmd_with_shell(cmd, True)
    # RABBITMQ_SERVER_ADDITIONAL_ERL_ARGS="+S 4:4"
    cmd = "echo '" + config_key + "=\"+S " + str(number_of_scheduler) + ":" + str(number_of_scheduler) + "\"' >> " + conf_mq
    logger.info("Configuring the RabbitMQ scheduler [" +str(number_of_scheduler) + "]")
    ret, out, err = execute_cmd_with_shell(cmd, True)
    if ret != 0:
        print_error("Failed to configure the RabbitMQ scheduler with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("rabbitmq-server")

def configure_celery(autoscale=True):
    config_celery = get_config(RESOURCE_CELERY)
    if config_celery is not None:
        autoscale = config_celery['autoscale']
    if autoscale:
        enable_celery_autoscale()
        return
    disable_celery_autoscale()
    global services_to_restart
    services_to_restart.append("celeryd")

def validate_resource(required_cpu=8, required_ram=32, ram_percent_error_margin=10.00):
    config_hardware = get_config(RESOURCE_HARDWARE)
    if config_hardware is not None:
        required_cpu = config_hardware['cpu']
        required_ram = config_hardware['ram']
        ram_percent_error_margin = config_hardware['ram_percent_error_margin']
    current_cpu = get_cpu_core_count()
    current_ram = get_ram()
    logger.info("Available CPU core [" + str(current_cpu) + "]")
    logger.info("Available RAM [" + str(current_ram) + "GB]")
    error = ""
    if current_cpu < required_cpu:
        error = error + "Ensure you have [" + str(required_cpu) + "] CPU core" + "\n"
    ram_percent_diff = get_percent_diff(required_ram, current_ram)
    # This will work if RAM is 3.6GB when checked for 4GB RAM validation
    # On FortiCloud Aailable RAM [31.26GB], percent diff is 2.31
    if ram_percent_diff > ram_percent_error_margin:
        error = error + "Ensure you have [" + str(required_ram) + "GB] RAM" + "\n"
    if error != "":
        print_error(error)
        sys.exit(1)

def perform_hardening():
    global args
    logger.info("Configuring " +  str(args.mode) + " mode")
    if not args.is_force:
        validate_resource()
    else:
        logger.info("'--force' is opted, skipping resource validation")
    configure_celery()
    configure_es()
    configure_pg()
    configure_tomcat()
    configure_php()
    configure_mq()
    if not args.skip_service_restart:
        restart_services()

def run_data_reset():
    backup_files = glob.glob(DIR_DATA_RESET + '/DR_BACKUP_*.tgz')
    if (backup_files is None) or len(backup_files) == 0:
        print_error("Backup data is not found for the data reset operation. Therefore, data reset cannot be performed.")
        sys.exit(1)
    msg =       "WARNING:\n"
    msg = msg + "Resetting data causes data loss, i.e., this operation will reset data to the point\n"
    msg = msg + "where you first saw the EULA on the console. This means that all your connector\n"
    msg = msg + "configurations, created users, created playbooks, playbook execution logs, audit\n"
    msg = msg + "logs, etc. will be lost."
    response = take_choice_input(msg=msg)
    logger.debug("User entered [" + response + "]")
    if response.lower() != 'yes':
        sys.exit(0)
    cmd = "csadm db --restore " + str(backup_files[0]) + " --no-interaction --reset-data"
    ret = os.system(cmd)
    exit_code = 0
    if ret != 0:
        exit_code = 1
    sys.exit(exit_code)

def disk_management():
    s_args = ' '.join(sys.argv[2:])
    cmd_disk = "/opt/cyops/scripts/csadm/commands/disk.sh --" + s_args
    ret = os.system(cmd_disk)
    exit_code = 0
    if ret != 0:
        exit_code = 1
    sys.exit(exit_code)

def define_parser_subcmd_data_reset(parser, subparsers):
    parser_data_reset = subparsers.add_parser('data-reset')
    parser_data_reset.set_defaults(operation='data-reset')

def define_parser_subcmd_config(parser, subparsers):
    parser_config = subparsers.add_parser('config')

    parser_config.set_defaults(operation='config')
    parser_config.add_argument('--mode',
                               help='System mode',
                               choices=MODE_ALL,
                               dest='mode',
                               required=True)

    parser_config.add_argument('--skip-service-restart',
                               help=argparse.SUPPRESS,
                               dest='skip_service_restart',
                               action='store_true',
                               required=False)

    parser_config.add_argument('--force',
                               help=argparse.SUPPRESS,
                               dest='is_force',
                               action='store_true',
                               required=False)

# Create argument parser for disk command
def define_parser_subcmd_disk(subparsers):
    # Help of disk action is displayed using parser_disk.print_help, if user enters just "csadm system disk"
    # Need to set it a global variable for that
    global parser_disk
    parser_disk = subparsers.add_parser('disk', help='Disk Management')
    parser_disk.set_defaults(operation='disk')

    visible_options = "{expand-lv}"
    sub_parser_disk = parser_disk.add_subparsers(metavar=visible_options)

    sub_parser_disk_expand_lv = sub_parser_disk.add_parser("expand-lv", help="Expand the logical volume.")
    sub_parser_disk_expand_lv.set_defaults(operation='expand-lv')

    # Command to get list of logical volumes
    # Expansion of swap and root partition is not supported. It should be handled manually
    cmd_get_logical_volume_list = "lvs -o lv_name --noheadings | grep -vw swap | grep -vw root | sed 's/ *//g' | tr '\n' ' ' | sed 's/.$//' "
    stream = os.popen(cmd_get_logical_volume_list)
    cmd_op_get_logical_volume_list = stream.read()
    l_logical_volumes = []
    l_logical_volumes = cmd_op_get_logical_volume_list.split(' ')
    stream.close()

    # Required parameter for expand-lv
    sub_parser_disk_expand_lv.add_argument('--logical-volume',
                                           help='Name of the logical volume to expand. Expansion of swap and root logical volumes is not supported.',
                                           choices=l_logical_volumes,
                                           dest='logical_volume',
                                           required=True)

    # --disk and --use-vg cannot be used together.
    # Either of them is required though
    group_sub_parser_disk_expand_lv = sub_parser_disk_expand_lv.add_mutually_exclusive_group(required=True)

    # Scan for new disks
    d_scsi_host="/sys/class/scsi_host"
    for s_host in os.listdir(d_scsi_host):
        f_host_full_path=d_scsi_host + "/" + s_host
        s_cmd="echo '- - -' >>" + f_host_full_path + "/scan"
        os.system(s_cmd)

    # Get the list of disks
    cmd_get_disk_list = "lsblk -o KNAME,TYPE | grep disk | awk '{print $1}' | tr '\n' ' ' | sed 's/.$//'"
    stream = os.popen(cmd_get_disk_list)
    cmd_op_get_disk_list = stream.read()
    l_disks = []
    l_disks = cmd_op_get_disk_list.split(' ')
    stream.close()

    # Either of single disk name or "all" are possible choices of this parameter
    group_sub_parser_disk_expand_lv.add_argument('--disk',
                                           help="Disk name to expand the logical volume.",
                                           choices=l_disks,
                                           dest='use_disk',
                                           required=False)

    # --use-vg is a flag
    group_sub_parser_disk_expand_lv.add_argument('--use-vg',
                                           help="Consume the available free disk space from the volume group this logical volume is part of.",
                                           dest='use_vg',
                                           action='store_true',
                                           required=False)

    # If --size is not mentioned, use 100% of either disk or vg
    sub_parser_disk_expand_lv.add_argument('--size',
                                           help="Size in gigabytes (GB), to be consumed from the disk or volume group this logical volume is part of. Omit this option for 100 percent expansion from the disk or the volume group.",
                                           type=check_positive_integer,
                                           required=False)
                                        
    sub_parser_disk_expand_lv.add_argument('--validate',
                                           help="Validates the inputs passed for csadm system disk expand-lv command.",
                                           dest='validate',
                                           action='store_true',
                                           required=False)

def define_parser_subcmd_fortimonitor(parser, subparsers):
    fsr.fortimonitor.handlers.cli.define_parser(parser, subparsers)
def define_parser(parser, subparsers):
    define_parser_subcmd_fortimonitor(parser, subparsers)

    # Disk option is not applicable for docker instance
    if not is_docker():
        define_parser_subcmd_disk(subparsers)

    if is_enterprise_instance():
        define_parser_subcmd_config(parser, subparsers)
    if is_forticloud_instance():
        define_parser_subcmd_data_reset(parser, subparsers)

# TBD: to display example commands for disk

def start():
    parser = argparse.ArgumentParser(prog="csadm system", formatter_class=CustomHelpFormatter)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'

    # Disk option is not applicable for docker instance
    if is_docker():
        visible_options = "{fortimonitor}"
        subparsers = parser.add_subparsers(metavar=visible_options)
    else:
        visible_options = "{fortimonitor, disk}"
        subparsers = parser.add_subparsers(metavar=visible_options)

    define_parser(parser, subparsers)

    global args
    args = parser.parse_args()

    if not 'operation' in dir(args):
        parser.print_help(sys.stderr)
        sys.exit(1)
    if str(args.operation) == "config":
        populate_config()
        perform_hardening()
        return
    elif str(args.operation) == "disk":
        parser_disk.print_help(sys.stderr)
        sys.exit(1)
    elif str(args.operation) == "expand-lv":
        disk_management()
        return
    elif str(args.operation) == "data-reset":
        run_data_reset()
        return
    elif str(args.operation).startswith("forti-monitor"):
        fsr.fortimonitor.handlers.cli.handle_cli_ops(args)
        return

    # Execution should not come to this stage. If comes, then print help
    parser.print_help(sys.stderr)

if __name__ == '__main__':
    start()
sys.exit(0)
