""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import sys, json, os, socket, signal, ssl, copy
import OpenSSL.crypto as crypto
from collections import OrderedDict
sys.path.append("/opt/cyops/scripts/")
from csadm.consts import *
from csadm.message_constants import *

from fsr.commonlibs.utils import is_docker
from csadm.logs.logger_defs import get_logger
from csadm.utils import *

logger = get_logger(__name__, RSYSLOG_LOG_FILE)

set_logger_for_utils(logger)

def signal_handler(signal, frame):
    sys.exit(1)

def get_license_details():
    cmd =       "/opt/cyops/scripts/api_caller.py "
    cmd = cmd + " --endpoint \"https://localhost/api/auth/license/?param=license_details&node_id=SELF\" "
    cmd = cmd + " --method get "
    cmd = cmd + " --silent "
    serial_number = "License not found on the system"
    vd = "License not found on the system"
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        return (serial_number, vd)
    license_details = None
    try:
        license_details = json.loads(out)
    except Exception as e:
        logger.error(str(e))
        return (serial_number, vd)

    nodes_details = license_details['nodes']
    if nodes_details:
        current_node = list(nodes_details.values())[0]
        current_node_details = current_node['details']
        serial_number = current_node_details['serial_no']
        vd = "enterprise"
        if current_node_details['is_distributed']:
            vd = current_node_details['role']
    return (serial_number, vd)

def notify_gateway_service(server_details):
    cmd_source = ""
    if CMD_SOURCE_IN_SDJ in server_details.keys():
        cmd_source = server_details[CMD_SOURCE_IN_SDJ]
    elif INPUT_ARGS_CMD_SOURCE in server_details.keys():
        cmd_source = server_details[INPUT_ARGS_CMD_SOURCE]

    if cmd_source == CMD_SOURCE_UI:
        # If configuration is done via UI, then tomcat service restart is NOT required.
        return
    logger.info("Notifying gateway service of syslog config change")
    cmd = "/opt/cyops/scripts/api_caller.py --method POST --endpoint https://localhost/api/gateway/config/syslog/reconcile --payload '{}'"
    execute_cmd(cmd, True)

def is_port_opened_in_selinux(protocol, port):
    format_cmd= "semanage port -l | grep -w {0} | grep {1}"
    cmd = format_cmd.format(port, protocol)
    ret, out, err = execute_cmd_with_shell(cmd, ignore_failure=True)
    if ret == 0:
        return True
    return False

def open_selinux_port(protocol, port):
    if protocol == PROTOCOL_RELP:
        # For selinux, there is no protocol called relp. It is tcp for relp.
        protocol = PROTOCOL_TCP
    format_cmd_port_add = "semanage port -a -t syslogd_port_t -p {0} {1}"
    format_cmd_port_modify = "semanage port -m -t syslogd_port_t -p {0} {1}"
    cmd = format_cmd_port_add.format(protocol, port)
    if is_port_opened_in_selinux(protocol, port):
        cmd = format_cmd_port_modify.format(protocol, port)
    logger.info(mconsts_15.format(str(port), protocol))
    execute_cmd_with_shell(cmd)

def install_selinux_policy_package(module_path):
    # -i is for persist across reboot
    cmd = "semodule -i " + module_path
    logger.info(mconsts_26.format(module_path))
    #
    # Do not fail the operation if it fails to install the selinux policy package.
    # If installation fails, then it will result to not forward the contents of
    # /var/log/cyops/cyops-api/last_system_publish.log.
    #
    execute_cmd(cmd, True)

def check_if_already_added(server_to_add, all_servers):
    for server in all_servers:
        if server[SERVER_IP_IN_SDJ] == server_to_add[SERVER_IP_IN_SDJ] and \
            server[SERVER_PORT_IN_SDJ] == server_to_add[SERVER_PORT_IN_SDJ]:
            print_error(mconsts_5)
            return True
    return False

def copy_tls_certs(server_details):
    uuid = server_details[UUID_IN_SDJ]
    ca_cert = server_details[CA_CERT_IN_SDJ]
    client_cert = server_details[CLIENT_CERT_IN_SDJ]
    client_key = server_details[CLIENT_KEY_IN_SDJ]
    target_dir = RSYSLOG_CERT_DIR + "/" + uuid
    cmd = "mkdir -p " + target_dir
    execute_cmd(cmd)

    ca_cert_path_target_dir = target_dir + "/" + CA_CERT_PEM
    client_cert_path_target_dir = None
    client_key_path_target_dir = None

    # Copy CA cert file
    cmd = "cp " + ca_cert + " " + ca_cert_path_target_dir
    if ca_cert != ca_cert_path_target_dir:
        # In case of update-config, there could be possibility that no change in cert.
        # So, we should not copy same cert file again to same location.
        execute_cmd(cmd)

    if client_cert is not None:
        #Copy client cert file
        client_cert_path_target_dir = target_dir + "/" + CLIENT_CERT_PEM
        cmd = "cp " + client_cert + " " + client_cert_path_target_dir
        if client_cert != client_cert_path_target_dir:
            execute_cmd(cmd)
        #Copy client key file
        client_key_path_target_dir = target_dir + "/" + CLIENT_KEY_PEM
        cmd = "cp " + client_key + " " + client_key_path_target_dir
        if client_key != client_key_path_target_dir:
            execute_cmd(cmd)

    # Now we have copied the input TLS certs to the target dir. Here after, the code
    # should refer the paths from target dir
    server_details[TLS_CERT_DIR_IN_SDJ] = target_dir
    server_details[CA_CERT_IN_SDJ] = ca_cert_path_target_dir
    server_details[CLIENT_CERT_IN_SDJ] = client_cert_path_target_dir
    server_details[CLIENT_KEY_IN_SDJ] = client_key_path_target_dir

def prepare_server_details_dict(input_args, is_update_config_operation=False):
    server_ip = input_args[INPUT_ARGS_SERVER_IP]
    port = input_args[INPUT_ARGS_SERVER_PORT]
    protocol = input_args[INPUT_ARGS_SERVER_PROTOCOL]
    tls = None
    if input_args[INPUT_ARGS_SERVER_TLS] is not None:
        tls = False
        if input_args[INPUT_ARGS_SERVER_TLS].lower() in TLS_YES_CHOICES:
            tls = True

    ca_cert = input_args[INPUT_ARGS_CA_CERT]
    client_cert = input_args[INPUT_ARGS_CLIENT_CERT]
    client_key = input_args[INPUT_ARGS_CLIENT_KEY]
    _filter = input_args[INPUT_ARGS_FILTER]
    config_name = input_args[INPUT_ARGS_CONFIG_NAME]
    cmd_source = input_args[INPUT_ARGS_CMD_SOURCE]

    server_details = OrderedDict()
    uuid = input_args[INPUT_ARGS_UUID]
    if uuid is None:
        uuid = get_uuid()
    server_details[UUID_IN_SDJ] = uuid
    if is_update_config_operation:
        server_details[UUID_IN_SDJ] = input_args[INPUT_ARGS_UUID]

    server_details[CONFIG_NAME_IN_SDJ] = config_name
    server_details[SERVER_IP_IN_SDJ] = server_ip
    server_details[SERVER_PORT_IN_SDJ] = port
    server_details[SERVER_PROTOCOL_IN_SDJ] = protocol
    server_details[SERVER_TLS_IN_SDJ] = tls
    server_details[CA_CERT_IN_SDJ] = ca_cert
    server_details[CLIENT_CERT_IN_SDJ] = client_cert
    server_details[CLIENT_KEY_IN_SDJ] = client_key
    server_details[FILTER_IN_SDJ] = _filter

    # Initialize the tls_cert_dir to None at start
    server_details[TLS_CERT_DIR_IN_SDJ] = None
    server_details[CMD_SOURCE_IN_SDJ] = cmd_source
    return server_details

def save_server_details(server_details, update_directly=False, is_update_config_op=False):
    all_servers = None
    if not update_directly:
        all_servers = read_json(RSYSLOG_SERVER_DETAILS_JSON)
        if all_servers is None:
            all_servers = []
        if not is_update_config_op:
            all_servers.append(server_details)
        else:
            for i in range(len(all_servers)):
                if all_servers[i][UUID_IN_SDJ] == server_details[UUID_IN_SDJ]:
                    all_servers[i] = server_details
                    break
    else:
        all_servers = []
        if (server_details is not None) and len(server_details) > 0:
            all_servers.append(server_details)

    logger.info(mconsts_14.format(RSYSLOG_SERVER_DETAILS_JSON))
    save_json(RSYSLOG_SERVER_DETAILS_JSON, all_servers)

def setup_conf_file_for_rsyslog_log_forwarding(server_details):
    #TODO : Due to UI limitation, currently we are not supporting more
    # than one syslog server to configure. Hence, we don't need to check for
    # already added servers. Hence, for now, setting already_added_servers to None.
    # already_added_servers = read_json(RSYSLOG_SERVER_DETAILS_JSON)
    already_added_servers = None

    # There could be situation where we are updating already added server itself. Hence,
    # if uuid macthes with any already added server, then do not consider it.
    if already_added_servers is not None:
        for i in range(len(already_added_servers)):
            if already_added_servers[i][UUID_IN_SDJ] == server_details[UUID_IN_SDJ]:
                del already_added_servers[i]
                break
    need_to_prepend_application_files_monitoring_conf = False
    _filter = server_details[FILTER_IN_SDJ]
    if _filter != FILTER_NONE:
        filters_list = _filter.split(",")
        if FILTER_APPLICATION in filters_list:
            need_to_prepend_application_files_monitoring_conf = True
        elif already_added_servers is not None:
            for server in already_added_servers:
                _filter = server[FILTER_IN_SDJ]
                filters_list = _filter.split(",")
                if FILTER_APPLICATION in filters_list:
                    need_to_prepend_application_files_monitoring_conf = True
                    break

    if need_to_prepend_application_files_monitoring_conf:
        cmd = "cat " + APPLICATION_FILE_LIST_TO_WATCH + " " + RSYSLOG_CONF_FILE_SRC +" " + "> " + RSYSLOG_CONF_FILE_TARGET
        execute_cmd_with_shell(cmd)
        return
    cmd = "cp " + RSYSLOG_CONF_FILE_SRC + "  " + RSYSLOG_CONF_FILE_TARGET
    execute_cmd(cmd)

def restart_rsyslog():
    serv = 'rsyslog'
    cmd = "systemctl restart " + serv
    logger.info(mconsts_18.format(serv))
    execute_cmd(cmd)

def unlink_conf_file_for_rsyslog_log_forwarding():
    cmd = "rm -f " + RSYSLOG_CONF_FILE_TARGET
    logger.info(mconsts_19.format(RSYSLOG_CONF_FILE_TARGET))
    execute_cmd(cmd)
    cmd = "rm -f " + RSYSLOG_SERVER_DETAILS_FILE
    logger.info(mconsts_19.format(RSYSLOG_SERVER_DETAILS_FILE))
    execute_cmd(cmd)

def get_filter_based_condition(server_details):
    in_filter = server_details[FILTER_IN_SDJ]
    list_of_input_filters = in_filter.split(",")
    if FILTER_NONE in list_of_input_filters:
        list_of_input_filters.remove(FILTER_NONE)

    if len(list_of_input_filters) == 0:
        return None

    deep_copy_filter_all = copy.deepcopy(FILTER_ALL)
    deep_copy_filter_all.remove(FILTER_NONE)
    if set(list_of_input_filters) == set(deep_copy_filter_all):
        return None

    final_if_condition = "if ( "
    need_or_condition = False
    for _filter in list_of_input_filters:
        if need_or_condition:
            final_if_condition = final_if_condition + " or "
        if _filter == FILTER_AUDIT:
            final_if_condition = final_if_condition + AUDIT_LOG_DEFINITION
        if _filter == FILTER_APPLICATION:
            final_if_condition = final_if_condition + APPLICATION_LOG_DEFINITION
        need_or_condition = True
    final_if_condition = final_if_condition + " ) then {"
    return final_if_condition

def prepare_udp_format(server_details):
    server_host = server_details[SERVER_IP_IN_SDJ]
    is_ipv6 = server_details[SERVER_IPV6_IN_SDJ]
    if is_ipv6 and is_valid_v6_ip(server_host):
        #IPV6 need to specify like
        #[2002:c0a8:63:0:20ea:3141:8ee1:2f69]
        server_host = "[" + server_host + "]"
    contents_for_udp = UDP_FORMAT_IN_SYSLOG.format(server_host, str(server_details[SERVER_PORT_IN_SDJ]))
    return contents_for_udp

def prepare_tcp_format(server_details):
    server_host = server_details[SERVER_IP_IN_SDJ]
    is_ipv6 = server_details[SERVER_IPV6_IN_SDJ]
    if is_ipv6 and is_valid_v6_ip(server_host):
        # IPV6 need to specify like
        # [2002:c0a8:63:0:20ea:3141:8ee1:2f69]
        server_host = "[" + server_host + "]"
    contents_for_tcp = TCP_FORMAT_IN_SYSLOG.format(server_host, str(server_details[SERVER_PORT_IN_SDJ]))
    return contents_for_tcp

def prepare_relp_format(server_details):
    server_host = server_details[SERVER_IP_IN_SDJ]
    is_ipv6 = server_details[SERVER_IPV6_IN_SDJ]
    if is_ipv6 and is_valid_v6_ip(server_host):
        # IPV6 need to specify like
        # [2002:c0a8:63:0:20ea:3141:8ee1:2f69]
        server_host = "[" + server_host + "]"
    contents_for_relp = RELP_FORMAT_IN_SYSLOG.format(server_host, str(server_details[SERVER_PORT_IN_SDJ]))
    return contents_for_relp

def prepare_tls_tcp_format(server_details):
    ca_cert_pem = server_details[CA_CERT_IN_SDJ]
    client_cert_pem = server_details[CLIENT_CERT_IN_SDJ]
    client_key_pem = server_details[CLIENT_KEY_IN_SDJ]
    server_host = server_details[SERVER_IP_IN_SDJ]
    is_ipv6 = server_details[SERVER_IPV6_IN_SDJ]
    if is_ipv6 and is_valid_v6_ip(server_host):
        # IPV6 need to specify like
        # [2002:c0a8:63:0:20ea:3141:8ee1:2f69]
        server_host = "[" + server_host + "]"
    port = server_details[SERVER_PORT_IN_SDJ]
    contents_for_tls_tcp = TLS_TCP_FORMAT_IN_SYSLOG_WITHOUT_CLIENT_CERT.format(
                            server_host,
                            str(port),
                            ca_cert_pem
                            )
    if client_cert_pem is not None:
        contents_for_tls_tcp = TLS_TCP_FORMAT_IN_SYSLOG_WITH_CLIENT_CERT.format(
            server_host,
            str(port),
            ca_cert_pem,
            client_cert_pem,
            client_key_pem
        )

    return contents_for_tls_tcp

def prepare_tls_relp_format(server_details):
    ca_cert_pem = server_details[CA_CERT_IN_SDJ]
    client_cert_pem = server_details[CLIENT_CERT_IN_SDJ]
    client_key_pem = server_details[CLIENT_KEY_IN_SDJ]
    server_ip = server_details[SERVER_IP_IN_SDJ]
    port = server_details[SERVER_PORT_IN_SDJ]
    contents_for_tls_relp = TLS_RELP_FORMAT_IN_SYSLOG_WITHOUT_CLIENT_CERT.format(
                            server_ip,
                            str(port)
                            )
    if client_cert_pem is not None:
        permitted_peer = get_san_names(server_details)
        contents_for_tls_relp = TLS_RELP_FORMAT_IN_SYSLOG_WITH_CLIENT_CERT.format(
            server_ip,
            str(port),
            ca_cert_pem,
            client_cert_pem,
            client_key_pem,
            permitted_peer
        )

    return contents_for_tls_relp

def get_config_start_comment(server_details):
    uuid = server_details[UUID_IN_SDJ]
    return CONFIG_START_COMMENT.format(uuid)

def get_forward_format(server_details):
    product_version = get_product_version()
    serial_number, vd = get_license_details()
    return FSR_FORWARD_FORMAT.format(str(product_version), str(serial_number), str(vd))

def get_config_end_comment(server_details):
    uuid = server_details[UUID_IN_SDJ]
    return CONFIG_END_COMMENT.format(uuid)

def get_rsyslog_format_string_for_server(server_details):
    # If no filter provided, then no need to forward anything.
    in_filter = server_details[FILTER_IN_SDJ]
    if in_filter == FILTER_NONE:
        # return empty string
        return ""
    contents_for_current_server = ""
    filter_based_condition = get_filter_based_condition(server_details)
    if server_details[SERVER_PROTOCOL_IN_SDJ] == PROTOCOL_UDP:
        contents_for_current_server = prepare_udp_format(server_details)

    elif server_details[SERVER_PROTOCOL_IN_SDJ] == PROTOCOL_TCP:
        # For TCP, it could be possible that tls in either yes or no
        if server_details[SERVER_TLS_IN_SDJ]:
            contents_for_current_server = prepare_tls_tcp_format(server_details)
        else:
            contents_for_current_server = prepare_tcp_format(server_details)

    elif server_details[SERVER_PROTOCOL_IN_SDJ] == PROTOCOL_RELP:
        # For RELP, it could be possible that tls in either yes or no
        if server_details[SERVER_TLS_IN_SDJ]:
            contents_for_current_server = prepare_tls_relp_format(server_details)
        else:
            contents_for_current_server = prepare_relp_format(server_details)

    if filter_based_condition is not None:
        contents_for_current_server = filter_based_condition + "\n" + contents_for_current_server + "\n}"

    forward_format = get_forward_format(server_details)
    contents_for_current_server = forward_format + "\n" + contents_for_current_server

    return contents_for_current_server


# Server details files is at /etc/rsyslog.d/server-details
def add_details_in_server_details_file(all_servers = []):

    contents_for_all_server = ""

    for server_details in all_servers:
        config_start_comment = get_config_start_comment(server_details)
        config_end_comment = get_config_end_comment(server_details)
        contents_for_current_server = get_rsyslog_format_string_for_server(server_details)
        contents_for_current_server = config_start_comment + "\n" + contents_for_current_server + "\n" + config_end_comment
        contents_for_all_server = contents_for_all_server + contents_for_current_server + "\n\n"

    logger.info(mconsts_14.format(RSYSLOG_SERVER_DETAILS_FILE))
    # TODO: Due to UI limitation, currently we are not supporting more
    # than one syslog server to configure. Hence, we don't need to check for
    # already added servers. Hence, for now, append_if_already_exists=False.
    #
    write_file(RSYSLOG_SERVER_DETAILS_FILE, contents_for_all_server, append_if_already_exists=False)

def validate_tcp_port(server_details):
    server_ip = server_details[SERVER_IP_IN_SDJ]
    port = server_details[SERVER_PORT_IN_SDJ]
    server_ip_and_port = (server_ip, int(port))
    logger.info(mconsts_21.format(server_ip, str(port)))
    err = ""
    try:
        sock = socket.socket(get_server_address_family(server_details), socket.SOCK_STREAM)
        # Timeout in 10 secs
        sock.settimeout(10)
        sock.connect(server_ip_and_port)
        sock.close()
        logger.info(mconsts_22.format(server_ip, str(port)))
        return
    except Exception as e:
        err = str(e)

    # Here came means it is failed to reach on port
    print_error(TCP_ERROR_CHECKLIST.format(err,
                                          server_ip,
                                           str(port)))
    sys.exit(1)

def get_server_address_family(server_details):
    # IPV4 default
    addr_family = socket.AF_INET
    if server_details[SERVER_IPV6_IN_SDJ]:
        addr_family = socket.AF_INET6
    return addr_family

def get_tls_args(server_details):
    ca_cert = server_details[CA_CERT_IN_SDJ]
    client_cert = server_details[CLIENT_CERT_IN_SDJ]
    client_key = server_details[CLIENT_KEY_IN_SDJ]
    tls_args = {
        'ca_certs': ca_cert,
        'cert_reqs': ssl.CERT_REQUIRED,
        'server_side': False
    }
    if client_cert is not None:
        tls_args['certfile'] = client_cert
        tls_args['keyfile'] = client_key
    return tls_args

def get_san_names(server_details):
    tls_args = get_tls_args(server_details)
    server_ip = server_details[SERVER_IP_IN_SDJ]
    port = server_details[SERVER_PORT_IN_SDJ]
    server_ip_and_port = (server_ip, int(port))
    try:
        sock = socket.socket(get_server_address_family(server_details), socket.SOCK_STREAM)
        # Timeout in 10 secs
        sock.settimeout(10)
        wrapped_socket = ssl.wrap_socket(sock, **tls_args)
        wrapped_socket.connect(server_ip_and_port)
        cert_bin = wrapped_socket.getpeercert(True)
        wrapped_socket.close()
        x509 = crypto.load_certificate(crypto.FILETYPE_ASN1, cert_bin)
        common_name = x509.get_subject().CN
        count = 0
        san_names = None
        while count < x509.get_extension_count():
            if x509.get_extension(count).get_short_name().decode("utf-8") == "subjectAltName":
                # san_names will come like below.
                # DNS:fsrprimary.fortinet.com, DNS:localhost, IP Address:127.0.0.1
                san_names = str(x509.get_extension(count))
                break
            count = count + 1
        final_str_san_names = ""
        if (san_names is not None) and (san_names != ""):
            list_of_san_names = san_names.split(",")
            for san_name in list_of_san_names:
                if san_name.strip().startswith("DNS:"):
                    final_str_san_names = final_str_san_names + '"' + san_name.split(":")[1].strip() + '",'
        final_str_san_names = final_str_san_names + '"' + common_name + '"'
        return final_str_san_names
    except Exception as e:
        print_error(str(e))
        sys.exit(1)

def validate_tls_certs(server_details):
    server_ip = server_details[SERVER_IP_IN_SDJ]
    port = server_details[SERVER_PORT_IN_SDJ]
    ca_cert = server_details[CA_CERT_IN_SDJ]
    server_ip_and_port = (server_ip, int(port))
    tls_args = get_tls_args(server_details)

    logger.info(mconsts_11)
    try:
        sock = socket.socket(get_server_address_family(server_details), socket.SOCK_STREAM)
        # Timeout in 10 secs
        sock.settimeout(10)
        wrapped_socket = ssl.wrap_socket(sock, **tls_args)
        wrapped_socket.connect(server_ip_and_port)
        wrapped_socket.close()
        logger.info(mconsts_27)
    except Exception as e:
        err = str(e)
        print_error(TLS_ERROR_CHECKLIST.format(err,
                                      server_ip,
                                      str(port),
                                      ca_cert))
        sys.exit(1)

def add_rsyslog_server(input_args, convert_input_args_to_sdj_format=True, interactive=True, check_if_added=True):
    server_details = input_args
    if INPUT_ARGS_NO_INTERACTION in server_details.keys():
        interactive = input_args[INPUT_ARGS_NO_INTERACTION]
    if convert_input_args_to_sdj_format:
        server_details = prepare_server_details_dict(input_args)
    all_servers = read_json(RSYSLOG_SERVER_DETAILS_JSON)
    if all_servers is None:
        all_servers = []

    server_details[SERVER_IPV6_IN_SDJ] = False
    if is_ipv6_server(server_details[SERVER_IP_IN_SDJ]):
        server_details[SERVER_IPV6_IN_SDJ] = True

    if check_if_added:
        if check_if_already_added(server_details, all_servers):
            # Here, the exit 1 is required as when this CLI gets called from UI,
            # it should show appropriate error msg.
            sys.exit(1)

    # TODO: Due to UI limitation, currently we are not supporting more
    # than one syslog server to configure. So now, keep the uuid same as of old.
    # this is required as filters for old configuration in UI is not removed.
    # Hence instead of adding new filter, keep the uuid same, so that UI filters
    # will remain same.
    #
    uuid_of_already_configured_server = None

    if len(all_servers) > 0:
        # If --no-interaction is used, then we want to go ahead and add
        # new configuration by removing old configuration, if there any.
        selected_choice = "yes"
        if interactive:
            selected_choice = take_choice_input(mconsts_32)

        if selected_choice.lower() == "no":
            if not interactive:
                print_error(mconsts_34)
            sys.exit(1)
        # Copy the old UUID to new configuration
        uuid_of_already_configured_server = all_servers[0][UUID_IN_SDJ]
        server_details[UUID_IN_SDJ] = uuid_of_already_configured_server

    handle_pre_checks_for_tcp_and_copy_certs(server_details)

    if not is_docker():
        open_selinux_port(server_details[SERVER_PROTOCOL_IN_SDJ],server_details[SERVER_PORT_IN_SDJ])
        if not os.path.exists(CUSTOM_SELINUX_POLICY_PACKAGE_FLAG):
            install_selinux_policy_package(CUSTOM_SELINUX_POLICY_PACKAGE)
            cmd = "touch " + CUSTOM_SELINUX_POLICY_PACKAGE_FLAG
            execute_cmd(cmd, True)
    setup_conf_file_for_rsyslog_log_forwarding(server_details)
    add_details_in_server_details_file([server_details])
    # On RockyLinux, the syslog socket is not getting created automatically.
    # Need to remove the comment ;Requires=syslog.socket from rsyslog.service file
    logger.debug("Uncommenting syslog.socket from rsyslog.service")
    cmd = "sed -i 's/;Requires=syslog.socket/Requires=syslog.socket/g' /usr/lib/systemd/system/rsyslog.service"
    execute_cmd_with_shell(cmd, True)
    cmd = 'systemctl daemon-reload'
    execute_cmd_with_shell(cmd, True)
    cmd = "sed -i 's#SysSock.Use *= *\"off\"#SysSock.Use=\"on\"#g' /etc/rsyslog.conf"
    execute_cmd_with_shell(cmd, True)
    restart_rsyslog()
    #TODO: Due to UI limitation, currently we are not supporting more
    # than one syslog server to configure. Hence, we don't need to check for
    # already added servers. Hence, for now, update_directly=True.
    #
    save_server_details(server_details, update_directly=True)
    notify_gateway_service(server_details)
    logger.info(mconsts_12)

def handle_pre_checks_for_tcp_and_copy_certs(server_details):

    if server_details[CA_CERT_IN_SDJ] is not None:
        validate_tls_certs(server_details)

    elif(server_details[SERVER_PROTOCOL_IN_SDJ] == PROTOCOL_TCP) or \
         (server_details[SERVER_PROTOCOL_IN_SDJ] == PROTOCOL_RELP):
        # RELP also listen over TCP
        validate_tcp_port(server_details)

    if server_details[CA_CERT_IN_SDJ] is not None:
        copy_tls_certs(server_details)

def update_rsyslog_server(server_details, input_args):

    server_details[SERVER_IPV6_IN_SDJ] = False
    if is_ipv6_server(server_details[SERVER_IP_IN_SDJ]):
        server_details[SERVER_IPV6_IN_SDJ] = True

    handle_pre_checks_for_tcp_and_copy_certs(server_details)

    if not is_docker():
        open_selinux_port(server_details[SERVER_PROTOCOL_IN_SDJ],server_details[SERVER_PORT_IN_SDJ])
        if not os.path.exists(CUSTOM_SELINUX_POLICY_PACKAGE_FLAG):
            install_selinux_policy_package(CUSTOM_SELINUX_POLICY_PACKAGE)
            cmd = "touch " + CUSTOM_SELINUX_POLICY_PACKAGE_FLAG
            execute_cmd(cmd, True)
    setup_conf_file_for_rsyslog_log_forwarding(server_details)
    update_details_in_server_details_file(server_details)
    restart_rsyslog()
    save_server_details(server_details, is_update_config_op=True)
    notify_gateway_service(server_details)
    logger.info(mconsts_12)

# Server details files is at /etc/rsyslog.d/server-details
def remove_details_from_server_details_file(server_to_remove):

    server_details_file_contents = []
    if not os.path.exists(RSYSLOG_SERVER_DETAILS_FILE):
        return

    with open(RSYSLOG_SERVER_DETAILS_FILE) as file_server_details:
        server_details_file_contents = file_server_details.readlines()

    start_line = get_config_start_comment(server_to_remove)
    end_line = get_config_end_comment(server_to_remove)

    flag = 1
    server_details_file_contents_after_removing_server = []

    for line in server_details_file_contents:
        line = line.strip()
        if line.startswith(start_line):
            flag = 0
        if flag:
            server_details_file_contents_after_removing_server.append(line)
        if line.startswith(end_line):
            flag = 1

    logger.info(mconsts_14.format(RSYSLOG_SERVER_DETAILS_FILE))

    with open(RSYSLOG_SERVER_DETAILS_FILE, 'w') as file_server_details:
        for line in server_details_file_contents_after_removing_server:
            file_server_details.write("%s\n" % line)

# Server details files is at /etc/rsyslog.d/server-details
def update_details_in_server_details_file(server_details):
    if not os.path.exists(RSYSLOG_SERVER_DETAILS_FILE):
        return

    contents_for_current_server = get_rsyslog_format_string_for_server(server_details)
    server_details_file_contents = []

    with open(RSYSLOG_SERVER_DETAILS_FILE) as file_server_details:
        server_details_file_contents = file_server_details.readlines()

    start_line = get_config_start_comment(server_details)
    end_line =  get_config_end_comment(server_details)

    flag = 1
    server_details_file_contents_after_updating_server = []

    for line in server_details_file_contents:
        line = line.strip()
        if line.startswith(start_line):
            flag = 0
        if flag:
            server_details_file_contents_after_updating_server.append(line)
        if line.startswith(end_line):
            server_details_file_contents_after_updating_server.append(start_line)
            for content_line in contents_for_current_server.splitlines():
                server_details_file_contents_after_updating_server.append(content_line)
            server_details_file_contents_after_updating_server.append(end_line)
            flag = 1

    logger.info(mconsts_14.format(RSYSLOG_SERVER_DETAILS_FILE))

    with open(RSYSLOG_SERVER_DETAILS_FILE, 'w') as file_server_details:
        for line in server_details_file_contents_after_updating_server:
            file_server_details.write("%s\n" % line)

def remove_rsyslog_server(input_args):
    all_servers = read_json(RSYSLOG_SERVER_DETAILS_JSON)
    if all_servers is None:
        logger.info(mconsts_16)
        sys.exit(0)

    uuid = input_args[INPUT_ARGS_UUID]
    server_configured = False
    server_to_remove = None
    for i in range(len(all_servers)):
        if all_servers[i][UUID_IN_SDJ] == uuid:
            server_to_remove = all_servers[i]
            del all_servers[i]
            server_configured = True
            break

    if not server_configured:
        print_error(mconsts_20.format(uuid))
        sys.exit(1)

    if len(all_servers) != 0:
        remove_details_from_server_details_file(server_to_remove)
    else:
        unlink_conf_file_for_rsyslog_log_forwarding()
    restart_rsyslog()
    save_server_details(all_servers, update_directly=True)
    notify_gateway_service(input_args)
    logger.info(mconsts_17)

def show_rsyslog_configuration(input_args, silent=False):
    print_in_json = input_args[INPUT_ARGS_JSON_FLAG]
    uuid = input_args[INPUT_ARGS_UUID]
    server_details = read_json(RSYSLOG_SERVER_DETAILS_JSON)
    if (server_details is None) or len(server_details) == 0:
        if print_in_json:
            if silent:
                return []
            print("[]")
        else:
            print(mconsts_2)
        sys.exit(0)

    if uuid is not None:
        server_details_matching_with_input_uuld = None
        for server in server_details:
            if server[UUID_IN_SDJ] == uuid:
                server_details_matching_with_input_uuld = server
                break
        if server_details_matching_with_input_uuld is None:
            print_error(mconsts_20.format(uuid))
            sys.exit(1)
        server_details =[server_details_matching_with_input_uuld]

    if print_in_json:
        for server in server_details:
            if server[CA_CERT_IN_SDJ] is not None:
                server[CA_CERT_IN_SDJ] = read_file(server[CA_CERT_IN_SDJ])
            if server[CLIENT_CERT_IN_SDJ] is not None:
                server[CLIENT_CERT_IN_SDJ] = read_file(server[CLIENT_CERT_IN_SDJ])
            if server[CLIENT_KEY_IN_SDJ] is not None:
                server[CLIENT_KEY_IN_SDJ] = read_file(server[CLIENT_KEY_IN_SDJ])
            if (server[FILTER_IN_SDJ] is not None) and (server[FILTER_IN_SDJ] != FILTER_NONE):
                # We have other filters also in audit log, hence filter shows conflicts. This is while API
                # for UI.
                server['syslog_filters'] = server[FILTER_IN_SDJ].split(",")
            else:
                server['syslog_filters'] = []

        if silent:
            return server_details
        print(json.dumps(server_details))
        sys.exit(0)

    for server in server_details:
        for col_to_hide in LIST_OF_COL_TO_HIDE_FROM_SDJ_DURING_SHOW_CONFIG:
            del server[col_to_hide]
    print_dataset(server_details)

def convert_to_absolute_paths(input_args, in_args_to_convert = [INPUT_ARGS_CA_CERT, INPUT_ARGS_CLIENT_CERT, INPUT_ARGS_CLIENT_KEY]):
    for in_arg in in_args_to_convert:
        if (input_args[in_arg] is not None) and input_args[in_arg] != "":
            #Convert to absolute path first before proceeding
            input_args[in_arg] = os.path.abspath(input_args[in_arg])

def check_file_paths(input_args, in_args_to_check = [INPUT_ARGS_CA_CERT, INPUT_ARGS_CLIENT_CERT, INPUT_ARGS_CLIENT_KEY]):
    err = ""
    at_least_one_missing = False
    for in_arg in in_args_to_check:
        if (input_args[in_arg] is not None) and input_args[in_arg] != "":
            if not os.path.exists(input_args[in_arg]):
                err = err + mconsts_7.format(input_args[in_arg])
                at_least_one_missing = True
    if at_least_one_missing:
        print_error(err)
        sys.exit(1)

def validate_filters(comma_separated_filter_list):
    if comma_separated_filter_list is None:
        return
    is_invalid_filter_found = False
    list_of_input_filters = comma_separated_filter_list.split(",")
    for _filter in list_of_input_filters:
        if _filter not in FILTER_ALL:
            is_invalid_filter_found = True
            break
    if is_invalid_filter_found:
        print_error(mconsts_28.format(",".join(FILTER_ALL)))
        sys.exit(1)

def validate_add_server_input_arguments(input_args):
    if not is_valid_hostname_or_ip(input_args[INPUT_ARGS_SERVER_IP]):
        sys.exit(1)

    if input_args[INPUT_ARGS_SERVER_PROTOCOL] == PROTOCOL_UDP:
        # For UDP protocol, TLS will be NOT be used
        input_args[INPUT_ARGS_SERVER_TLS] = "false"
        input_args[INPUT_ARGS_CA_CERT] = None
        input_args[INPUT_ARGS_CLIENT_CERT] = None
        input_args[INPUT_ARGS_CLIENT_KEY] = None

    validate_filters(input_args[INPUT_ARGS_FILTER])
    check_file_paths(input_args)

def validate_update_server_input_arguments(server_details):
    if not is_valid_hostname_or_ip(server_details[SERVER_IP_IN_SDJ]):
        sys.exit(1)
    if (server_details[SERVER_PROTOCOL_IN_SDJ] == PROTOCOL_UDP) or (not server_details[SERVER_TLS_IN_SDJ]):
        # For UDP protocol, TLS will be NOT be used
        server_details[SERVER_TLS_IN_SDJ] = False
        server_details[CA_CERT_IN_SDJ] = None
        server_details[CLIENT_CERT_IN_SDJ] = None
        server_details[CLIENT_KEY_IN_SDJ] = None
        server_details[TLS_CERT_DIR_IN_SDJ] = None

    validate_filters(server_details[FILTER_IN_SDJ])
    files_to_check = [CA_CERT_IN_SDJ, CLIENT_CERT_IN_SDJ, CLIENT_KEY_IN_SDJ]
    check_file_paths(server_details, files_to_check)

def get_config_details(uuid):
    all_servers = read_json(RSYSLOG_SERVER_DETAILS_JSON)
    if all_servers is None:
        return None
    details_to_return = None
    for server in all_servers:
        if server[UUID_IN_SDJ] == uuid:
            details_to_return = server
            break
    return details_to_return

def get_rsyslog_configuration():
    input_args_for_show_config = {
        INPUT_ARGS_JSON_FLAG: True,
        INPUT_ARGS_UUID: None
    }
    server_details = show_rsyslog_configuration(input_args_for_show_config, silent=True)
    return server_details

def export_config_json(in_export_path):
    server_details = get_rsyslog_configuration()
    if len(server_details) == 0:
        print(mconsts_2)
        return
    export_path = os.path.abspath(os.getcwd())
    if in_export_path is not None:
        if in_export_path != "dummy_value":
            export_path = os.path.abspath(in_export_path)
    if not os.path.isdir(export_path):
        print_error(mconsts_6.format(export_path))
        sys.exit(1)
    export_file_name = EXPORT_CONFIG_JSON_FILE_NAME
    export_path_including_file_name = export_path + "/" + export_file_name
    logger.info(mconsts_31.format(export_path_including_file_name))
    save_json(export_path_including_file_name, server_details)

def export_config_base64():
    server_details = get_rsyslog_configuration()
    if len(server_details) == 0:
        print(mconsts_2)
        return
    print(base64_encode(server_details))

def export_config(input_args):
    export_path = input_args[INPUT_ARGS_EXPORT_PATH]
    export_format = input_args[INPUT_ARGS_EXPORT_FORMAT]
    if export_format == EXPORT_FORMAT_JSON:
        export_config_json(export_path)
    elif export_format == EXPORT_FORMAT_BASE64:
        export_config_base64()

def refresh_config(input_args):
    import_config_input_args = {}
    if not os.path.isfile(RSYSLOG_SERVER_DETAILS_JSON):
        return
    import_config_input_args[INPUT_ARGS_PAYLOAD] = RSYSLOG_SERVER_DETAILS_JSON
    # non interactive way
    import_config_input_args[INPUT_ARGS_NO_INTERACTION] = False
    import_config(import_config_input_args, check_if_already_added=False)

def import_config(input_args, check_if_already_added=True):
    payload = input_args[INPUT_ARGS_PAYLOAD]
    # payload can be path to json file or can be base64 encoded string.
    in_json_path = None
    payload_format = None
    server_details = None
    if os.path.isfile(payload):
        in_json_path = payload
        payload_format = EXPORT_FORMAT_JSON
    else:
        server_details = base64_decode(payload)

    if payload_format == EXPORT_FORMAT_JSON:
        server_details = read_json(in_json_path)

    if server_details is None:
        print_error(mconsts_30)
        sys.exit(1)

    for server_detail in server_details:
        if 'syslog_filters' in server_detail.keys():
            # We add 'syslog_filters' key for UI. Not required. If we do not delete,
            # then it will be visible in show-config output.
            del server_detail['syslog_filters']

        uuid = server_detail[UUID_IN_SDJ]
        tmp_ca_cert = "/tmp/" + uuid + "-ca-cert.pem"
        tmp_client_cert = "/tmp/" + uuid + "-client-cert.pem"
        tmp_client_key = "/tmp/" + uuid + "-client-key.pem"
        if server_detail[CA_CERT_IN_SDJ] is not None:
            write_file(tmp_ca_cert, server_detail[CA_CERT_IN_SDJ])
            server_detail[CA_CERT_IN_SDJ] = tmp_ca_cert

        if server_detail[CLIENT_CERT_IN_SDJ] is not None:
            write_file(tmp_client_cert, server_detail[CLIENT_CERT_IN_SDJ])
            server_detail[CLIENT_CERT_IN_SDJ] = tmp_client_cert

        if server_detail[CLIENT_KEY_IN_SDJ] is not None:
            write_file(tmp_client_key, server_detail[CLIENT_KEY_IN_SDJ])
            server_detail[CLIENT_KEY_IN_SDJ] = tmp_client_key

        add_rsyslog_server(server_detail, convert_input_args_to_sdj_format=False, interactive=input_args[INPUT_ARGS_NO_INTERACTION], check_if_added=check_if_already_added)
        cmd = "rm -f " + tmp_ca_cert + " " + tmp_client_cert + " " + tmp_client_key
        execute_cmd(cmd, True)

def handle_rsyslog_operations(input_args):
    # If user press ctrl + c during execution, then exit the script without showing stack trace.
    signal.signal(signal.SIGINT, signal_handler)

    if INPUT_ARGS_CMD_SOURCE in input_args.keys():
        if input_args[INPUT_ARGS_CMD_SOURCE] == CMD_SOURCE_UI:
            unset_logger_for_utils()
            disable_fancy_output()

    if str(input_args['operation']) == 'rsyslog-add-server':
        validate_add_server_input_arguments(input_args)
        convert_to_absolute_paths(input_args)
        add_rsyslog_server(input_args)
        return

    elif str(input_args['operation']) == 'rsyslog-show-config':
        show_rsyslog_configuration(input_args)
        return

    elif str(input_args['operation']) == 'rsyslog-remove-server':
        remove_rsyslog_server(input_args)
        return

    elif str(input_args['operation']) == 'rsyslog-update-server':
        input_uuid = input_args[INPUT_ARGS_UUID]
        configured_details = get_config_details(input_uuid)
        if configured_details is None:
            print_error(mconsts_29.format(input_uuid))
            sys.exit(1)
        server_details = prepare_server_details_dict(input_args, is_update_config_operation=True)

        for config_param in server_details.keys():
            if server_details[config_param]:
                # Input parameter is provided. No need to go further in loop.
                continue
            #
            # boolean parameter needs special handling. For eg. --tls.
            # --tls need special handling. Here, --tls is boolean parameter.
            # if --tls set to false during update-config,
            # then above 'if' condition will fail and call will come here.
            # Hence, --tls or any other boolean parameter should be not be considered
            # from configured_details.
            # Now, assume if user wanted to disable the previous configured TLS
            # and just want to use TCP,
            # if we do not check for boolean parameter, then it will be used
            # from configured_details which user don't want.
            #
            if type(server_details[config_param]) != bool:
                server_details[config_param] = configured_details[config_param]

        validate_update_server_input_arguments(server_details)
        path_to_convert_to_absolute = [CA_CERT_IN_SDJ, CLIENT_CERT_IN_SDJ, CLIENT_KEY_IN_SDJ]
        convert_to_absolute_paths(server_details, path_to_convert_to_absolute)
        update_rsyslog_server(server_details, input_args)

    elif str(input_args['operation']) == 'rsyslog-import-config':
        import_config(input_args)

    elif str(input_args['operation']) == 'rsyslog-export-config':
        export_config(input_args)

    elif str(input_args['operation']) == 'rsyslog-refresh-config':
        refresh_config(input_args)