""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import subprocess
import uuid
import shlex, sys, json, os, ipaddress, re
import yaml
sys.path.append("/opt/cyops/scripts/")

from csadm.message_constants import *
from csadm.consts import *
import tabulate
import base64

_common_utils_logger = None
_show_fancy_output = True

def set_logger_for_utils(logger):
    global _common_utils_logger
    _common_utils_logger = logger

def unset_logger_for_utils():
    global _common_utils_logger
    _common_utils_logger = None

def disable_fancy_output():
    global _show_fancy_output
    _show_fancy_output = False

def get_dash_string():
    return "------------------------------------------------"

def print_error(msg):
    err_msg = msg + "\n"

    if _show_fancy_output:
        dash_string = get_dash_string()
        err_msg = dash_string + "\n"
        err_msg = err_msg + "Error:" + "\n"
        err_msg = err_msg + msg + "\n"
        err_msg = err_msg + dash_string + "\n"

    if _common_utils_logger is not None:
        _common_utils_logger.error(err_msg)
        return
    sys.stderr.write(err_msg)

def execute_cmd(cmd, ignore_failure=False):
    cmd = shlex.split(cmd)
    pipes = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    std_out, std_err = pipes.communicate()
    if (pipes.returncode != 0) and (not ignore_failure):
        # an error happened!
        print_error(mconsts_1.format(cmd, std_err.strip()))
        sys.exit(1)
    ret = pipes.returncode
    std_out = std_out.strip().decode("utf-8")
    std_err = std_err.strip().decode("utf-8")
    return (ret, std_out, std_err)

def execute_cmd_with_shell(cmd, ignore_failure=False):
    pipes = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    std_out, std_err = pipes.communicate()
    if (pipes.returncode != 0) and (not ignore_failure):
        # an error happened!
        print_error(mconsts_1.format(cmd, std_err.strip()))
        sys.exit(1)
    ret = pipes.returncode
    std_out = std_out.strip().decode("utf-8")
    std_err = std_err.strip().decode("utf-8")
    return (ret, std_out, std_err)

def read_json(json_path):
    if not os.path.exists(json_path):
        return None
    try:
        with open(json_path, 'r') as json_contents:
            return json.load(json_contents)
    except Exception as e:
        if _common_utils_logger is not None:
            _common_utils_logger.debug(str(e))
        return None

def read_file(file):
    data = None
    try:
        with open(file, "r") as myfile:
            data = myfile.read()
    except Exception as e:
        if _common_utils_logger is not None:
            _common_utils_logger.debug(str(e))
    return data

def write_file(file_path, contents, append_if_already_exists=False):
    if contents is None:
        return
    fh = None
    file_already_exists = False
    if append_if_already_exists:
        if os.path.exists(file_path):
            fh = open(file_path, "a")
            file_already_exists = True

    if not file_already_exists:
        fh = open(file_path, "w")
    fh.write(contents)
    fh.close()

def save_json(json_path, contents):
    json_ = json.dumps(contents, indent=4)
    write_file(json_path, json_)


def print_dataset(dataset):
    header = None
    rows = None
    if isinstance(dataset, list):
        header = dataset[0].keys()
        rows = [x.values() for x in dataset]
    if isinstance(dataset, dict):
        header = dataset.keys()
        rows = [dataset.values()]
    print(tabulate.tabulate(rows, header))


def is_valid_v4_ip(addr):
    try:
        ip = ipaddress.IPv4Address(addr)
        return True
    except Exception as e:
        return False

def is_valid_v6_ip(addr):
    try:
        ipaddress.IPv6Address(addr)
        return True
    except Exception as e:
        return False

def get_getents_hosts_entry(server_ip):
    host_to_ip = "getent hosts " + server_ip
    return execute_cmd(host_to_ip, True)

def get_ip_using_getents(host):
    host_to_ip = "getent hosts " + host + " | awk '{ print $1 ; exit }'"
    return execute_cmd_with_shell(host_to_ip, True)

def is_ipv6_server(host):
    if is_valid_v6_ip(host):
        return True
    ret, out, err = get_ip_using_getents(host)
    if ret != 0:
        return False
    if is_valid_v6_ip(out):
        return True
    return False

def is_valid_hostname_or_ip(server_ip):

    if is_valid_v4_ip(server_ip):
        return True
    if is_valid_v6_ip(server_ip):
        return True

    ret, ip, err = get_getents_hosts_entry(server_ip)
    if ret == 0:
        return True
    print_error(mconsts_8.format(server_ip))
    sys.exit(1)

def get_uuid():
    return str(uuid.uuid4())

def base64_encode(payload):
    base64_payload = base64.b64encode(bytes(json.dumps(payload), 'utf-8')).decode(encoding="utf-8")
    return base64_payload

def base64_decode(payload):
    output = None
    try:
        output = base64.b64decode(payload)
        output = json.loads(output)
    except Exception as e:
        if _common_utils_logger is not None:
            _common_utils_logger.debug(str(e))
    return output

def read_input(msg):
    in_put = input(msg)
    return in_put.strip()

def print_in_dash_string(input_msg, print_using_logger=True):
    msg = input_msg + "\n"

    if  _show_fancy_output:
        dash_string = get_dash_string()
        msg = dash_string + "\n"
        msg = msg + input_msg + "\n"
        msg = msg + dash_string + "\n"

    if print_using_logger:
        if _common_utils_logger is not None:
            _common_utils_logger.info(msg)
            return

    print(msg)

def take_choice_input(msg=None, validation_msg=None, valid_choices=['yes', 'no'], default=None):
    if msg is not None:
        print_in_dash_string(msg)
    if validation_msg is None:
        validation_msg = "Do you want to continue? [yes/no]:"
    wrong_input = True
    input_ = None
    while wrong_input:
        input_ = read_input(validation_msg)
        if input_.lower() not in valid_choices:
            if default is None or input_ != "":
                print(mconsts_33)
                continue
        wrong_input = False
    if input_ == "":
        return default
    return input_

def get_product_version():
    version_file = "/opt/cyops-ui/cyops_version.json"
    if not os.path.isfile(version_file):
        return "version-not-found"
    data=json.load(open(version_file,"r"))
    return data['version'].split("-")[0]

def read_yaml(conffile):
    with open(conffile, 'r') as config_file_fp:
        local_config = yaml.load(config_file_fp, Loader=yaml.FullLoader)
    return local_config

def is_safe_path(basedir, path):
    requested_path = os.path.relpath(path, start=basedir)
    requested_path = os.path.normpath(os.path.join(basedir, requested_path))
    common_prefix = os.path.commonprefix([requested_path, basedir])
    if common_prefix is not basedir:
        return False
    return True
