""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import subprocess
import re


# =============================== Code Start ================================

def run_command_and_return_output(command=[]):
    try:
        with subprocess.Popen(command, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
            out, err = proc.communicate()
            cmd_output = {
                "output": out,
                "error": err,
                "returncode": proc.returncode
            }
            return cmd_output
    except (ValueError, FileNotFoundError) as err_val:
        pass
    return None


def check_if_PGSQL_directory_is_mounted():
    return run_command_and_return_output(["mountpoint", "-q", "/var/lib/pgsql"])["returncode"]


def get_PGDATA_LVM():
    try:
        df_command = subprocess.Popen(["df", "-h"], text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if product_type_check():
            grep_command = subprocess.Popen(["grep", "-i", "/var/lib/fsr-data"], stdin=df_command.stdout, text=True,
                                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            grep_command = subprocess.Popen(["grep", "-i", "/var/lib/pgsql"], stdin=df_command.stdout, text=True,
                                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = grep_command.communicate()
        if grep_command.returncode != 0:
            return None
        return out.split()[0]
    except (ValueError, FileNotFoundError) as err_val:
        pass
    return None


def get_cpu_details():
    data = run_command_and_return_output(["nproc"])
    return {
        "CPU": re.sub('\\n', '', data["output"]),
        "error": data["error"],
        "errorCode": data["returncode"],
    }


def get_RAM_details():
    data = run_command_and_return_output(["free", "-g"])
    return {"RAM": re.findall("[\d]+", data["output"])[0], "errorCode": data["returncode"], "error": data['error']}


def get_disk_latency():
    try:
        pg_lvm = get_PGDATA_LVM()
        if pg_lvm is None or pg_lvm.strip() == "":
            return {
                "output":"",
                "error": "could not find lvm for /var/lib/pgsql",
                "errorCode": 1
            }
        letany_data = subprocess.Popen(["sudo", "-S", "-p", "\'\'", "/bin/ioping", "-c", "20", pg_lvm.strip()], text=True,
                                       stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Here it was not accepting ["awk","\"FNR","==","25\""] so used the command directly and ''' as using " in between
        awk_command = subprocess.Popen([''' awk "FNR == 25" '''], stdin=letany_data.stdout, text=True,
                                       stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, err = awk_command.communicate()
        # For checking if correc data is selected or not
        # print(out)
        return {
            "disk_latency": out.split("/")[-2],
            "error": err,
            "errorCode": awk_command.returncode
        }

    except (ValueError, FileNotFoundError) as err_val:
        pass
    return None


def get_disk_IOPS():
    command = "sudo fio --randrepeat=1 --ioengine=libaio --direct=1 --gtod_reduce=1 --name=fsrIOPS --filename=/var/lib/pgsql/random_read_write.fio --bs=8k --iodepth=64 --size=1G --readwrite=randrw --rwmixread=50"
    data = run_command_and_return_output(command.split())
    read_IOPS = re.findall('read: IOPS=[0-9\.a-zA-Z,]+', data["output"])[0]
    # print(read_IOPS)
    read_IOPS = re.sub("(read: IOPS=|,)", "", read_IOPS)
    write_IOPS = re.findall('write: IOPS=[0-9\.a-zA-Z,]+', data["output"])[0]
    # print(write_IOPS)
    write_IOPS = re.sub("(write: IOPS=|,)", "", write_IOPS)
    return {"read": read_IOPS, "write": write_IOPS, "error": data["error"], "errorCode": data["returncode"]}


def get_shmall_and_shmmax():
    page_size = run_command_and_return_output(["getconf", "PAGE_SIZE"])
    phys_size = run_command_and_return_output(["getconf", "_PHYS_PAGES"])
    shmall = round(int(re.sub('\n', "", phys_size["output"])) / 2)
    shmmax = shmall * round(int(re.sub('\n', "", page_size["output"])))
    return {
        "shmall": shmall,
        "shmmax": shmmax,
        "error": {
            "page_size": page_size["error"],
            "phys_size": phys_size["error"]
        },
        "errorCode": 0 if page_size["returncode"] == 0 and phys_size["returncode"] == 0 else 1
    }


def get_system_config():
    if check_if_commands_are_installed() == 1:
        return {
            "error": "check if following commands are installed or not \n1) mountpoint\n2) df\n3) nproc\n4) free\n5) ioping\n6) fio\n"
        }

    if check_if_PGSQL_directory_is_mounted() == 0:
        cpu = get_cpu_details()
        ram = get_RAM_details()
        disk_latency = get_disk_latency()
        iops = get_disk_IOPS()
        shmall_and_shmmax = get_shmall_and_shmmax()
        if iops["errorCode"] != 0 or cpu["errorCode"] != 0 or ram["errorCode"] != 0 or disk_latency["errorCode"] != 0 or shmall_and_shmmax["errorCode"] != 0:
            return {
                "error": "command failed",
                "error_log": {
                    "cpu": cpu["error"],
                    "ram": ram["error"],
                    "iops": iops["error"],
                    "disk_latency": disk_latency["error"],
                    "shmall_and_shmmax": shmall_and_shmmax["error"]
                }
            }
        return {
            "CPU": cpu["CPU"],
            "RAM": ram["RAM"],
            "Disk IOPS - Read": iops["read"],
            "Disk IOPS - Write": iops["write"],
            "Disk latency": disk_latency["disk_latency"],
            "SHMALL": str(shmall_and_shmmax["shmall"]),
            "SHMMAX": str(shmall_and_shmmax["shmmax"])
        }
    return {
        "error": "No PGSQL directory is mounted"
    }


def check_if_commands_are_installed():
    mountpoint = run_command_and_return_output(["mountpoint", "--help"])
    df = run_command_and_return_output(["df", "--help"])
    nproc = run_command_and_return_output(["nproc", "--help"])
    free = run_command_and_return_output(["free", "--help"])
    ioping = run_command_and_return_output(["ioping", "--help"])
    fio = run_command_and_return_output(["fio", "--help"])
    getconf = run_command_and_return_output(["getconf", "--help"])
    if mountpoint is not None and df is not None and nproc is not None and free is not None and ioping is not None and fio is not None and getconf is not None:
        return 0
    return 1

# here we have added this check for forticloud platform where there is not pgsql mount we only have fsr-data mount
def product_type_check():
    product_type_list = ["forticloud,forticloud-secure-message-exchange"]
    try:
        with open("/etc/cyops-release", "r") as file_cyops_release:
            product_type = file_cyops_release.read().strip().split("=")[1]
            if product_type in product_type_list:
                return True
    except FileNotFoundError as file_not_found:
        pass
    return False


# =============================== Code End ================================

def main():
    print(get_system_config())


if __name__ == '__main__':
    main()
