#!/bin/sh
# Copyright start
# Copyright (C) 2008 - 2024 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end

set -e

check_root() {
    if [ $EUID -ne 0 ]; then
        echo "You need to be root to perform this command."
        exit 1
    fi
}

init() {
    openssl_conf="/opt/cyops/certs/leaf.openssl.conf"
    fsr_docker_utils_sh="/opt/cyops/scripts/docker/utils.sh"
    source $fsr_docker_utils_sh
}

restart_services() {
    if [ "X$2" = "X--skip-services-restart" ]; then
        #
        # Since nginx conf files get updated, need at least nginx restart.
        # This is for container where IPV6 gets enabled during boot and we don't restart
        # nginx during container provisioning to show the progress bar on UI.
        #
        echo "Restarting nginx"
        systemctl restart nginx
        return
    fi
    csadm services --restart
}

generate_certs() {
    local curr_hostname=$1
    # To-Do
    # Do only if ::1 is not present in self signed certs
    if ! grep -q "IP\.2" $openssl_conf; then 
        sed -i 's/\s*IP.1\s*=\s*127.0.0.1/IP.1 = ::1\nIP.2 = 127.0.0.1/g' $openssl_conf
    fi

    if ! grep -q "DNS\.2" $openssl_conf; then
        sed -i 's/\s*DNS\.1\s*=\s*fortisoar\.localhost/DNS.1 = fortisoar.localhost\nDNS.2 = localhost/g' $openssl_conf
    fi
    csadm certs --generate $curr_hostname --skip-mq-cert-generation
}

enable_ipv6_in_sysctl() {
    local keys=("net.ipv6.conf.all.disable_ipv6
                net.ipv6.conf.default.disable_ipv6
                net.ipv6.conf.lo.disable_ipv6
                net.ipv6.conf.all.accept_redirects
                net.ipv6.conf.default.accept_redirects
                net.ipv6.route.flush
                ")
    local file=/etc/sysctl.conf
    for key in $keys; do
        grep -q "^\s*$key" /etc/sysctl.conf || echo "$key = 0" >> $file
        sed -i "s/$key =.*$/$key = 0/" $file
    done
    sysctl -p /etc/sysctl.conf
}

clear_api_cache() {
    rm -rf /opt/cyops-api/app/cache/prod
    restorecon -R /opt/cyops-api
    systemctl restart nginx
    sudo -u nginx php /opt/cyops-api/bin/console cache:clear --no-interaction
}

process_confirmation() { 
    local in_msg=$1
    local exit_on_non_aggreement=$2
    local confirmation_msg="Do you want to continue? [yes/no]:"
    local input
    local correct_input=0
    local dash_str=`get_dash_line`
    local error_msg="\n${dash_str}\nError:\n"
          error_msg="${error_msg}Entered value is invalid\n"
          error_msg="${error_msg}${dash_str}"

    while [ "X$correct_input" = "X0" ]; do
        echo -e $in_msg
        read -p "$confirmation_msg" input
        if echo "$input" | grep -iE "^yes$|^no$|^y$|^n$" >/dev/null ;then
            correct_input=1
        else
            echo -e $error_msg
        fi
    done

    if [ "X$exit_on_non_aggreement" = "X1" ]; then
        if echo "$input" | grep -iE "^no$|^n$" >/dev/null ;then
            echo "Exiting..."
            exit 0
        fi
    fi
}

get_dash_line() {
    echo "-------------------------------------------------"
}

take_confirmation_for_ipv6_enablement() {
    local dash_str=$(get_dash_line)
    local note_msg="${dash_str}\nNote:\n"
          note_msg="${note_msg}Running this command will enable IPv6 protocol on your system, which requires\n"
          note_msg="${note_msg}restarting of the network and product services.\n"
          note_msg="${note_msg}${dash_str}"
    process_confirmation "$note_msg" 1
}

reboot_the_system() {
    local dash_str=$(get_dash_line)
    local note_msg="${dash_str}\nNote:\n"
          note_msg="${note_msg}You need to re-run command 'csadm network ipv6 --enable'\n"
          note_msg="${note_msg}after system reboot to complete IPv6 configuration.\n"
          note_msg="${note_msg}${dash_str}"
    echo -e $note_msg
    read -p "Press enter/return to reboot..." opt
    reboot
}

modify_ifcfg() {
    local interfaces=$(ip link | awk -F: '$0 !~ "lo|vir|wl|^[^0-9]"{print $2;getline}' | awk '{$1=$1};1')
    for interface in $interfaces; do
        local ifcfg_file_path="/etc/sysconfig/network-scripts/ifcfg-${interface}"
        if [ ! -f $ifcfg_file_path ]; then
            continue
        fi
        echo "IPV6INIT=yes" >> $ifcfg_file_path
        echo "DHCPV6C=yes" >> $ifcfg_file_path 
    done
}

modify_rc_local() {
    rc_local=/etc/rc.local
    etc_issue=/etc/issue
    etc_issue_net=/etc/issue.net
    # First check whether IPV6 entry is there in rc.local. If NOT, then add it.
    if ! grep -qi "ipv6" $rc_local; then
        # In below command, we are finding ipv6 excluding ::1 and fe80 (link local ipv6)
        # Why we are sleeping for 10 secs: This is required as system need some time to actually get the IPV6 from publisher.
        # Please note that number 10 is try and error basis number got from QA.
        echo "sleep 10" >> $rc_local
        echo "IPV6=\$(ip -6 addr | grep inet6 | awk -F '[ \t]+|/' '{print \$3}' | grep -v ^::1 | grep -v ^fe80)" >> $rc_local
        echo "echo \"IPv6: \$IPV6\" | tee --append $etc_issue $etc_issue_net" >> $rc_local
    fi
}

_enable_ipv6() {
    if ! is_docker; then
        take_confirmation_for_ipv6_enablement
        local reboot_required=0
        if grep -q "ipv6.disable=1" /etc/default/grub; then
            # ipv6 is disabled in grub
            sed -i 's/ipv6.disable=1/ipv6.disable=0/g' /etc/default/grub
            grub2-mkconfig -o /boot/grub2/grub.cfg
            reboot_required=1
        fi
        enable_ipv6_in_sysctl
        modify_ifcfg
        modify_rc_local
        if [ "X$reboot_required" = "X1" ]; then
            reboot_the_system
        else
            systemctl restart NetworkManager
        fi
    fi

    # Update nginx configuration files
    /opt/cyops/scripts/update_nginx_listen_config.sh

    local cyops_conf_files=("/opt/cyops-api/config/parameters_prod.yaml
                       /opt/cyops-workflow/sealab/sealab/config.ini 
                       /opt/cyops-integrations/integrations/configs/config.ini 
                       /opt/cyops/configs/database/db_config.yml 
                       /etc/cyops/config.yml 
                       /opt/cyops/configs/cyops-api/parameter_change.yml 
                       /opt/cyops-auth/utilities/das.ini 
                       /opt/cyops-routing-agent/postman/config.ini
                       /etc/elasticsearch/elasticsearch.yml")

    for conf in $cyops_conf_files; do
        echo "Updating $conf"
        sed -i 's/127.0.0.1/localhost/g' $conf
    done

    #
    # In above call, we are replacing the IP 127.0.0.1 with localhost.
    # For FortiMonitor, we add rule for /nginx_status page to access for only localhost
    # IPs i.e 127.0.0.1 and ::1. Nginx allow only IPs in 'allow' directive.
    # Hence we need to change 'allow localhost' to the 'allow 127.0.0.1', since above
    # call changed 127.0.0.1 to localhost.
    #

    local nginx_api_conf="/etc/nginx/conf.d/cyops-api.conf"
    if [ -f "$nginx_api_conf" ]; then
        sed -i "s/allow localhost/allow 127.0.0.1/" "$nginx_api_conf"
    fi

    if is_docker && grep -q "IP\.2" $openssl_conf; then
        restart_services "$@"
        return
    fi
    local curr_hostname=$(hostname)
    generate_certs $curr_hostname
    clear_api_cache
    restart_services "$@"
}

print_help() {
   local help_type=$1
   local help_one_liner="IPv6 operations"
   if [ "detailed" = "$help_type" ]; then
       echo "$help_one_liner"
       echo "usage: csadm ipv6 [options]"
       echo ""
       echo "subcommand options are:"
       echo "    --enable                    Enable IPv6 protocol. Requires system reboot."
       echo "    --help                      Show this text"
   elif [ "short" = "$help_type" ]; then
       echo "$help_one_liner"
   fi
   exit 0
}

##################################
# MAIN/main/Main starts here
##################################
check_root
# read the input options
if [ $# -lt 1 ]; then
    print_help 'detailed'
fi
init
case "$1" in
    --enable)
        _enable_ipv6 "$@"
        break
        ;;
    *)
        print_help 'detailed'
        break
        ;;
esac
exit 0
