#!/bin/sh
# Copyright start
# Copyright (C) 2008 - 2024 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end

remove_file(){
    local file_to_remove=$1
    if [ -f "$file_to_remove" ]; then
        rm -f $file_to_remove
    fi
}

read_password(){
    read -sp $'Enter the password:\n' password
    if [ "X$password" = "X" ]; then
        exit 1
    fi
    echo "$password"
}

logcollect(){
    local LOGPATH=$1
    local LOG_FILE_PASSWORD=$2

    set -e
    if [ "X$LOGPATH" = "X" ]; then
        echo "Enter path where logs should be collected(if not specified, logs are collected in the current working directory): "
        read LOGPATH
    fi
    
    if [ -z $LOGPATH ]; then
        # User has not entered any value, hence consider PWD
        LOGPATH=$(pwd)
    fi

    if [ ! -d $LOGPATH ]; then
        echo "The directory [$LOGPATH] does not exist. Enter a valid directory path."
        exit 1
    fi

    # User can enter relative path like '.' i.e current directory or ../../. Convert it to absolute path
    LOGPATH=$(cd $LOGPATH; pwd)

    local LOGFILE_NAME="fortisoar-logs"
    local BUILDAREA="$LOGPATH/fsr-log-collect-buildarea/$LOGFILE_NAME"
    local LOGFILE_TAR_GZ="$LOGFILE_NAME.tar.gz"
    local FINAL_LOGFILE="$LOGPATH/$LOGFILE_TAR_GZ"

    if [ "X$LOG_FILE_PASSWORD" = "X" ]; then
        # Input password is not provided, remove the .tar.gz file created earlier
        remove_file $FINAL_LOGFILE
    else
        # Input password is provided, remove the .tar.gz.gpg file created earlier
        remove_file $FINAL_LOGFILE.gpg
    fi

    if [ -d $BUILDAREA ]; then
        rm -rf $BUILDAREA
    fi

    mkdir -p $BUILDAREA

    logpaths="/var/log/cyops \
              /var/log/nginx \
              /var/log/elasticsearch \
              /var/log/messages* \
              /var/log/audit \
              /var/log/rabbitmq \
              /var/log/php-fpm/ \
              /opt/cyops-tomcat/logs/ \
              /var/log/maillog* \
              /var/log/fm-agent/ \
              /var/log/leapp/"

    # 2> /dev/null will ignore any warnings related to non-existing path.
    available_size=$(df -B1 $BUILDAREA --output=avail | tail -1)
    threshold_size=$(echo "$available_size*0.8" | bc)
    threshold_size_int=$(printf "%.0f" $threshold_size)

    logs_size=$(du -sh --total -b $logpaths 2> /dev/null |  tail -1  | awk '{print $1}')
    tar_size=$(tar -czf - $logpaths 2> /dev/null | wc -c)
    gpg_size=$tar_size

    # the directory where all logs are collected before compression can be deleted when the password protection is enabled
    size_required_for_tar=$((logs_size+tar_size))
    size_required_for_gpg=$((tar_size+gpg_size))
    required_size=$size_required_for_tar

    if [ "X$LOG_FILE_PASSWORD" != "X" ]; then
        if [ $size_required_for_gpg -gt $size_required_for_tar ]; then
            required_size=$size_required_for_gpg
        fi
    fi

    echo "$(date) Total logs size: $(numfmt --to=iec $logs_size)"
    echo "$(date) Available size: $(numfmt --to=iec $available_size)"
    echo "$(date) Compressed log file size: $(numfmt --to=iec $tar_size)"
    echo "$(date) Required size: $(numfmt --to=iec $required_size)"
    echo "$(date) Threshold size: $(numfmt --to=iec $threshold_size_int)"

    if [ "$required_size" -gt "$threshold_size_int" ]; then
        echo "$(date) not enough space available"
        exit 1
    fi

    echo "$(date) Collecting logs"
    for item in $logpaths
    do
        if [ -f $item ] || [ -d $item ]; then
            cp -r --parents $item $BUILDAREA
        fi
    done


    tar czf $(dirname $BUILDAREA)/$LOGFILE_TAR_GZ -C $(dirname $BUILDAREA) $LOGFILE_NAME/

    if [ "X$LOG_FILE_PASSWORD" = "X" ]; then
        # Input password is not provided
        mv $(dirname $BUILDAREA)/$LOGFILE_TAR_GZ $FINAL_LOGFILE
        echo "$(date) Logs collected at: "$FINAL_LOGFILE
        rm -rf $(dirname $BUILDAREA)
        return
    fi
    rm -rf $BUILDAREA

    # Input password is provided, make the gpg with input password

    gpg --yes --batch --passphrase=$LOG_FILE_PASSWORD -o $FINAL_LOGFILE.gpg --cipher-algo AES256 --symmetric $(dirname $BUILDAREA)/$LOGFILE_TAR_GZ

    rm -rf $(dirname $BUILDAREA)

    echo "$(date) Logs collected at: "$FINAL_LOGFILE.gpg
}

subcmd_help(){
    local help_type=$1
    local s_help_one_liner="Collect logs"
    if [ "detailed" = "$help_type" ]; then
        echo "usage: csadm log [-h] [--collect [LOG_PATH]] [--password [LOG_FILE_PASSWORD]]"
        echo ""
        echo "subcommand options are:"
        echo "  --collect [LOG_PATH]              Perform log collection."
        echo "                                    Collected logs are bundled into fortisoar-logs.tar.gz"
        echo "  --password [LOG_FILE_PASSWORD]    Password-protects the log file,i.e., users require the password to extract"
        echo "                                    the log file. Collected logs are bundled into fortisoar-logs.tar.gz.gpg"
        echo "  -h, --help                        show this help message and exit"
    elif [ "short" = "$help_type" ]; then
        echo "$s_help_one_liner"
    fi
}

subcmd_implementation(){
    case "$1" in
        --collect)
            # $3 will be --password and $4 will be password input
            local input_password=$4
            local input_log_file_path=$2
            if [ "X$2" = "X--password" ]; then
                # Since --collect is optional argument, we need this if condition.
                input_log_file_path=""
                input_password=$3
            fi
            if [ "X$input_password" = "X" ]; then
                # Input password is not provided. Now it is required that whether --password was
                # provided in input or not. If yes, then we need to read the password.
                if [ "X$2" = "X--password" ] || [ "X$3" = "X--password" ]; then
                    input_password=$(read_password)
                    if [ $? -ne 0 ]; then
                        # Password is empty
                        echo "Password must not be empty"
                        exit 1
                    fi
                fi
            fi
            logcollect "$input_log_file_path" $input_password
            break ;;

        -h|--help)
            subcmd_help "detailed"
            break ;;

        *)
            subcmd_help "detailed"
            break ;;
    esac
}

main(){
    subcmd_implementation "$@"
}

main "$@"
exit 0
