#!/bin/sh
# Copyright start
# Copyright (C) 2008 - 2024 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end

# The script performs:
# 1. Addition of newly attached disks (if added)
# 2. Entension of the partitions (in of case free disk space)
#   to lvdata logical volume

set -x

init(){
    # pgsql, es and mq are mounted from logical volume: lvdata
    s_logical_volume="lvdata"

    # Get the full path of the logical volume
    s_logical_volume_path=`df -kh | grep -w $s_logical_volume | awk '{print $1}'`

    # Get the file system path on which lvdata is mounted
    s_filesystem_lvdata=`df -kh | grep -w $s_logical_volume | awk '{print $6}'`

    i_free_disk_space_percent_threshold=5

    flag_extend_logical_volume=0
}

add_new_disks(){
    # Overall logic:
    # Run disk scan command
    # Get the list of all disks
    # Loop through the list
    #   its new disk if its entry is not found in the output of pvs command
    # prepare the list of new disks
    # Get name of the disk. Ex: sdi
    # Loop through each disk
    #   partition the disk
    #       parted --script --align optimal /dev/sdi mklabel gpt mkpart primary 0% 100%
    #       o/p: partioned /dev/sdi1
    #   create physical volume
    #       pvcreate /dev/sdi1
    #       o/p: /dev/sdi1
    #       add newly created physical volume to existing volume group
    #           vgextend vgdata /dev/sdi1
    #           o/p: vgdata with additional disk space

    # Scan each scsi host device for new disks
    for host in /sys/class/scsi_host/*; do
        echo "- - -" >> $host/scan
    done

    # Get the list of disks
    l_disks=`lsblk -o KNAME,TYPE | \
             grep disk | \
             awk '{print $1}'`

    # Loop through the list
    for s_disk in $l_disks; do
        #   its new disk if its entry is not found in the output of pvs command
        pvs | grep /dev/$s_disk
        if [ $? -ne 0 ]; then
            l_new_disks="$l_new_disks /dev/$s_disk"

            # Set the flag to extend the logical volume
            flag_extend_logical_volume=1
        fi
    done

    # Loop through each disk
    for s_new_disk_device in $l_new_disks ; do
        # Get the disk name from device name
        s_new_disk_name=`echo $s_new_disk_device | sed "s#/dev/\(.*\)#\1#g"`

        # Create the partition
        parted --script \
               --align optimal \
               $s_new_disk_device \
               mklabel gpt \
               mkpart primary 0% 100%

        sleep 5

        # Create the physical volume
        s_new_physical_volume=${s_new_disk_device}1
        pvcreate $s_new_physical_volume

        # Add the newly created physical volume to existing volume group
        vgextend vgdata $s_new_physical_volume
    done
}

extend_existing_partitions(){
    # Get the list of partition names of the volume group on which lvdata is mounted
    l_existing_partitions=`lvs --segments $s_logical_volume_path -o +devices | \
                          grep $s_logical_volume | \
                          awk '{print $7}' | \
                          cut -d\( -f1 | \
                          uniq`

    for s_existing_partition in $l_existing_partitions; do

        # Get the disk name of $s_existing_partition
        s_disk_name=`echo $s_existing_partition | head -c -2`

        # Get the partition number
        i_partition_number=`echo $s_existing_partition | grep -Eo '[0-9]+$'`

        # Get physical volume name
        s_physical_volume=$s_disk_name$i_partition_number

        # Calculate the free disk percentage on partition
        i_free_disk_space_percent=`parted $s_disk_name unit '%' print free | \
                                            grep 'Free Space' | \
                                            tail -n1 | \
                                            awk '{print $3}' | \
                                            cut -d. -f1`

        # Extend the LVM only if free disk space on the partition is more than set percentage threshold
        if [ $i_free_disk_space_percent -gt $i_free_disk_space_percent_threshold -a \
             ! -z "$s_logical_volume_path" ];then
            # Grow the partition to its maximum available space
            growpart $s_disk_name $i_partition_number

            # Grow the physical volume to fully occupy the partition
            pvresize $s_physical_volume

            # Set the flag to extend the logical volume
            flag_extend_logical_volume=1
        fi
    done
}

# Extend the logical volume to 100% and
# grow the filesystem to 100%
extend_logical_volume(){
    # Extend the logical volume to fully occupy the physical volume
    lvextend -l +100%FREE $s_logical_volume_path

    # Grow the filesystem to fully occupy the logical volume
    xfs_growfs $s_filesystem_lvdata
}

#######
# Main
#######

init

extend_existing_partitions

add_new_disks

if [ $flag_extend_logical_volume -eq 1 ];then
    extend_logical_volume
fi

exit
